
#include "cmbcosmo.h"
#include "fatalerror.h"

#include <cmath>
#include <iostream>


using namespace std;

static double sq(const double &x) { return x*x;}
static double cub(const double &x) { return x*x*x;}

CmbCosmo::CmbCosmo(const Cosmology &C, const double OmegabH2, const double H0) : Cosmology(C.Omegak()), obh2(OmegabH2), h0(H0)
{
  actualCosmo = C.Clone();
  double omegar = 2.469e-5*(1+0.2271/* extra contribution from nu's */ *3.04 /* N_nu */)/(h0*h0/10000);
  actualCosmo->SetOmegar(omegar);
  *(Cosmology *)this = *actualCosmo;
}

void CmbCosmo::dump(std::ostream &stream) const
{
  stream << " CmbCosmo on ";actualCosmo->dump(stream);
}

double CmbCosmo::zstar() const
{// eq 66,67,68 Komatsu 2009
  double omegamh2 = Omegam()*sq(h0/100);
  double g1=0.0783*pow(obh2,-0.238)/(1.+ 39.5*pow(obh2,0.763));
  double g2 = 0.560/(1+21.1*pow(obh2,1.81));
  return 1048*(1.+0.00124*pow(obh2,-0.738))*(1+g1*pow(omegamh2,g2));
}


double CmbCosmo::zdrag() const
{ // eq 4 from Eisenstein Hu, 1998
  
  

  double omegamh2 = Omegam()*sq(h0/100);
  double b1=0.313*pow(omegamh2,-0.419)*(1+0.607*pow(omegamh2,0.674));
  double b2 = 0.238*pow(omegamh2,0.223);
  return 1291*(1.+ b1*pow(obh2,b2))* pow(omegamh2,0.251)/
    (1+0.659*pow(omegamh2,0.828));
}


/* it actually misses the c/H0 prefactor */
double CmbCosmo::rs(const double Z) const
{// eq 6,7,8
  double amax= 1./(1+Z);
  double amin=0;
  int nstep = 100; 
  double step = (amax-amin)/nstep;
  double a=amin+0.5*step;
  double sum = 0;
  double density_ratio = 3*obh2/sq(h0/100)/(4.*omegar/(1+0.2271*3.04));
  /* Dark energy is supposed to be subdominant before recombination. 
     However, rather than neglecting it, we consider its eos to be constant.
  */
  double w = W(1100);
  if (w>1) w= 1;
  double omegam = Omegam();
  double omegak = Omegak();
  double omegar = Omegar();
  double omegax = 1-omegam-omegak;
  
  // set omegax = 0 for now
#warning rs with omega_DE = 0;
  omegax = 0;
  for (int i=0; i<nstep ; ++i , a+= step)
    {
      double a4h2 = (omegam+omegak*a)*a+omegar+omegax*pow(a,1-3.*w);
      double fact = 1.+density_ratio*a;
      sum += 1./sqrt(a4h2*fact);
    } 
  
  return sum*step/sqrt(3.);
}
  
double CmbCosmo::l_a(const double Z,const int NStep) const
{
  return M_PI*Dm(Z,NStep)/rs(Z);
}

double CmbCosmo::R(const double Z,const int NStep) const  
{
  return sqrt(Omegam())*Dm(Z,NStep);
}

#ifdef STORAGE
#include <cstring> // for strcmp
#include <cstdlib> // for abort

int CmbCosmo::GetParamRank(const char *Name) const
{
  /* check compatibility with routines above ! */
  static const char *names[6]={"Omegam", "Omegal", "w0", "wa","OmegabH2", "H0"};
  for (int k=0; k<6; ++k)
    if (strcmp(Name,names[k]) ==0) return k;
  FatalError("CmbCosmo::GetParamRank,unknown param : "+string(Name));
  abort();
}
#endif

void CmbCosmo::SetParam(const int IPar, const double Val)
{ 
  actualCosmo->SetParam(IPar, Val); 
  *(Cosmology *)this=*actualCosmo;
} 
