#include <cmath>
#include <assert.h>

#include "cosmology.h"
#include "cosmology_prior.h"
#include "matvect.h"
#include "snfitexception.h"
#include "neutrinocosmo.h"

static double sq(const double &x) {return x*x;}

// priors
//===============================================================



Mat Eisenstein2005_Prior::Weight(const Cosmology *Cosmo) const
{
  Mat W(1,1);
  W(0,0)= 1./sq(0.017);
  return W;
}
//! the equation 4 and 5 of astro-ph 0501117  
Vect Eisenstein2005_Prior::Residual(const Cosmology *Cosmo) const
{
  double z1 = 0.35;
  double A=sqrt(Cosmo->Omegam())*pow(Cosmo->E(z1),-1./3.)*
    pow(Cosmo->Dm(z1)/z1,2./3.);
  Vect R(1);
  R(0) = A-0.469;
  return R;
}

/***********************  OmegaM_Prior   ***************/ 

Mat OmegaM_Prior::Weight(const Cosmology *Cosmo) const
{
  Mat W(1,1);
  W(0,0)= 1./sq(omega_m_error);
  return W;
}

Vect OmegaM_Prior::Residual(const Cosmology *Cosmo) const
{
  Vect R(1);
  R(0) = Cosmo->Omegam()-omega_m_val;
  return R;
}

double Prior::Chi2(const Cosmology *Cosmo) const
{
  Vect R = Residual(Cosmo);
  return scalar_product(R,Weight(),R);
}


#include "cmbcosmo.h"






using namespace std;


// utility

static const CmbCosmo* get_cmbcosmo(const Cosmology* C, bool Fatal=true)
{
  const CmbCosmo *c = dynamic_cast<const CmbCosmo *>(C);
  if (Fatal && !c)
    throw(SnfitException(string(" in ")+__FILE__+", need a CmbCosmo"));
  return c;
}
      


/**************** Wmap5CosmoPrior *********************/

/* Five-Year Wilkinson Microwave Anisotropy Probe (WMAP) Observations: 
   Cosmological Interpretation
Komatzu et al, section 5.4

Accepted paper but no journal ref yet (dec 2008)

http://cdsads.u-strasbg.fr/abs/2008arXiv0803.054
http://lambda.gsfc.nasa.gov/product/map/dr3/pub_papers/fiveyear/cosmology/wmap_5yr_cosmo.pdf

P.A.


*/



Mat Wmap5CosmoPrior::Weight(const Cosmology *cosmo) const
{// table 11 of Komatsu et al
  Mat W(3,3);
  W(0,0) = 1.800;
  W(1,0) = W(0,1) = 27.968;
  W(2,0) = W(0,2) = -1.103;
  W(1,1) = 5667.577;
  W(2,1) = W(1,2) = -92.263;
  W(2,2) = 2.923;
  return W;
}

Vect Wmap5CosmoPrior::Residual(const Cosmology *Cosmo) const
{// data from table 10 of Komatsu et al
  
  const CmbCosmo *c= get_cmbcosmo(Cosmo);

  Vect r(3);
  double zstar = c->zstar();
  r(0) = c->l_a(zstar)- 302.10;
  r(1) = c->R(zstar) - 1.710;
  r(2) = zstar - 1090.04;
  return r;
}


/**************** Wmap7CosmoPrior *********************/
/*
Next round : WMAP 7
Komatsu et al : 1001.4538
p26 for the distance prior.
http://fr.arxiv.org/abs/1001.4538
*/



Mat Wmap7CosmoPrior::Weight(const Cosmology *cosmo) const
{// table 9 p26 (1001.4538)
  Mat W(3,3);
  W(0,0) = 2.305;
  W(1,0) = W(0,1) = 29.698;
  W(2,0) = W(0,2) = -1.333;
  W(1,1) = 6825.270;
  W(2,1) = W(1,2) = -113.180;
  W(2,2) = 3.414;
  return W;
}

Vect Wmap7CosmoPrior::Residual(const Cosmology *Cosmo) const
{
  // data from table 9 of Komatsu et al
  //chosen first column
  
  const CmbCosmo *c= get_cmbcosmo(Cosmo);

  Vect r(3);
  double zstar = c->zstar();
  r(0) = c->l_a(zstar)- 302.09;
  r(1) = c->R(zstar) - 1.725;
  r(2) = zstar - 1091.3;
  return r;
}




/**************** Wmap9CosmoPrior *********************/
/*
Next round : WMAP 9
Hinshaw et al. 2012 : 1212.5256v1
submitted to ApJs
p21 for the distance prior.
http://adsabs.harvard.edu/abs/2012arXiv1212.5226H
*/



Mat Wmap9CosmoPrior::Weight(const Cosmology *cosmo) const
{// table 11 p21 (1212.5256)
  Mat W(3,3);
  W(0,0) = 3.182;
  W(1,0) = W(0,1) = 18.253;
  W(2,0) = W(0,2) = -1.429;
  W(1,1) = 11887.879;
  W(2,1) = W(1,2) = -193.808;
  W(2,2) = 4.556;
  return W;
}

Vect Wmap9CosmoPrior::Residual(const Cosmology *Cosmo) const
{
  // Value given for Eq. 17 in Hinshaw et al. (2012)
  //chosen first column
  
  const CmbCosmo *c= get_cmbcosmo(Cosmo);

  Vect r(3);
  double zstar = c->zstar();
  r(0) = c->l_a(zstar)- 302.40;
  r(1) = c->R(zstar) - 1.7246;
  r(2) = zstar - 1090.88;
  return r;
}


/****************   Percival10Prior (BAODR7) ***********************/
//http://cdsads.u-strasbg.fr/abs/2010MNRAS.401.2148P


static double sqr(const double &x) { return x*x;}

Vect  BAODR7Prior::Residual(const Cosmology *Cosmo) const
{
  /*  d_z $B"a(B r_s (z_d )/DV (z);
  DV (z) $B"a(B [(1 +z)2 D**2 cz/H (z)]**(1/3)
  */
  const CmbCosmo *c= get_cmbcosmo(Cosmo);

  double rs = c->rs(c->zdrag());  
  double z1 = 0.2;
  double z2 = 0.35;
  // definition of Dv page 2152, bottom of first column.
  /* ignore the  c/H0 prefactor because CmbCosmo::rs does the same */
  double dv1 = pow(sqr(c->Dm(z1))*z1/c->E(z1),1./3.);
  double dv2 = pow(sqr(c->Dm(z2))*z2/c->E(z2),1./3.);
  Vect res(2);
  // values from eq 3 (p 2155)
  // checked that vanilla cosmology yields values reasonnably close to those.
  res(0) = rs/dv1-0.1905;
  res(1) = rs/dv2-0.1097;
  return res;
}


Mat  BAODR7Prior::Weight(const Cosmology *Cosmo) const
{
  // eq 5 p 2155
  Mat w(2,2);
  w(0,0) = 30124;
  w(0,1) = w(1,0) = -17227;
  w(1,1) = 86977;
  return w;
}

/* Anderson et al. 2012 CMASS sample at z~0.57, Dv(0.57)/rs=13.67+-0.22, independent from BAODR7Prior */

Vect  BAODR9Prior::Residual(const Cosmology *Cosmo) const
{
  const CmbCosmo *c= get_cmbcosmo(Cosmo);

  double rs = c->rs(c->zdrag());  
  double z = 0.57;
  double dv = pow(sqr(c->Dm(z))*z/c->E(z),1./3.);
  Vect res(1);
  res(0) = rs/dv-1./13.67;
  return res;
}
Mat  BAODR9Prior::Weight(const Cosmology *Cosmo) const
{
  Mat w(1,1);
  w(0,0) = 1./sqr(0.22/sqr(13.67)); // residual is rs/dv , not dv/rs
  return w;
}
/* Sanchez et al. 2012 CMASS sample at z~0.57, A(zm)=0.444 +- 0.014, to used INSTEAD of BAODR9Prior when no CMB  */

static double A_for_BAO(const double& z, const Cosmology *Cosmo);

Vect  BAODR9Sanchez2012Prior::Residual(const Cosmology *Cosmo) const
{
  Vect res(1);
  res(0) = A_for_BAO(0.57,Cosmo)-0.444;
  return res;
}

Mat  BAODR9Sanchez2012Prior::Weight(const Cosmology *Cosmo) const
{
  Mat w(1,1);
  w(0,0) = 1./sqr(0.014); // 
  return w;
}



  
/****************   WiggleZ  ***********************/
// 

static double Dv_for_BAO(const double& z, const Cosmology *Cosmo) {
  return pow(sqr(Cosmo->Dm(z))*z/Cosmo->E(z),1./3.);
}

static double A_for_BAO(const double& z, const Cosmology *Cosmo) {
  return Dv_for_BAO(z,Cosmo)*sqrt(Cosmo->Omegam())/z;
}

Vect WiggleZPrior::Residual(const Cosmology *Cosmo) const
{
  
  // wigglez Blake 2011. table 3
  Vect res(3);
  const double z[]={0.44,0.6,0.73};
  const double a[]={0.474,0.442,0.424};
  for(int i=0;i<3;i++) {
    res(i)=A_for_BAO(z[i],Cosmo)-a[i];
    if(isnan(res(i))) res(i)=1.e10;
  }
  return res;
}

Mat  WiggleZPrior::Weight(const Cosmology *Cosmo) const
{
  // weight matrix is Blake 2011. table 2
  Mat w(3,3);
  w(0,0) = 1040.3; w(0,1) = -807.5; w(0,2) = 336.8; 
  w(1,1) = 3720.3; w(1,2) = -1551.9;
  w(2,2) = 2914.9;

  w(1,0) = w(0,1);
  w(2,0) = w(0,2);
  w(2,1) = w(1,2);
  
  return w;
}
    
Vect SixdFGSPrior::Residual(const Cosmology *Cosmo) const
{
  
  //  Blake 2011. table 3
  Vect res(1);
  res(0) = A_for_BAO(0.106,Cosmo)-0.526;
  return res;
}



Mat  SixdFGSPrior::Weight(const Cosmology *Cosmo) const
{
  // Blake 2011. table 3
  Mat w(1,1);
  w(0,0) = 1./sqr(0.028); 
  return w;
}
 


/* Shoes prior is the H0 prior from the H maser stuff 
   http://cdsads.u-strasbg.fr/abs/2009ApJ...699..539R
   H 0 = 74.2 ± 3.6 km s-1 Mpc-1,
*/
Vect ShoesPrior::Residual(const Cosmology *Cosmo) const
{
  const CmbCosmo *c= get_cmbcosmo(Cosmo, /* fatal = */ false);
  Vect res(1);
  if (!c) return res;
  res(0) = c->H0()-74.2;
  return res;
}


Mat ShoesPrior::Weight(const Cosmology *Cosmo) const
{
  Mat w(1,1);
  w(0,0) = 1./sqr(3.6);
  return w;
}

/* Riess 2011
   http://cdsads.u-strasbg.fr/abs/2011ApJ...730..119R
   H0 = 73.8 +- 2.4 km s−1 Mpc−1
*/
Vect H0Riess2011Prior::Residual(const Cosmology *Cosmo) const
{
  const CmbCosmo *c= get_cmbcosmo(Cosmo, /* fatal = */ false);
  Vect res(1);
  if (!c) return res;
  res(0) = c->H0()-73.8;
  return res;
}


Mat H0Riess2011Prior::Weight(const Cosmology *Cosmo) const
{
  Mat w(1,1);
  w(0,0) = 1./sqr(2.4);
  return w;
}


/////////////////////////////////////////////////////////////////////
/* 
   ONLY FOR SIMULATION TESTS !! 
   astro-ph 0803.1616v1 : R at 0.32%   R = sqrt(Omegam*h^2)*r(z_cmb). 
*/
#define NSTEP_FOR_PlanckPrior 200
LightPlanckPrior::LightPlanckPrior() {
  double om_ref = 0.27;
  GeneralCosmo cosmo(om_ref,1-om_ref,-1,0);
  z_ref  = 1088;
  R_ref = sqrt(om_ref)*cosmo.Dm(z_ref,NSTEP_FOR_PlanckPrior);
}

Vect LightPlanckPrior::Residual(const Cosmology *cosmo) const {
  // R = sqrt(Omegam*h^2)*r(z_cmb)
  // relative value R(cosmo)/R(ref)-1
  Vect R(1);
  R(0) = sqrt(cosmo->Omegam())*cosmo->Dm(z_ref,NSTEP_FOR_PlanckPrior)/R_ref-1;
  return R;
}
  
Mat LightPlanckPrior::Weight(const Cosmology *Cosmo) const {
  Mat W(1,1);
  W(0,0) = 1/sqr(0.0032);
  return W;
}
/////////////////////////////////////////////////////////////////////

FullPlanckPrior::FullPlanckPrior() {
  CmbCosmo cosmo(GeneralCosmo(0.27,0.73,-1,0),0.02199,70.);
  double zstar = cosmo.zstar();
  R_ref = cosmo.R(zstar);
  la_ref = cosmo.l_a(zstar);
  obh2_ref = cosmo.OBH2();

  W.allocate(3,3);
  W(0,0) = 0.303492e-04;
  W(1,1) = 0.951881e-02;
  W(0,1) = 0.297688e-03;
  W(2,2) = 0.279464e-07;
  W(0,2) = -0.545532e-06; 
  W(1,2) = -0.759752e-05;
  
  W(1,0) = W(0,1);
  W(2,0) = W(0,2);
  W(2,1) = W(1,2);
  
  W.CholeskyInvert("U");
  W.Symmetrize("U");
}

Vect FullPlanckPrior::Residual(const Cosmology *Cosmo) const
{
  
  const CmbCosmo *c= get_cmbcosmo(Cosmo);
  // R, la, omegabh2, based on table I and II of astro-ph 0803.1616v1
  Vect r(3);
  double zstar = c->zstar();
  r(0) = c->R(zstar) - R_ref;
  r(1) = c->l_a(zstar)- la_ref;
  r(2) = c->OBH2() - obh2_ref;
  return r;
}

Mat FullPlanckPrior::Weight(const Cosmology *Cosmo) const
{
  return W;
}
/////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////
/* from email of J. Marriner 212-05-24
   1                      w_0   (not w_p!)
   2                      w_a
   3                      Omega_{D.E.}
   4                      Omega_K
   5                      Omega_{mh^2}
   6                      Omega_b h^2
   7                      n_s
   8                      ln P       ( here P=A=(1.84 delta_ksi*10^4)^2 )
 */
DETFStageIIPrior::DETFStageIIPrior() {
  
  OmegamOmegaxW0WaCosmo cosmo(0.27,0.73,-1,0);
  
  CmbCosmo c(cosmo,0.02199,70.);
  ref.allocate(6);
  ref(0) = cosmo.W0();
  ref(1) = cosmo.Wa();
  ref(2) = cosmo.Omegax();
  ref(3) = cosmo.Omegak();
  ref(4) = cosmo.Omegam()*sqr(c.H0()/100);
  ref(5) = c.OBH2();
  
  Mat Wall(8,8); // from an email of J. Marriner 212-05-24
  
  Wall(0,0)=248439;
  Wall(0,1)=70417.1;
  Wall(0,2)=845764;
  Wall(0,3)=-2.69057e+006;
  Wall(0,4)=3.90626e+006;
  Wall(0,5)=-8.72762e+006;
  Wall(0,6)=-46969.7;
  Wall(0,7)=-3230.83;
  Wall(1,0)=70417.1;
  Wall(1,1)=19967.5;
  Wall(1,2)=239909;
  Wall(1,3)=-763681;
  Wall(1,4)=1.10903e+006;
  Wall(1,5)=-2.4776e+006;
  Wall(1,6)=-13253.8;
  Wall(1,7)=-862.033;
  Wall(2,0)=845764;
  Wall(2,1)=239909;
  Wall(2,2)=2.88999e+006;
  Wall(2,3)=-9.17943e+006;
  Wall(2,4)=1.33398e+007;
  Wall(2,5)=-2.97948e+007;
  Wall(2,6)=-159698;
  Wall(2,7)=-10455.6;
  Wall(3,0)=-2.69057e+006;
  Wall(3,1)=-763681;
  Wall(3,2)=-9.17943e+006;
  Wall(3,3)=2.96045e+007;
  Wall(3,4)=-4.31792e+007;
  Wall(3,5)=9.64792e+007;
  Wall(3,6)=470018;
  Wall(3,7)=1506.08;
  Wall(4,0)=3.90626e+006;
  Wall(4,1)=1.10903e+006;
  Wall(4,2)=1.33398e+007;
  Wall(4,3)=-4.31792e+007;
  Wall(4,4)=6.51125e+007;
  Wall(4,5)=-1.36725e+008;
  Wall(4,6)=-197282;
  Wall(4,7)=4777.9;
  Wall(5,0)=-8.72762e+006;
  Wall(5,1)=-2.4776e+006;
  Wall(5,2)=-2.97948e+007;
  Wall(5,3)=9.64792e+007;
  Wall(5,4)=-1.36725e+008;
  Wall(5,5)=3.76464e+008;
  Wall(5,6)=1.55231e+006;
  Wall(5,7)=-79277.1;
  Wall(6,0)=-46969.7;
  Wall(6,1)=-13253.8;
  Wall(6,2)=-159698;
  Wall(6,3)=470018;
  Wall(6,4)=-197282;
  Wall(6,5)=1.55231e+006;
  Wall(6,6)=213636;
  Wall(6,7)=1385.29;
  Wall(7,0)=-3230.83;
  Wall(7,1)=-862.033;
  Wall(7,2)=-10455.6;
  Wall(7,3)=1506.08;
  Wall(7,4)=4777.9;
  Wall(7,5)=-79277.1;
  Wall(7,6)=1385.29;
  Wall(7,7)=11019.3;
  
  Wall.CholeskyInvert("U");
  Wall.Symmetrize("U");
  // now a covariance
  W = Wall.SubBlock(0,5,0,5);
  W.CholeskyInvert("U");
  W.Symmetrize("U");
  // now a weight matrix again
}



Vect DETFStageIIPrior::Residual(const Cosmology *Cosmo) const
{
  const CmbCosmo *c= get_cmbcosmo(Cosmo);
  const OmegamOmegaxW0WaCosmo * cosmo = dynamic_cast<const OmegamOmegaxW0WaCosmo *>(c->ExtractCosmo());
  if(!cosmo) {
    cout << "FATAL DETFStageIIPrior can only deal with OmegamOmegaxW0WaCosmo or OmegamW0WaCosmo" << endl;
    cout << *Cosmo << endl;
    abort();
  }
 

  Vect r(6);
  r(0) = cosmo->W0();
  r(1) = cosmo->Wa();
  r(2) = cosmo->Omegax();
  r(3) = cosmo->Omegak();
  r(4) = cosmo->Omegam()*sqr(c->H0()/100);
  r(5) = c->OBH2();
  r -= ref;
  return r;
}

Mat DETFStageIIPrior::Weight(const Cosmology *Cosmo) const
{
  return W;
}
/////////////////////////////////////////////////////////////////////

/*
  Email of Rahul 2012-05-25
  Hi All,

  We did the calculations a little differently from the way Julien mentioned, but I agree that the main difference should mostly be due to the choice of the prior. So, assuming that Julien is only trying to get a rough match, let us discuss the prior here. (I would be happy to elaborate on the details of the calculations if people are interested).
  The prior (Stage II + Planck), I was supplied was a 4 by 4 Fisher matrix, attached here as PLANCKFisherHu.txt . 
  
  1 = Omega_DE
  2 = w0
  3 = wa 
  4 = Omega_K
  
We did not assume a flat universe, but yes, this is much more constraining than a Planck prior. 

This is essentially equivalent to the matrix John sent, except for ...

61937.78684808781	17311.727012864303	4803.949386846843	-81675.61386615083
17311.7270128643	5748.593002851787	1545.312687098513	-25989.62996183936
4803.94938684684	1545.3126870985125	422.4515286555771	-7329.538156366885
-81675.6138661508	-25989.629961839364	-7329.538156366886	234893.15678664352
 */



DETFStageIIPriorOfRahul::DETFStageIIPriorOfRahul() {
  
  OmegamOmegaxW0WaCosmo cosmo(0.27,0.73,-1,0);
  
  ref.allocate(4);
  ref(0) = cosmo.Omegax();
  ref(1) = cosmo.W0();
  ref(2) = cosmo.Wa();
  ref(3) = cosmo.Omegak();
  
  W.allocate(4,4); // from an email of Rahul 2012-05-25
  
  W(0,0)=61937.78684808781;	W(1,0)=17311.727012864303;	W(2,0)=4803.949386846843;	W(3,0)=-81675.61386615083;
  W(0,1)=17311.7270128643;	W(1,1)=5748.593002851787;	W(2,1)=1545.312687098513;	W(3,1)=-25989.62996183936;
  W(0,2)=4803.94938684684;	W(1,2)=1545.3126870985125;	W(2,2)=422.4515286555771;	W(3,2)=-7329.538156366885;
  W(0,3)=-81675.6138661508;	W(1,3)=-25989.629961839364;	W(2,3)=-7329.538156366886;	W(3,3)=234893.15678664352;
  
  // re-symmetrize for numerical diff.
  W.Symmetrize("U");
  
}



Vect DETFStageIIPriorOfRahul::Residual(const Cosmology *Cosmo) const
{
  
  
  const OmegamOmegaxW0WaCosmo * cosmo = dynamic_cast<const OmegamOmegaxW0WaCosmo *>(Cosmo);
  if(!cosmo) {
    cout << "FATAL DETFStageIIPrior can only deal with OmegamOmegaxW0WaCosmo or OmegamW0WaCosmo" << endl;
    cout << *Cosmo << endl;
    abort();
  }
 

  Vect r(4);
  r(0) = cosmo->Omegax();
  r(1) = cosmo->W0();
  r(2) = cosmo->Wa();
  r(3) = cosmo->Omegak();
  
  r -= ref;
  return r;
}

Mat DETFStageIIPriorOfRahul::Weight(const Cosmology *Cosmo) const
{
  return W;
}
/////////////////////////////////////////////////////////////////////

/*

  paramètres

0  n      
1  omh2    
2  obh2    
3  okh2    
4  odeh2   
5  w0
6  wa 

on va virer n

 */

HutererPlanckPrior::HutererPlanckPrior() {

  cout << "HutererPlanckPrior::HutererPlanckPrior" << endl;

  OmegamOmegaxW0WaCosmo cosmo(0.27,0.73,-1,0);
  CmbCosmo c(cosmo,0.02199,70.);
  c.ns = 0.963;
  c.delta = -19.9628;
  double h = c.H0()/100;  
  double h2 = h*h;
  
  ref.allocate(8);
  int i=0;
  ref(i++) = c.ns;
  ref(i++) = cosmo.Omegam()*h2;
  ref(i++) = c.OBH2();
  ref(i++) = cosmo.Omegak()*h2;
  ref(i++) = cosmo.Omegax()*h2;
  ref(i++) = c.delta;
  ref(i++) = cosmo.W0();
  ref(i++) = cosmo.Wa();
  
  
  W.allocate(8,8);

  W(0,0)=199164.965680;
W(0,1)=11824.095037;
W(0,2)=1354962.152108;
W(0,3)=1091332.711880;
W(0,4)=-70694.206920;
W(0,5)= -932.302866;
W(0,6)=-37327.950312;
W(0,7)=-10482.784385;
W(1,0)=11824.095037;
W(1,1)=31925826.303230;
W(1,2)=-87057799.089560;
W(1,3)=-68299258.874580;
W(1,4)=4424280.411313;
W(1,5)=-9694.912793;
W(1,6)=2336108.240765;
W(1,7)=656047.781413;
W(2,0)=1354962.152108;
W(2,1)=-87057799.089560;
W(2,2)=342661772.130900;
W(2,3)=210221174.992500;
W(2,4)=-13617679.633550;
W(2,5)=12943.912536;
W(2,6)=-7190406.270515;
W(2,7)=-2019277.188837;
W(3,0)=1091332.711880;
W(3,1)=-68299258.874580;
W(3,2)=210221174.992500;
W(3,3)=157215540.155300;
W(3,4)=-10184087.589300;
W(3,5)= 5505.208449;
W(3,6)=-5377401.233701;
W(3,7)=-1510132.145239;
W(4,0)=-70694.206920;
W(4,1)=4424280.411313;
W(4,2)=-13617679.633550;
W(4,3)=-10184087.589300;
W(4,4)=659703.486844;
W(4,5)= -356.615669;
W(4,6)=348336.590090;
W(4,7)=97823.141550;
W(5,0)= -932.302866;
W(5,1)=-9694.912793;
W(5,2)=12943.912536;
W(5,3)= 5505.208449;
W(5,4)= -356.615669;
W(5,5)= 2638.064090;
W(5,6)= -188.300181;
W(5,7)=  -52.880219;
W(6,0)=-37327.950312;
W(6,1)=2336108.240765;
W(6,2)=-7190406.270515;
W(6,3)=-5377401.233701;
W(6,4)=348336.590090;
W(6,5)= -188.300181;
W(6,6)=183928.662521;
W(6,7)=51652.568524;
W(7,0)=-10482.784385;
W(7,1)=656047.781413;
W(7,2)=-2019277.188837;
W(7,3)=-1510132.145239;
W(7,4)=97823.141550;
W(7,5)=  -52.880219;
W(7,6)=51652.568524;
W(7,7)=14505.557745; 

 
  

  // re-symmetrize for numerical diff.
  W.Symmetrize("U");
  
  // remove first row and column
  //W = W.SubBlock(1,W.SizeX()-1,1,W.SizeY()-1);
  
  //W.writeFits("W.fits");
}



Vect HutererPlanckPrior::Residual(const Cosmology *Cosmo) const
{
  //cout << "HutererPlanckPrior::Residual" << endl;
  
  const CmbCosmo *c= get_cmbcosmo(Cosmo);
  const OmegamOmegaxW0WaCosmo * cosmo = dynamic_cast<const OmegamOmegaxW0WaCosmo *>(c->ExtractCosmo());
  if(!cosmo) {
    cout << "FATAL DETFStageIIPrior can only deal with OmegamOmegaxW0WaCosmo or OmegamW0WaCosmo" << endl;
    cout << *Cosmo << endl;
    abort();
  }
 
  double h = c->H0()/100;
  double h2 = h*h;
  
  Vect r(8);
  int i=0;
  r(i++) = c->ns;
  r(i++) = cosmo->Omegam()*h2;
  r(i++) = c->OBH2();
  r(i++) = cosmo->Omegak()*h2;
  r(i++) = cosmo->Omegax()*h2;
  r(i++) = c->delta;
  r(i++) = cosmo->W0();
  r(i++) = cosmo->Wa();
  
  
  r -= ref;

  
  cout << "residuals " << endl;
  cout << "ns = " << c->ns  << endl;
  cout << "delta = " << c->delta  << endl;
  
  return r;
}

Mat HutererPlanckPrior::Weight(const Cosmology *Cosmo) const
{
  //cout << "HutererPlanckPrior::Weight" << endl;
  return W;
}


Mat BAOB14Prior::Weight(const Cosmology *Cosmo) const
{
  Mat W(3,3);
  W(0,0) = 4444.;
  W(1,1) = 215156.;
  W(2,2) = 721487.;
  return W;
}

Vect BAOB14Prior::Residual(const Cosmology *Cosmo) const
{
  const NuCosmo *cosmo = dynamic_cast<const NuCosmo*>(Cosmo);
  if(cosmo==0) { 
    cout << "not a neutrino cosmology" << endl;
    abort();
  }
  double rs_val = cosmo->rs(cosmo->zdrag());
  Vect res(3);
  res(0) = rs_val/Dv_for_BAO(0.106,cosmo) - 0.336;
  res(1) = rs_val/Dv_for_BAO(0.35,cosmo) - 0.1126;
  res(2) = rs_val/Dv_for_BAO(0.57,cosmo) - 0.07315;
  return res;
}

PlanckB14Prior::PlanckB14Prior() {
  
  Mat C(3,3);

  C(0,0) = 0.79039; C(0,1)=-4.0042; C(0,2)=0.80608;
  C(1,1) = 66.950; C(1,2)=-6.9243;
  C(2,2) = 3.9712;
  C *= 1.e-7;

  C(1,0) = C(0,1);
  C(2,0) = C(0,2);
  C(2,1) = C(1,2);
  
  C.CholeskyInvert("S");
  W=C;
}

Mat PlanckB14Prior::Weight(const Cosmology *Cosmo) const
{
  return W;
}

  
Vect PlanckB14Prior::Residual(const Cosmology *Cosmo) const
{
  const NuCosmo *cosmo = dynamic_cast<const NuCosmo*>(Cosmo);
  if(cosmo==0) {
    cout << "not a neutrino cosmology" << endl;
    abort();
  }
  double h = cosmo->H0()/100.;
  double wb = cosmo->OBH2();
  
  double omegam = Cosmo->Omegam();
  double wm = omegam*(h*h);
  
  assert(cosmo->NeutrinoMass1()==0.06); // eV
  assert(cosmo->NeutrinoMass2()==0.); // eV
  assert(cosmo->NeutrinoMass3()==0.); // eV
  assert(cosmo->Neff()==3.046);
  assert(cosmo->Tcmb()==2.7255);
  
  double omega_nu =  cosmo->Omega_neutrino();
  // cout << "Omega_neutrinos = " << omega_nu << endl;
  double wn = omega_nu*h*h;
  
  double zstar = cosmo->zstar();  
  
  double wc = wm-wb-wn; // need to subtract neutrinos as well
  
  Vect res(3);
  res(0) = wb - 0.022065; 
  res(1) = wc - 0.1199;
  
  double theta_MC = cosmo->rs_camb(zstar)/cosmo->Da(zstar)/(1+zstar);
  res(2) = 100*theta_MC - 1.041;
  
  return res;
}




/////////////////////////////////////////////////////////////////////

/*********** bricolage : a prior that enables to use BAODR7 on its own */
#ifdef STORAGE
class OmegabPrior : public Prior
{
  private :
    double omega_m_val;
  double omega_m_error;

  public :
    OmegaM_Prior(const double& om, const double& err){
    omega_m_val = om;
    omega_m_error = err;
  };


  std::string Name() const {return "Om";}
  Mat Weight(const Cosmology *Cosmo=0) const;
  Vect Residual(const Cosmology *Cosmo) const;
  
};
  
#endif


Prior* FindPrior(const std::string &Name)
{
  if (Name == "WMAP5") return new Wmap5CosmoPrior();
  else if (Name=="E05") return new Eisenstein2005_Prior();
  else if (Name=="WMAP7") return new Wmap7CosmoPrior();
  else if (Name=="WMAP9") return new Wmap9CosmoPrior();
  else if (Name=="BAODR7") return new BAODR7Prior();
  else if (Name=="BAODR9") return new BAODR9Prior();
  else if (Name=="BAODR9Sanchez2012") return new BAODR9Sanchez2012Prior();
  else if (Name=="SHOES") return new ShoesPrior();
  else if (Name=="WiggleZ") return new WiggleZPrior();
  else if (Name=="6dFGS") return new SixdFGSPrior();
  else if (Name=="H0Riess2011") return new H0Riess2011Prior();
  else if (Name=="LightPlanck") return new LightPlanckPrior();
  else if (Name=="FullPlanck") return new FullPlanckPrior();
  else if (Name=="DETFStageII") return new DETFStageIIPrior();
  else if (Name=="DETFStageIIR") return new DETFStageIIPriorOfRahul();
  else if (Name=="HPlanck") return new HutererPlanckPrior();
  else if (Name=="BAOB14") return new BAOB14Prior();
  else if (Name=="PlanckB14") return new PlanckB14Prior();
  
cout << " FindPrior : No prior named " << Name << endl;
  cout << " Available priors : " 
       <<"WMAP5" <<',' 
       << "E05"  <<','
       << "WMAP7" <<',' 
       << "WMAP9" <<',' 
       << "BAODR7" <<',' 
       << "BAODR9" <<',' 
       << "BAODR9Sanchez2012" <<',' 
       << "SHOES" <<','
       << "WiggleZ" <<','
       << "6dFGS" <<','
       << "H0Riess2011" <<','
       << "LightPlanck" <<','
       << "FullPlanck" <<','
       << "DETFStageII" <<','
       << "DETFStageIIR" <<','
       << "HPlanck"  <<','
       << "BAOB14"   <<','
       << "PlanckB14" 
       << endl;
  exit(1);
}

