#ifndef COSMOLOGY_PRIOR_H
#define COSMOLOGY_PRIOR_H


#include <string>

#include "countedref.h"
#include "matvect.h"

class Mat;
class Vect;
class Cosmology;

class Prior : public RefCount {

  public :

  virtual std::string Name() const = 0;
  virtual ~Prior(){};
  virtual Mat Weight(const Cosmology *Cosmo=0) const = 0;
  virtual Vect Residual(const Cosmology *Cosmo) const = 0;
  virtual double Chi2(const Cosmology *Cosmo) const;
  virtual bool NeedCmb() const {return false;}

};


Prior* FindPrior(const std::string &Name);



class Eisenstein2005_Prior : public Prior {
 public :

  std::string Name() const { return "Eisenstein2005";}
  
  Mat Weight(const Cosmology *Cosmo=0) const;
  Vect Residual(const Cosmology *Cosmo) const;
};


class OmegaM_Prior : public Prior {
  private :
    double omega_m_val;
  double omega_m_error;

  public :
    OmegaM_Prior(const double& om, const double& err){
    omega_m_val = om;
    omega_m_error = err;
  };


  std::string Name() const {return "Om";}
  Mat Weight(const Cosmology *Cosmo=0) const;
  Vect Residual(const Cosmology *Cosmo) const;
  
};

struct Wmap5CosmoPrior : public Prior
{
  std::string Name() const {return "WMAP5";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const;
  bool NeedCmb() const { return true;}
};


struct Wmap7CosmoPrior : public Prior
{
  std::string Name() const {return "WMAP7";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return true;}
};  

struct Wmap9CosmoPrior : public Prior
{
  std::string Name() const {return "WMAP9";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return true;}
};  


struct BAODR7Prior : public Prior
{
  std::string Name() const {return "BAODR7";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return true;}
};    

/* Anderson et al. 2012 CMASS sample at z~0.57, Dv(0.57)/rs=13.67+-0.22, independent from BAODR7Prior */
struct BAODR9Prior : public Prior
{
  std::string Name() const {return "BAODR9";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return true;}
};    

/* Sanchez et al. 2012 CMASS sample at z~0.57, A(zm)=0.444 +- 0.014, to used INSTEAD of BAODR9Prior when no CMB */
struct BAODR9Sanchez2012Prior : public Prior
{
  std::string Name() const {return "BAODR9Sanchez2012";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return true;}
};    

/* Shoes prior is the H0 prior from the H maser stuff 
   http://cdsads.u-strasbg.fr/abs/2009ApJ...699..539R
   H 0 = 74.2 ± 3.6 km s-1 Mpc-1,
*/
struct ShoesPrior : public Prior
{
  std::string Name() const {return "SHOES";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return true;}
};

/* Riess 2011
   http://cdsads.u-strasbg.fr/abs/2011ApJ...730..119R
   H0 = 73.8 +- 2.4 km s−1 Mpc−1
 */
struct H0Riess2011Prior : public Prior
{
  std::string Name() const {return "H0Riess2011";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return true;}
};


/* Blake 2011. table 2
   http://cdsads.u-strasbg.fr/abs/2011MNRAS.418.1707B
*/
struct WiggleZPrior : public Prior
{
  std::string Name() const {return "WiggleZ";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return false;} // using A(z)
};  

/* Blake 2011. table 3
   http://cdsads.u-strasbg.fr/abs/2011MNRAS.418.1707B
 */
struct SixdFGSPrior : public Prior
{
  std::string Name() const {return "6dFGS";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return false;} // using A(z)
};  

/* ONLY FOR SIMULATION TESTS !! ASSUMES OMEGAM=0.27 !! 
 based on astro-ph 0803.1616v1
*/
struct LightPlanckPrior : public Prior
{
  
  double z_ref;
  double R_ref;
  LightPlanckPrior(); // 
  std::string Name() const {return "LightPlanck";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return false;} // using A(z)
};  

/* ONLY FOR SIMULATION TESTS !! ASSUMES OMEGAM=0.27 !!
   based on astro-ph 0803.1616v1
*/ 
struct FullPlanckPrior : public Prior
{
  double R_ref;
  double la_ref;
  double obh2_ref;
  Mat W;
  FullPlanckPrior(); // 
  std::string Name() const {return "FullPlanck";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return true;} 
};  

/* ONLY FOR SIMULATION TESTS !! 
   as used by DES 
*/
struct DETFStageIIPrior : public Prior
{
  Vect ref;
  Mat W;
  DETFStageIIPrior(); // 
  std::string Name() const {return "DETFStageII";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return true;} 
};  

/* 
   another version of planck prior used by DES 
 */
struct DETFStageIIPriorOfRahul : public Prior
{
  Vect ref;
  Mat W;
  DETFStageIIPriorOfRahul(); // 
  std::string Name() const {return "DETFStageIIR";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return false;} 
};  

/* based on Huterer, Dragan
 */
struct HutererPlanckPrior : public Prior
{
  Vect ref;
  Mat W;
  HutererPlanckPrior(); // 
  std::string Name() const {return "HPlanckPrior";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return true;} 
};  

/* the combinaison of BAO data used in Betoule2014 */
struct BAOB14Prior : public Prior
{
  BAOB14Prior(){}; // 
  std::string Name() const {return "BAOB14";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return true;} 
};  

/* the Planck prior used in Betoule2014 */
struct PlanckB14Prior : public Prior
{
  Mat W;
  PlanckB14Prior(); // 
  std::string Name() const {return "PlanckB14";}
  Vect Residual(const Cosmology *Cosmo) const;
  Mat Weight(const Cosmology *Cosmo=0) const; 
  bool NeedCmb() const { return true;} 

  double Im(const double & mc2divkT, bool fermion=true) const;
  double rho_nu_div_rho_crit0(const double& z, const double& Neff, const double &mc2_in_eV) const;
  

};  



#endif 
