#include "instrument.h"
#include "filterwheel.h"
#include "datacards.h"
#include "pathutils.h"
#include "fileutils.h"

/* 1 Angst */
#define WAVELENGTH_STEP 5.

/* The whole thing should be redesigned to allow for a more abstract
   definition of what may vary in an Instrument transmission.
   As it is now, it only provides variations of the filters as a function
   of something. One may need at some point that e.g. the atmospheric 
   transmission depends on something.
   We'll think about it when it becomes necessary....
*/


//#define DEBUG

static Sampled1DFunction *read_trans(const DataCards& cards,
				     const string & TagName,
				     const string &PathToMainFile)
{
#ifdef DEBUG
  cout << "in read_trans :" << endl;
  cout << "  TagName = " << TagName << endl;
  cout << "  PathToMainFile = " << PathToMainFile << endl;
#endif
  
  if (!cards.HasKey(TagName))
    {
      cout << "WARNING : no card named " << TagName << " in " << PathToMainFile << " setting in to constant==1 " << endl; 
      return new Constant1DFunction(1.);
    }

  const char *value = cards.SParam(TagName).c_str();
  char *endptr = NULL;
  double val = strtod(value, &endptr);
  if (endptr !=  value) // this is a number
    {
      return new Constant1DFunction(val);
    }
  else // one or more file names
    {
      vector<string> files = cards.AllParams(TagName);
      if (files.size() == 1)
	{	  
	  string s = AssembleFileName(PathToMainFile,files[0]);
	  return new  General1DFunction(s);
	}
      vector<Sampled1DFunction*> terms;
      for (unsigned k=0; k< files.size(); ++k)
	{
	  string s = AssembleFileName(PathToMainFile,files[k]);
	  //	  cout << "INFO : reading " << s << endl;
	  terms.push_back(new  General1DFunction(s));
	}
      General1DFunction *res = new General1DFunction(Product(5,terms));
      for (unsigned k=0; k<terms.size(); ++k) delete terms[k];
      return res;
    }
}

Instrument::~Instrument() { 
  //cout << "Instrument::~Instrument()" << endl;
  delete filters;
  delete optics_transmission;
  delete mirror_reflectivity;
  delete mean_qe;
  delete atmospheric_transmission;
} 

Instrument::Instrument(const string & Name) 
{
  optics_transmission = NULL;
  mirror_reflectivity = NULL;
  mean_qe = NULL;
  atmospheric_transmission = NULL;
  filters = NULL;
  filename = LocateFileFromCard(Name)+"/instrument.cards";

#ifdef DEBUG
  cout << " Reading '" << filename << "' for instrument " << Name 
       << std::endl;
#endif
  
  DataCards cards(filename);
  string pathToMainFile = DirName(filename);
  // instrumentName = BaseName(pathToMainFile);
  instrumentName = Name;

  optics_transmission =      read_trans(cards, "OPTICS_TRANS", pathToMainFile);
  mirror_reflectivity =      read_trans(cards, "MIRROR_REFLECTIVITY", pathToMainFile);
  mean_qe =                       read_trans(cards, "QE", pathToMainFile);
  atmospheric_transmission = read_trans(cards, "ATMOSPHERIC_TRANS", pathToMainFile);
  
  if (cards.HasKey("QE_PER_CCD"))
    {
      vector<string> files = cards.AllParams("QE_PER_CCD");
      for (unsigned k=0; k< files.size(); ++k)
	{
	  string s = AssembleFileName(pathToMainFile,files[k]);
	  //	  cout << "INFO : reading " << s << endl;
	  qe_per_ccd.push_back(new  General1DFunction(s));
	}
    }

  if (cards.HasKey("MIRROR_AREA"))
    {
      mirror_area = cards.DParam("MIRROR_AREA");
#ifdef DEBUG    
      cout << " mirror area (cm^2) " << mirror_area << endl;
#endif
    }
  else mirror_area = 1;

  if (cards.HasKey("PIX_SIZE"))
    {
      pix_size = cards.DParam("PIX_SIZE");
#ifdef DEBUG      
      cout << " pixel size(\") " << pix_size << endl;
#endif
    }
  else pix_size = 0;
  if (cards.HasKey("RADIALLY_VARIABLE_FILTERS"))
    {
      filters = new RadiallyVariableFilterWheel(AssembleFileName(pathToMainFile,
								 cards.SParam("RADIALLY_VARIABLE_FILTERS")),Name); 
    }
  else if (cards.HasKey("FILTERS"))
    filters = new NoParamFilterWheel(AssembleFileName(pathToMainFile, 
						      cards.SParam("FILTERS")),
				     Name);

  else 
    cout << "WARNING Instrument::Instrument : NO filterwheel found in " << filename  << endl;
  
  if (cards.HasKey("EFF_FILTERS")) {
    cout << "WARNING Instrument::Instrument ignoring EFF_FILTERS in file " << filename << endl
	 << "WARNING Instrument::Instrument Please remove this deprecated line" << endl;
  }

  if (cards.HasKey("PSF_CORRECTION_FILTERS")) {
    PSFCorrections = FilterWheel(AssembleFileName(pathToMainFile, 
						  cards.SParam("PSF_CORRECTION_FILTERS")),Name);
    cout << "INFO Instrument::Instrument found PSF correction filters " << cards.SParam("PSF_CORRECTION_FILTERS") << endl;
  }
if (cards.HasKey("CHROMATIC_CORRECTIONS")) {
    ChromaticCorrections = FilterWheel(AssembleFileName(pathToMainFile, 
						  cards.SParam("CHROMATIC_CORRECTIONS")),Name);
    cout << "INFO Instrument::Instrument found chromatic corrections " << cards.SParam("CHROMATIC_CORRECTIONS") << endl;
 }


#ifdef DEBUG
  cout << " for instrument " << Name << endl;
  cout << " filterwheel content " << endl << *filters << endl;
#endif

}

Filter Instrument::EffectiveFilterByBand(const std::string &Band,
					 int ccd,
					 const IndexParams* P) const{

  const Sampled1DFunction* QE = mean_qe;
  if(ccd>=0) QE = qe_per_ccd[ccd];
  
  if (Band == "NO_FILTER") {

    General1DFunction transmission = Product(WAVELENGTH_STEP,4, 
				       mirror_reflectivity, optics_transmission, 
				       QE, atmospheric_transmission);
    transmission *= mirror_area;
    return Filter(transmission,Band /* == "No_Filter" */, instrumentName);
  }
  
  
  Filter filter = filters->GetFilterByBand(Band, P);
  
  const Filter* chromatic_correction = 0;
  for (list<Filter>::const_iterator it = ChromaticCorrections.begin(); it != ChromaticCorrections.end() ; ++it) {
    if (it->Band() == Band) {
      chromatic_correction = &(*it);
      break;
    }
  }
  
  General1DFunction transmission;
  
  if(chromatic_correction == 0) {
    transmission = Product(WAVELENGTH_STEP,5, 
			   mirror_reflectivity, optics_transmission, &filter, 
			   QE, atmospheric_transmission);
  } else {
    transmission = Product(WAVELENGTH_STEP,6, 
			   mirror_reflectivity, optics_transmission, &filter,  
			   QE, atmospheric_transmission, 
			   chromatic_correction);
  }
  
  transmission *= mirror_area;
  return Filter(transmission,Band /* == "No_Filter" */, instrumentName);
}

Filter Instrument::EffectiveFilterByBand(const string &Band, 
					 const IndexParams* P) const
{
  return EffectiveFilterByBand(Band,-1,P); // ccd=-1 means use average QE
}


FilterWheel Instrument::EffectiveFilters(const IndexParams*P) const
{
  std::vector<std::string> names = filters->BandNames();
  FilterWheel wheel;
  for (std::vector<std::string>::const_iterator i=names.begin(); i != names.end(); ++i)
    {
      wheel.push_back(EffectiveFilterByBand(*i, P));
    }
  return wheel;
}

FilterWheel Instrument::EffectiveFilters(int ccd, const IndexParams*P) const
{
  std::vector<std::string> names = filters->BandNames();
  FilterWheel wheel;
  for (std::vector<std::string>::const_iterator i=names.begin(); i != names.end(); ++i)
    {
      wheel.push_back(EffectiveFilterByBand(*i, ccd, P));
    }
  return wheel;
}

bool Instrument::HasPSFCorrectionForBand(const std::string &Band) const 
{ 
  for (list<Filter>::const_iterator it = PSFCorrections.begin(); it != PSFCorrections.end() ; ++it)
    if (it->Band() == Band) return true;
  return false;
}

const Filter& Instrument::PSFCorrectionForBand(const std::string &Band) const
{
  return PSFCorrections.GetFilterByBand(Band);
}




#ifdef STORAGE /* I think it is not used */
//! if you have an effective filter in hand, returns the corresponding raw one 
const Filter Instrument::GetRawFilter(const Filter &EffectiveFilter)  const
{
  int pos=1;
  bool found = false;
  for (FilterCIterator it = effective_filters.begin(); it != effective_filters.end(); ++it)
    {
      const Filter &eff_filter = *it;
      if (eff_filter.Name() == EffectiveFilter.Name()) {found = true; break;}
      pos++;
    }
  if (!found)
    {
      cerr << "Instrument::GetRawFilter called with an argument which is not an effective filter : " 
	   << EffectiveFilter.Name() << endl;
      cerr << "crashing now " << endl;
      abort();
    }
  return filters.GetFilter(pos);
}
#endif /* STORAGE */
