#ifndef INSTRUMENT__H
#define INSTRUMENT__H

#include "filter.h"
#include "filterwheel.h"
#include <string>
#include <map>
#include <vector>

class Instrument {
private:
  double mirror_area, pix_size;
  Sampled1DFunction *optics_transmission;
  Sampled1DFunction *mirror_reflectivity;
  Sampled1DFunction *mean_qe;
  std::vector<Sampled1DFunction *> qe_per_ccd;
  Sampled1DFunction *atmospheric_transmission;
  
  FilterCollection *filters; // this is virtual class that handles 
  std::string filename;
  std::string instrumentName;
  
  // for PSF photometry wavelength dependent correction
  FilterWheel PSFCorrections;
  
  // to use instead of PSF corrections :
  FilterWheel ChromaticCorrections;

public:

  //! the given file contains some numerical data and data filenames
  Instrument(const std::string &Name);
  //! name
  const std::string &Name() const {return instrumentName;}

  //!the full file name
  const std::string &Filename() const {return filename;}

  //! in cm^2
  double MirrorArea() const {return mirror_area;};
  //! in arcsec
  double PixelSize() const {return pix_size;};
  //! in arcsec^2
  double PixelArea() const {return pix_size*pix_size;};

  //! return the filter + optics + qe overall transmission
  Filter EffectiveFilterByBand(const std::string &Band, const IndexParams* P = 0) const;

  //! return the filter + optics + qe(ccd) overall transmission
  Filter EffectiveFilterByBand(const std::string &Band, int ccd, const IndexParams* P = 0) const;
  
  //! the filter collection (one per band)
  FilterWheel EffectiveFilters(const IndexParams*P = NULL) const;

  //! the filter collection (one per band)
  FilterWheel EffectiveFilters(int ccd, const IndexParams*P = NULL) const;
 
  
  const FilterCollection* Filters() const { return filters;}
  
  const Sampled1DFunction *  OpticsTransmission() const { return optics_transmission; }
  const Sampled1DFunction *  MirrorReflectivity() const { return mirror_reflectivity; }
  const Sampled1DFunction *  MeanQE() const { return mean_qe; }
  const Sampled1DFunction *  AtmosphericTransmission() const { return atmospheric_transmission; }

  ~Instrument();

  // for PSF photometry wavelength dependent correction
  bool HasPSFCorrectionForBand(const std::string &Band) const;
  const Filter& PSFCorrectionForBand(const std::string &Band) const; 
  
  

#ifdef STORAGE  
  //! the whole filter wheel (+overall transmission)
  const FilterWheel& EffectiveFilters() const {return effective_filters;};

  //! the total optical transmission without filters(mirror optics and qe)
  General1DFunction OverallTransmission() const;
 
  //! if you have an effective filter in hand, returns the corresponding raw one 
  const Filter& GetRawFilter(const Filter &EffectiveFilter) const;
#endif

  

};

//! only here to be derived (when there are several ways for an instrument to be variable
class IndexParams
{
 public :
  virtual ~IndexParams() {};
};


//! a class needed for pass over coordinates to the spatially variable filter stuff 
class FocalPlanePosition :  public IndexParams
{
 public :
  const double x;
  const double y;
  FocalPlanePosition(const double X, const double Y) : x(X), y(Y) {};
  ~FocalPlanePosition() {};
};

std::string InstrumentNameFromFilterName(const std::string &InstFilterName);
#endif /* INSTRUMENT__H */


