#include <vector>
#include <map>

#ifdef OPENMP
#include <omp.h>
#endif

#include <matvect.h>
#include <local_global_solver.h>

using namespace std;



 /*
   
   A =  W0    A01^t A02^t A03^t ...  
        A01   W1
	A02         W2
	A03               W3 


   B = B0	
       B1
       B2
       ...

   we want to solve :    
   A X = B    

   for i>0 (local set of parameters), equation is :
     A0i.X0 + Wi.Xi = Bi
     Xi = Wi^-1 (Bi-A0i.X0)

   for i=0 (global) : 
     W0.X0 + sum_i A0i^t ( Wi^-1 (Bi-A0i.X0) ) = B0
   
   need to solve :
   (W0 - sum_i A0i^t Wi^-1  A0i ) X0 = B0 - sum_i A0i^t Wi^-1 Bi
   with C0  = (W0 - sum_i A0i^t Wi^-1  A0i )^-1
   
   X0 = C0 B0 - sum_i C0 A0i^t Wi^-1 Bi
   Xi =  Wi^-1 Bi - Wi^-1 A0i ( C0 B0 - sum_j C0 A0j^t Wj^-1 Bj )
      =  ( -Wi^-1*A0i*C0 ) B0  + sum_j ( delta_ij Wi^-1  + Wi^-1 A0i C0 A0j^t Wj^-1 ) Bj 

   invert :
   
   A =  C0    C01 ...
        C01   C1
	C02         
	...

   C0  = (W0 - sum_i A0i^t Wi^-1  A0i )^-1
   C0i = -C0*A0i^t Wi^-1 (or transposed)
   Cij = delta_ij Wi^-1  + Wi^-1 A0i C0 A0j^t Wj^-1 
  
 */

typedef  map< int , vector<int> >::iterator mapiter;

int cholesky_solve_and_invert(Mat &A, Vect &B,  const char* UorL, bool invert = true) {
  int status = cholesky_solve(A,B,UorL);
  if(status!=0) return status;
  if(invert) status = cholesky_invert(A,UorL);
  return status;
}

int local_global_solve(Mat &A, Vect &B, const Vect& Index, const char* UorL, bool local_is_pos_def, bool global_is_pos_def, bool invert_global, bool invert_all) {
  
  vector<int> global_index;
  map< int , vector<int> > local_index;
  int max_local_index = 0;
  
  bool debug = true; 

  if(invert_all) invert_global = true;

  for(size_t i=0;i<Index.size();++i) {
    int id = Index(i);

    if(id<0) abort();

    if(id==0) 
      global_index.push_back(i);
    else {
      if(id>max_local_index)
	max_local_index=id;
      local_index[id].push_back(i);
    }
  }

  // only global
  if(local_index.size()==0) {
    cout << "WARNING local_global_solve : only global parameters" << endl;
    int status;
    if(global_is_pos_def) {
      status = cholesky_solve_and_invert(A,B,UorL,invert_global);
    } else {
      status = general_solve(A,B,invert_global,UorL);
    }
    return status;
  }
  
  // only local
  if(global_index.size()==0 && local_index.size()==1) {
    cout << "WARNING local_global_solve : only 1 set of local parameters" << endl;
    int status;
    if(local_is_pos_def) {
      status = cholesky_solve_and_invert(A,B,UorL,invert_global);
    } else {
      status = general_solve(A,B,invert_global,UorL);
    }
    return status;
  }
  
  
  size_t ng = global_index.size();

  
  
  if(0 && debug) {
    cout << A.SizeX() << " " << B.size() << " " << Index.size() << endl;
    cout << ng << " global parameters" << endl;
    for(mapiter it =local_index.begin() ; it != local_index.end(); it++) {
      cout << "local parameters " << it->first << " " << it->second.size() << endl;
    }
  }
  
  
  // invert local stuff
  vector<Mat> Wlinv(max_local_index+1);
  vector<Vect> Bl(max_local_index+1);

  const double* a = A.Data();
  const double* b = B.Data();
  int ntot = B.size();
  
  // can be parallelized
#pragma omp parallel for
  for(int iloc=0;iloc<=max_local_index;iloc++) {
    mapiter it = local_index.find(iloc);
    if(it == local_index.end()) continue;
    // for(mapiter it =local_index.begin() ; it != local_index.end(); it++) {
    
    const vector<int>& indices = it->second;
    
    size_t npar = indices.size();

    Mat Wl(npar,npar);
    Vect Bloc(npar);
    
    
    double *bl = Bloc.NonConstData();
    double *wl = Wl.NonConstData();
    
    for(size_t wj=0;wj<npar;wj++,bl++) {
      int aj = indices[wj];
      *bl = b[aj];
      for(size_t wi=0;wi<npar;wi++,wl++) {
	*wl = a[indices[wi]+aj*ntot]; // A(indices[wi],aj);
      }
    }

    if(ng==0) { // specific case where there is no global parameters
      int status;
      if(local_is_pos_def) {
	status = cholesky_solve_and_invert(Wl,Bloc,UorL,invert_global);
      } else {
	status = general_solve(Wl,Bloc,invert_global,UorL);
      }
      if(status!=0) abort();
      
      // save this in B
      for(size_t wi=0;wi<npar;wi++) {
	B(indices[wi]) = Bloc(wi);
      }
      continue;
    }

    if(debug) 
      if(iloc%100==0) cout << "inverting local " << it->first << " of size " << npar << endl;
    
    int status;
    if(local_is_pos_def) {
      status = posdef_invert(Wl,UorL);
    } else {
      Vect Bdummy(npar);
      status = general_solve(Wl,Bdummy,true,UorL);
    }
    if(status!=0) abort();
    Wl.Symmetrize(UorL);
    Wlinv[it->first]=Wl;
    Bl[it->first]=Bloc;
  } /* end of parallelized loop */

  if(ng==0) return 0;

  // solve global
  
  Mat Ag(ng,ng);
  Vect Bg(ng);

  double *bg = Bg.NonConstData();
  double *ag = Ag.NonConstData();

  for(size_t gj=0;gj<ng;gj++,bg++) {
    int aj = global_index[gj];
    *bg = b[aj];
    for(size_t gi=0;gi<ng;gi++,ag++) {
      *ag = a[global_index[gi]+aj*ntot]; // A(global_index[gi],aj);
    }
  }
  
  if(debug) cout << "subtract local terms to global "  << endl;

  // "L" <=> A(i,j)!=0 for i>=j 
  bool isL = ( UorL[0] == 'L' );
  vector<Mat> Agl(max_local_index+1);
  vector<Mat> AgltWi(max_local_index+1);
  
  #pragma omp parallel for
  for(int iloc=0;iloc<=max_local_index;iloc++) {
    mapiter it = local_index.find(iloc);
    if(it == local_index.end()) continue;
    
    //  for(mapiter it = local_index.begin() ; it != local_index.end() ; it++) {
    
    if(debug) 
      if(iloc%100==0)
	cout << "computing local term " << it->first << " (1st part, in parallel)" << endl;

    const vector<int>& local_indices = it->second;
    
    size_t nl = local_indices.size();
    Mat Agloc(ng,nl);
    
    double *agl = Agloc.NonConstData();
    
    for(size_t lj=0;lj<nl;lj++) {
      int aj = local_indices[lj];
      for(size_t gi=0;gi<ng;gi++,agl++) {
	int ai = global_index[gi];
	
	if( (isL && ai>=aj) || ( (!isL) && ai<=aj ) ) 
	  *agl = a[ai+aj*ntot]; // A(ai,aj);
	else
	  *agl = a[aj+ai*ntot]; // Agloc(gi,lj) = A(aj,ai);
      }
    }

    Agl[it->first]=Agloc;
    AgltWi[it->first]=Agloc.transposed()*Wlinv[it->first];
    

  } /* end of parallelized loop */
  
  for(int iloc=0;iloc<=max_local_index;iloc++) {
    mapiter it = local_index.find(iloc);
    if(it == local_index.end()) continue;
    
    if(debug) 
      if(iloc%100==0)
	cout << "computing local term " << it->first << " (2nd part, not in parallel)" << endl;
    
    // Ag -= AgltWi[it->first]*Agl[it->first];
    // Bg -= AgltWi[it->first]*Bl[it->first];
    // faster using DGEMM : computes  C = alpha*B*A + beta*C  
    // void DGEMM(Mat &C, const Mat& A, const Mat& B, const double& alpha, const double& beta);
    
    /*
      DGEMM(Ag,Agl[it->first],AgltWi[it->first],-1,1);
      DGEMM(Bg,Bl[it->first],AgltWi[it->first],-1,1);
    */

   
    
    // actually two nodes are used

    // uncomment this to parallize this part of the code
#define PARALLELIZED_LOCAL_GLOBAL

#ifdef PARALLELIZED_LOCAL_GLOBAL
    #pragma omp parallel 
    {
#pragma omp sections
      {
#pragma omp section	
{
#endif
	  
	  // Ag -= AgltWi[it->first]*Agl[it->first];
	  DGEMM(Ag,Agl[it->first],AgltWi[it->first],-1,1);
	  


#ifdef PARALLELIZED_LOCAL_GLOBAL
	  //cout << "Comp A term, THREAD=" <<  omp_get_thread_num() << " N_THREADS=" << omp_get_num_threads()  << endl; 
	}
#pragma omp section
	{
#endif

	  // Bg -= AgltWi[it->first]*Bl[it->first];
	  DGEMM(Bg,Bl[it->first],AgltWi[it->first],-1,1);

#ifdef PARALLELIZED_LOCAL_GLOBAL
	  //cout << "Comp B term, THREAD=" <<  omp_get_thread_num() << " N_THREADS=" << omp_get_num_threads()  << endl; 
	}
      } 
    } // end of parallelized sections
#endif

    

    
  }
  
  if(debug) cout << "solve (and invert if needed) global "  << endl;
  
  int status;
  if(global_is_pos_def) {
    status = cholesky_solve_and_invert(Ag,Bg,UorL,invert_global);
  } else {
    status = general_solve(Ag,Bg,invert_global,UorL);
  }
  if(status!=0) return status;
  

  double *nc_b        = B.NonConstData();
  {
    const double *bg = Bg.Data();
    
    for(size_t gi=0;gi<ng;gi++,bg++) {
      nc_b[global_index[gi]] = *bg;
    }
  }
  if(debug) cout << "solve (and invert if needed) local "  << endl;
  
  // solve local
#pragma omp parallel for
  for(int iloc=0;iloc<=max_local_index;iloc++) {
    mapiter it = local_index.find(iloc);
    if(it == local_index.end()) continue;
    //for(mapiter it =local_index.begin() ; it != local_index.end(); it++) {

    Vect Xl = Wlinv[it->first]*(Bl[it->first]-Agl[it->first]*Bg);
    const vector<int>& local_indices = it->second;
    size_t nl = local_indices.size();
    const double *x = Xl.Data();
    for(size_t li=0;li<nl;li++,x++) {
      nc_b[local_indices[li]]= *x;
    }
  } /* end of parallelized loop */
  
  if(! invert_global)  return 0;
  
  if(debug) cout << "reconstruct inverse matrix "  << endl;
  
  // need to reconstruct inverse matrix
  A.Zero();
  
  // global global : (W0 - sum_i A0i^t Wi^-1  A0i )^-1
  if(debug) cout << " global global term "  << endl;
  // first symmetrize to be safe
  Ag.Symmetrize(UorL);
  
  const Mat& Cg = Ag; // it is the covariance of the global parameters now

  for(size_t ig=0;ig<ng;ig++)
    for(size_t jg=0;jg<ng;jg++)
      A(global_index[ig],global_index[jg]) = Cg(ig,jg);
  
  if(!invert_all) return 0;
  
  
  vector<Mat> ClAgl(max_local_index+1);
  vector<Mat> ClAglCg(max_local_index+1);
  // local 
  for(mapiter it =local_index.begin() ; it != local_index.end(); it++) {
    const vector<int>& local_indices_i = it->second;
    size_t nl_i = local_indices_i.size();
    
    if(debug) 
      if(it->first % 100 == 0)
	cout << "local local diagonal block " << it->first << endl;
    
    // local local diagonal block : delta_ij Wi^-1
    const Mat& Cl = Wlinv[it->first];
    for(size_t li=0;li<nl_i;li++) {
      for(size_t lj=0;lj<nl_i;lj++) {
	A(local_indices_i[li],local_indices_i[lj]) = Cl(li,lj);
      }
    }

    if(debug) 
      if(it->first % 100 == 0)
	cout << "local global block " << it->first << endl;

    // local global : C0i = -C0*A0i^t Wi^-1 
    // Ag is actually inverse of Ag' here
    const Mat& Agloc_i = Agl[it->first];
    ClAgl[it->first]=Cl*Agloc_i;
    const Mat& ClAgl_i = ClAgl[it->first];
    ClAglCg[it->first]=ClAgl_i*Cg;
    const Mat& ClAglCg_i = ClAglCg[it->first];
    
    for(size_t ig=0;ig<ng;ig++) {
      int i = global_index[ig];
      for(size_t lj=0;lj<nl_i;lj++) {
	int j = local_indices_i[lj];
	A(j,i) = A(i,j) = -ClAglCg_i(ig,lj);
      }
    }
  }
  
  
  
  // local local covariant term  Wi^-1 A0i C0 A0j^t Wj^-1 
  for(mapiter it =local_index.begin() ; it != local_index.end(); it++) {
    const vector<int>& local_indices_i = it->second;
    size_t nl_i = local_indices_i.size();
    for(mapiter jt = it ; jt != local_index.end(); jt++) {     
      
      if(debug) 
	if(it->first % 100 == 0)
	  cout << "local local off diagonal block " << it->first << " " << jt->first << endl;
      
      const vector<int>& local_indices_j = jt->second;
      size_t nl_j = local_indices_j.size();

      Mat C = ClAglCg[it->first]*ClAgl[jt->first].transposed();
      
      for(size_t li=0;li<nl_i;li++) {
	int i = local_indices_i[li];
	for(size_t lj=0;lj<nl_j;lj++) {
	  int j = local_indices_j[lj];
	  A(i,j) += C(lj,li); 	  
	  if(i!=j)
	    A(j,i) += C(lj,li); 
	  
	}
      }
    }
  }
  
  

  return 0;
}
