#include "measureddata.h"
#include "dictfile.h"
#include "fatalerror.h"
#include <cmath>
#include <fstream>
#include <fstream>
#ifdef USECFITSIO
#include <fitsio.h>
#endif
#include <iomanip>
#include <string>
#include <map>

#include "fileutils.h" // DecomposeString
#include "instrument.h"
#include "sampledfunction.h"
#include "pathutils.h"
#include "hostabsorption.h"
#include "datafile.h"
#include "magsys.h"
#include "globalval.h"

#ifndef  WAVELENGTH_STEP
/* 10 Angst */
#define WAVELENGTH_STEP 10 
#endif

// c'est a cause du fait que minuit admet un nombre limite de characteres pour
// les noms des parametres et que j'utilise le nom du spectre pour
// identifier ces parametres de recalibration
#define MAX_NUMBER_OF_CHARACTERS_FOR_NAME_OF_SPECTRUM 6

//#define DEBUG

static double sqr(double x) {return x*x;}

double MeasuredPointList::GetSimulationZP() const{
  
  return GetMagSys(magSystem)->ZeroPoint(EffectiveFilter);
}

MeasuredPointList::MeasuredPointList() {
  has_focal_plane_coord = false;
  zero_point_uncertainty = 0;
}


void MeasuredPoint::StoreLine(DictFileCIterator &Line)
{
  const Dictionnary & dict = Line->GetDictionnary();
  for (Dictionnary::const_iterator i= dict.begin(); i!= dict.end(); ++i)
    {
      string tag = i->first;
      string val = Line->Value(tag);
      dict_file_line[tag] = val;
    }
}

MeasuredPoint::Val MeasuredPoint::Value(const std::string &Key, const bool DiesIfAbsent) const
{
  MyMap::const_iterator i =  dict_file_line.find(Key);
  if (i == dict_file_line.end())
    if (DiesIfAbsent) FatalError("did not find Key "+Key, true);
  return i->second;
}


bool  MeasuredPoint::HasKey(const std::string &Key) const
{
  MyMap::const_iterator i =  dict_file_line.find(Key);
  return ( i != dict_file_line.end());
}


//#ifdef STORAGE
// errorfloor = errorfloordval times the peak flux, see jha thesis page 177
void MeasuredPointList::ApplyErrorFloor(double errorfloordval) {

  // cerr << "WARNING MeasuredPointList::ApplyErrorFloor is disabled" << endl; return;
  cout << "applying error floor : " << errorfloordval << endl;
#ifdef DEBUG
  cout << "apply error floor " << endl;
#endif
  
  if(errorfloordval==0)
    errorfloordval = 0.01;
  bool has_covmat = CovMat.SizeX()>0;
  double fluxmax = -1000;
  for(MeasuredPointIterator i= begin(); i!= end(); ++i) {
    MeasuredPoint *p = *i;
    if(fluxmax < p->flux) {
      fluxmax = p->flux;
    } 
  }
  
  double errorfloor = fluxmax * errorfloordval ;
  int count = 0;
  for(MeasuredPointIterator i= begin(); i!= end(); ++i, ++count) {
    MeasuredPoint *p = *i;
    p->fluxerr = sqrt(sqr(errorfloor)+sqr(p->fluxerr));
    if(has_covmat)
      CovMat(count,count) += sqr(errorfloor);
  }
  
}
//#endif


void MeasuredPointList::ApplyScaling() {
#ifdef DEBUG
  cout << "apply scaling " << endl;
#endif
  // flux scaling
  // ================================================
  // we want to scale those fluxes in order to match with the simulation 
  // zero point.
  // flux -> flux* 10^{ [-ZP_{data} + ZP_{simul}(instrument,EffectiveFilter,magSystem)]/2.5 }
  
  // if WeightMat exists, we have to scale it a the same time

  // compute simulation zp
  double simulation_zeropoint = GetSimulationZP();
  
  double scaling=1;
  Vect scale(size()); // a vector of scalings
  int count = 0;
  for(MeasuredPointIterator i= begin(); i!= end(); ++i) {
    MeasuredPoint *p = *i;
    scaling =  pow(10.,0.4*(-p->zp+simulation_zeropoint));
    
#ifdef DEBUG
    cout << "ZP   : " << p->zp << " --> " << simulation_zeropoint << endl;
    cout << "Flux : " << p->flux << " --> " << p->flux * scaling << endl;
#endif
    
    p->flux *= scaling;
    p->fluxerr *= scaling;
    p->zp = simulation_zeropoint;
    scale(count) = scaling;
    count++;
  }
  
  if(CovMat.SizeX()>0) {
#ifdef DEBUG
    cout << "Apply scaling to covariance matrix" << endl;
    cout << "Before : " << CovMat << endl;
#endif
    for(size_t i = 0 ; i<CovMat.SizeX() ; i++)
      for(size_t j = 0 ; j<CovMat.SizeX() ; j++)
	CovMat(i,j)*=(scale(i)*scale(j));
  }
#ifdef DEBUG
    cout << "After : " << CovMat << endl;
#endif
}

void MeasuredPointList::CreateFilter(const string &instrument_name, const string &bandName, const GlobalVal &globalvals, const double& MWEBV) {
  instrumentName = instrument_name;
  Instrument instrument(instrumentName);
  
  
  
  if(globalvals.HasKey("X_FOCAL_PLANE")) {
    has_focal_plane_coord = true;
    focal_plane_x_coord  = globalvals.getDoubleValue("X_FOCAL_PLANE");
    focal_plane_y_coord  = globalvals.getDoubleValue("Y_FOCAL_PLANE");
    
    cout << "INFO MeasuredPointList::CreateFilter focal plane coords " <<  focal_plane_x_coord << " " << focal_plane_y_coord << endl;
    FocalPlanePosition P(focal_plane_x_coord,focal_plane_y_coord);
    EffectiveFilter = instrument.EffectiveFilterByBand(bandName,&P);
  }else{
    EffectiveFilter = instrument.EffectiveFilterByBand(bandName);
  }
  
  EffectiveFilterWithMilkyWay = EffectiveFilter;
  vector<Sampled1DFunction*> terms;
  terms.push_back(&EffectiveFilter);

  double data[2];

  if(fabs(MWEBV)>1.e-10) {  
    data[0] = MWEBV;
    data[1] = 3.1; 
    // mean Rv for the Galaxy (see E.L. Fitzpatrick astro-ph/9809387 , J.A. Cardelli et al., ApJ 425, 245 (1989) )
    cout << "INFO MeasuredPointList::CreateFilter Take into account Galactiv E(B-V)=" << MWEBV << endl;
    terms.push_back(new Computed1DFunction(&FluxExtinctionCardelli,data));
  }
  

  // add here PSF wavelength dependent correction if exists
  if(instrument.HasPSFCorrectionForBand(bandName)) {
    cout << "INFO MeasuredPointList::CreateFilter PSF correction for " << bandName << endl;
    terms.push_back(new Filter(instrument.PSFCorrectionForBand(bandName)));
    
  }
  
  if( terms.size()>1) { // already EffectiveFilter in there  
    
    (General1DFunction &) EffectiveFilterWithMilkyWay = Product(WAVELENGTH_STEP,terms);
    for(size_t i=1;i<terms.size();++i) // start at 1 cause dont want to delete EffectiveFilter
      delete terms[i];   
  }
} 

MeasuredPointList::MeasuredPointList(const string &FileName,bool applyscaling, double MWEBV)
{
  DictFile file(FileName);
  instrumentName = file.GlobalValue("INSTRUMENT");
  string bandName = file.GlobalValue("BAND");
  magSystem = file.GlobalValue("MAGSYS");
  
  CreateFilter(instrumentName,bandName,file.GlobalKeys,MWEBV);

  string key = "ZPERR_";
  if (file.GlobalKeys.HasKey(key)){
    zero_point_uncertainty = file.GlobalKeys.getDoubleValue(key);
    cout << "INFO MeasuredData::GetPointList using " << key << " = " <<  zero_point_uncertainty << endl;
  }
  else
    zero_point_uncertainty = 0;

  if(file.HasMat()) CovMat = file.GetMat();
  else if(file.HasGlobalKey("WEIGHTMAT")){
    WeightMatFileName = file.GlobalValue("WEIGHTMAT");
    string filename =  AddSlash(DirName(FullFileName(FileName)))+WeightMatFileName;
    if(FileExists(filename)) {
      ifstream stream(filename.c_str());
      Mat WeightMat;
      WeightMat.readASCII(stream);
      stream.close();
      WeightMat.Symmetrize("U"); // cause may be small differences when reading the file
      CovMat = WeightMat;
      CovMat.SymMatInvert();
      for(size_t i=0;i<WeightMat.SizeX();i++) {
	if(WeightMat(i,i)<1.e-20) {
	  cerr << "in MeasuredPointList::MeasuredPointList, Null diagonal value in weight map" << endl;
	  abort();
	}
      }
    }else{
      FatalError("cannot open matrix ASCII file "+filename,true);
    }
  }else{
    WeightMatFileName="";
  }
  
  // read extra information
  infos = GlobalVal(FileName);
  
    
  int index = 0;
  for (DictFileCIterator line = file.begin(); line != file.end(); ++line, ++index)
    {
      MeasuredPoint *m = new MeasuredPoint(EffectiveFilterWithMilkyWay,index);
      m->StoreLine(line);
      if (line->HasKey("Flux"))
	{
	  m->flux =  line->Value("Flux");
	  m->fluxerr =  line->Value("Fluxerr");
	  if (m->fluxerr <= 0)
	    FatalError(" negative or null error in "+FileName);
	  m->zp = line->Value("ZP");
	}
      else if (line->HasKey("Mag"))
	{
	  if (line->HasKey("Flux"))
	    FatalError("Cannot mix Flux And Mag in file"+FileName);
	  // have to compute the reference flux for mags
	  double mag = line->Value("Mag");
	  double errMag = line->Value("Magerr");
	  if(errMag==0)
	    errMag = 0.001;
	  m->zp = 0;
	  m->flux = pow(10.,0.4*(m->zp-mag));
	  // 0.921 = 0.4*log(10), with log(e) = 1.
	  m->fluxerr = m->flux*0.921*errMag;
	}
      m->date = line->Value("Date");
      
      push_back(m);
    }
  
  if(applyscaling) 
    ApplyScaling();
  

}


MeasuredPointList::iterator MeasuredPointList::erase(MeasuredPointList::iterator toremove) {
  if(CovMat.SizeX()==0 || size()==1) {
    return  ((list<CountedRef<MeasuredPoint> >*) this)->erase(toremove);
  }
  
  size_t index=0;
  MeasuredPointIterator  iter = begin();
  while(iter!=toremove) {
    if(iter==end()) {
      cout << "what a strange error, this iterator does not match any in the list" << endl;
      return toremove;
    }
    index++;
    ++iter;
  }
  //cout << "remove line and column at index " << index << endl;
  CovMat = CovMat.WithoutRows(index,index);
  CovMat = CovMat.WithoutColumns(index,index);
  
  return ((list<CountedRef<MeasuredPoint> >*) this)->erase(toremove);
}


#include <cstdio>

// old deprecated format
void MeasuredData::ReadLightFile(const string& lightFileName) {
  FILE *file = fopen(lightFileName.c_str(), "r");
  mwEbv = 0;
  name=BaseName(snpath);
  if(name=="" || name==".")
    name = "";
  
  float z_default = -1;
  if(file) {
    char line[1024];
    
    while (fgets(line,1024,file))
      {
	char *start = line;
	start = start+strspn(start," \t");
	if (*start == '#') continue;
	char *cr = strrchr(start,'\n'); if (cr) *cr = '\0';
	vector<string> words;
	DecomposeString(words,start," ");
	//      if (words.size() != 2)
	//	FatalError(" expect 'Tag value' pairs in "+lightFileName);
	if (words.size() < 2)  continue;
	if (words[0] == "NAME") name = words[1];
	if (words[0] == "SN") name = words[1];
	if (StringToUpper(words[0]) == "REDSHIFT") z_default = atof(words[1].c_str());
	if (words[0] == "z_heliocentric") z_heliocentric  = atof(words[1].c_str());
	if (words[0] == "z_cmb") z_cmb  = atof(words[1].c_str());
	if (words[0] == "MWEBV") mwEbv = atof(words[1].c_str());
	if (words[0] == "SET") survey = words[1];
	if (words[0] == "SURVEY") survey = words[1];
	
	//if (words[0] == "CCD") ccd = words[1];
	//if (words[0] == "Field") field = words[1];
	// any extra information is put in dictionnary
	infos.AddKey(words[0],words[1]); 
      }
    fclose(file);
    if(z_heliocentric<=-1 && z_default>=-1)
      z_heliocentric = z_default;
    
    
    
  }else{
    cout << "WARNING in MeasuredData:MeasuredData, could not read " << lightFileName << endl;
    cout << "        DO NOT FIT THIS SN WITHOUT FIXING THIS" << endl; 
  }
#ifdef DEBUG
  cout << " name = " << name << endl;
  cout << " z_heliocentric = " << z_heliocentric << endl;
  cout << " z_cmb = " << z_cmb << endl;
  cout << " mwEbv = " << mwEbv << endl;
#endif
  
  infos.AddKey("FILENAME",lightFileName);

}

void MeasuredData::Init() {
  name=""; 
  z_cmb=-1;
  z_heliocentric=-1;
  mwEbv=0;
  survey  = "unknown"; 
  day_max = -100000;
}

MeasuredData::MeasuredData() {
  Init();
};

MeasuredData::MeasuredData(const vector<string> &LcNames,bool applyscaling,  const vector<string> * discarded_bands)
{
  
  Init();
      
  if (LcNames.size() == 0) return;
  
  // look at first lc to see if in old or new format
  bool new_format = false;
  {
    DictFile file(LcNames[0]);
    new_format = file.HasGlobalKey("SN");
  }
  
  if(new_format) {
    cout << "INFO MeasuredData::MeasuredData reading lightcurves in new format (ignore lightfile)" << endl;
    for (size_t i=0 ; i < LcNames.size(); ++i) {
      ReadLightCurvesFile(LcNames[i]);
    }
  }else{
    cout << "INFO MeasuredData::MeasuredData reading lightcurves in old format (with lightfile)" << endl;

    ReadLightFile(DirName(LcNames[0])+"/lightfile"); 
    
    for (size_t i=0 ; i < LcNames.size(); ++i) {
      pointLists.push_back(new MeasuredPointList(LcNames[i],applyscaling,mwEbv));
    }
  }
  
  if (discarded_bands){
    for (int i = 0; i < discarded_bands->size(); i++){
      for (vector<CountedRef<MeasuredPointList> >::iterator j=  pointLists.begin(); j < pointLists.end(); j++){
	if (((*j)->GetEffectiveFilter().Name())== (*discarded_bands)[i]){
	  cout << "INFO MeasuredData::MeasuredData drop measurements in "<<discarded_bands[i] << endl;
	  pointLists.erase(j);
	}
      }
    }
  }
}


ostream &operator << (ostream &s, const MeasuredPointList &L)
{
  // ecrire ici les trucs:
  s <<"#Date :" <<endl;
  s <<"#Flux :" <<endl;
  s <<"#Fluxerr :"<<endl;
  s <<"#ZP :"<<endl;
  s <<"#end :" <<endl;
  s <<"@INSTRUMENT "<< L.GetInstrumentName() <<endl;
  s <<"@BAND "<< L.GetEffectiveFilter().Band() <<endl;
  s <<"@MAGSYS "<< L.GetMagSystem() <<endl;
  if(L.GetWeightMatFileName()!="")
     s <<"@WEIGHTMAT "<< L.GetWeightMatFileName() <<endl;
  for (MeasuredPointCIterator i = L.begin(); i != L.end(); ++i)
    s << *(*i);
  return s;
}

ostream &operator << (ostream &s, const MeasuredPoint &M)
{
  //s << " day = "  M.Day() << " flux = " << M.Flux() << " +/- " << M.FluxErr();
  s << M.Day() << " " << M.Flux() << " " << M.FluxErr() << " " << M.Zp()<< endl;
  return s;
}
  

void MeasuredData::Print(const int Level) const
{
  std::cout << "Name  : " <<  Name() << std::endl
	    << "z_heliocentric     : " << Zhelio() << std::endl 
	    << "z_cmb     : " << Zcmb() << std::endl
	    << "MWEBV: " << MWEBV() << std::endl
	    << "used bands : " << BandNames() << endl;

  if (Level >=2)
    {
      for (size_t i=0; i <NBands() ; ++i)
	{
	  const MeasuredPointList &pointList = PointList(i);
	  std::cout << " data in band " << pointList.BandName() << std::endl;
	  cout << pointList;
	}
    }
}


std::vector<std::string> MeasuredData::BandNames() const
{
  std::vector<std::string> result;
  for (size_t i =0; i < pointLists.size(); ++i)
    result.push_back(BandName(i));
  return result;
}


const MeasuredPointList * MeasuredData::GetMeasuredPointListByBand(const string& band) const
{
  for (size_t i =0; i < pointLists.size(); ++i) {
    if(pointLists[i]->GetEffectiveFilter().Band()==band) {
      return pointLists[i];
    }
  }
  return 0;
}

ostream &operator << (ostream &s, const std::vector<std::string> V)
{
  for (size_t i = 0; i < V.size(); ++i)
    s << V[i] << ' ';
  return s;
}



#ifdef USECFITSIO
static void fitserror(int status) {
  if(status) {
    cerr << " we got these successive errors:" << endl;
    fits_report_error(stderr, status);
    exit(status);
  }
}

static void read_a_fits_array(const string &FitsName,Vect &X, Vect &Y) {
  

  int status = 0;
  fitsfile *fptr = 0;
  fits_open_file(&fptr,FitsName.c_str(),0,&status); 
  fitserror(status); 
  
  char value[256];

  // first get the size of the spectrum
  fits_read_key(fptr, TSTRING, "NAXIS1", &value, NULL, &status); fitserror(status); 
  int ndata = atoi(value);
  X.allocate(ndata);
  Y.allocate(ndata);
  

  // read fluxes
  {
    float nullval = 0;
    int anynull;
    fits_read_img(fptr, TDOUBLE, 1, ndata, &nullval,  Y.NonConstData() , &anynull, &status); fitserror(status); 
  }
  
  // read scale
  fits_read_key(fptr, TSTRING, "CRPIX1", &value, NULL, &status); fitserror(status); 
  int ref_pix = atoi(value);
  fits_read_key(fptr, TSTRING, "CRVAL1", &value, NULL, &status); fitserror(status); 
  double ref_wavelength = atof(value);
  fits_read_key(fptr, TSTRING, "CDELT1", &value, NULL, &status); fitserror(status); 
  double wavelength_step = atof(value);
  
  for(int i=0;i<ndata;i++) {
    X(i) = ref_wavelength+(i+1-ref_pix)*wavelength_step;
  }
  fits_close_file(fptr, &status); fitserror(status);
}
#endif //  USECFITSIO

static void read_an_ascii_array(const string &FileName,Vect &X, Vect &Y,int Ycol) {
  DataFile dfile(FileName);
  int ndat = dfile.NDat();
  //cout << "ndata in fluxes = " << ndat << endl;
  X.allocate(ndat);
  Y.allocate(ndat);
  dfile.ReadCols(X.NonConstData(),Y.NonConstData(),ndat,Ycol);
}

static string fullpath(const string& SpectrumDescriptor, const string& filename) {
  return AddSlash(DirName(FullFileName(SpectrumDescriptor)))+filename;
}

void MeasuredSpectrum::ReadIUE(const DictFile& file) {
  
  size_t size = file.size();
  
  Vect Wavelength(size);
  Vect Flux(size);
  Vect Error(size);
  Vect Valid;
    
  int i=0;
  for (DictFileCIterator line = file.begin(); line != file.end(); 
       ++line, ++i) {
    Wavelength(i) = line->Value("WAVELENGTH");
    Flux(i) = line->Value("FLUX");
    Error(i) = line->Value("SIGMA");
  }
  
  LoadData(Wavelength,Flux,Error,Valid);  
 
}

void MeasuredSpectrum::Read(const DictFile& file) {
  
  size_t size = file.size();
  
  Vect Wavelength(size);
  Vect Flux(size);
  Vect Error;
  Vect Valid;
  
  if(file.begin()->HasKey("ERR"))
    Error.allocate(size);
  if(file.begin()->HasKey("VALID"))
    Valid.allocate(size);

  int i=0;
  for (DictFileCIterator line = file.begin(); line != file.end(); 
       ++line, ++i) {
    Wavelength(i) = line->Value("WAVE");
    Flux(i) = line->Value("FLUX");
    if(Error.Size()>0)
      Error(i) = line->Value("ERR");
    if(Valid.Size()>0)
      Valid(i) = line->Value("VALID");
  }
  LoadData(Wavelength,Flux,Error,Valid);
  
}


// in order to avoid duplicates
static void write_additional_info(const string& filename, ofstream& os, const GlobalVal& infos) {
  os.flush();
  GlobalVal written_infos(filename);
  GlobalVal infos_to_write;
  vector<string> keys = infos.Keys();
  for(size_t k=0;k<keys.size();k++) {
    const string& key = keys[k];
    if(key == "FILENAME") continue;
    if(! written_infos.HasKey(key))
      infos_to_write.AddKey(key,infos.getStringValues(key));
  }
  os << infos_to_write;
}

void MeasuredSpectrum::Write(const string& filename, const map<string,string>* globalkeys) const {
  
  ofstream os(filename.c_str());
  os << setprecision(8);
  os << "@FORMAT STD" << endl;
  if(sn_name!="") {
    os << "@SN " << sn_name << endl;
  }
  os << "@Date " << date << endl;
  os << "@Redshift " << redshift << endl;
  
  if(globalkeys) {
    for(map<string,string>::const_iterator it=globalkeys->begin(); it!=globalkeys->end(); ++it)
      os << "@" << it->first << " " << it->second << endl;
  }

  write_additional_info(filename,os,infos);
  
  os << "#WAVE : wavelength, observer frame, in A" << endl;
  os << "#FLUX : erg/s/cm2/A" << endl;
  os << "#ERR  : statistical uncertainty on flux" << endl;
  os << "#VALID : 1=ok 0=ignore" << endl;
  os << "#end" << endl;
  
  for(MeasuredSpectrum::const_iterator it=begin();it!=end();it++) {
    
    os << (*it)->Wavelength() 
       << " " << (*it)->Flux() 
       << " " << (*it)->FluxErr()
       << " " << (*it)->Valid()
       << endl;
  }
  os.close();
}

void MeasuredSpectrum::ReadVLT1(const DictFile& file) {
  
  size_t size = file.size();
  
  
  Vect Wavelength(size);
  Vect Flux(size);
  Vect Error(size);
  Vect GalaxyFlux(size);
  Vect GalaxyError(size);
  Vect Valid(size);
  

  int i=0;
  for (DictFileCIterator line = file.begin(); line != file.end(); ++line,++i) {
    // ! S.Baumont modif : copied by C.Balland
    //
    Valid(i)=0;
    if( line->Value("MASK") != 1 && line->Value("WAVE") > 4250. && line->Value("WAVE") < 10000. )
      Valid(i)=1;
    else
      Valid(i)=0;
    
    Wavelength(i) = line->Value("WAVE");
    Flux(i) = line->Value("SN_SPEC");
    Error(i) = line->Value("SN_ERR");
    GalaxyFlux(i) = line->Value("GAL_SPEC");
    GalaxyError(i) = line->Value("GAL_ERR");  
  }
  
  
  
  LoadData(Wavelength,Flux,Error,Valid);
  
  galaxy = new MeasuredSpectrum();
  galaxy->redshift = redshift;
  galaxy->is_a_galaxy = true;
  galaxy->LoadData(Wavelength,GalaxyFlux,GalaxyError,Valid);
}

void MeasuredSpectrum::ReadVLT0(const DictFile& file) {
  
  size_t size = file.size();
  
  Vect Wavelength(size);
  Vect Flux(size);
  Vect Error(size);
  Vect GalaxyFlux(size);
  Vect GalaxyError(size);
  Vect Valid(size);
  
  int i=0;
  for (DictFileCIterator line = file.begin(); line != file.end(); 
       ++line, ++i) {
    Wavelength(i) = line->Value("WAVE");
    Flux(i) = line->Value("SN_SPEC");
    Error(i) = line->Value("SN_ERR");
    GalaxyFlux(i) = line->Value("GAL_SPEC");
    GalaxyError(i) = line->Value("GAL_ERR");  
    Valid(i) = 1;  
  }
  
  LoadData(Wavelength,Flux,Error,Valid);
  
  galaxy = new MeasuredSpectrum();
  galaxy->redshift = redshift;
  galaxy->is_a_galaxy = true;
  galaxy->LoadData(Wavelength,GalaxyFlux,GalaxyError,Valid);
}



//==================== For Gemini Data (CB 18/04/07) ===========

void MeasuredSpectrum::ReadGEMINI(const DictFile& file) {
  
  size_t size = file.size();
  
  
  Vect Wavelength(size);
  Vect Flux(size);
  Vect Error(size);
  Vect GalaxyFlux(size);
  Vect GalaxyError(size);
  Vect Valid(size);
  
  int i=0;
  for (DictFileCIterator line = file.begin(); line != file.end(); ++line, ++i) {
    
    if( line->Value("MASK") != 1 && line->Value("WAVE") > 5300. && line->Value("WAVE") < 10000. )
      Valid(i)=1;
    else
      Valid(i)=0;
    
    Wavelength(i) = line->Value("WAVE");
    Flux(i) = line->Value("SN_SPEC");
    Error(i) = line->Value("SN_ERR");
    GalaxyFlux(i) = line->Value("GAL_SPEC");
    GalaxyError(i) = line->Value("GAL_ERR");
  }
  
  LoadData(Wavelength,Flux,Error,Valid);
  
  galaxy = new MeasuredSpectrum();
  galaxy->redshift = redshift;
  galaxy->is_a_galaxy = true;
  galaxy->LoadData(Wavelength,GalaxyFlux,GalaxyError,Valid);
}


//==================== For KECK Data (CB 21/05/07) ===========

void MeasuredSpectrum::ReadKECK(const DictFile& file) {
  
  size_t size = file.size();
  
  Vect Wavelength(size);
  Vect Flux(size);
  Vect Error(size);
  Vect GalaxyFlux(size);
  Vect GalaxyError(size);
  Vect Valid(size);

  int i=0;
  for (DictFileCIterator line = file.begin(); line != file.end(); ++line, ++i) {
    if( line->Value("MASK") != 1 && line->Value("WAVE") > 3400. && line->Value("WAVE") < 10000. )
      Valid(i)=1;
    else
      Valid(i)=0;
    
    Wavelength(i) = line->Value("WAVE");
    Flux(i) = line->Value("SN_SPEC");
    Error(i) = line->Value("SN_ERR");
    GalaxyFlux(i) = line->Value("GAL_SPEC");
    GalaxyError(i) = line->Value("GAL_ERR");  
  }
  
  LoadData(Wavelength,Flux,Error,Valid);
  
  galaxy = new MeasuredSpectrum();
  galaxy->redshift = redshift;
  galaxy->is_a_galaxy = true;
  galaxy->LoadData(Wavelength,GalaxyFlux,GalaxyError,Valid);
}

void MeasuredSpectrum::ApplyMask(const DictFile& file) {
  
  char key[100];
  for(int maskindex=0;maskindex<100;maskindex++) {
    sprintf(key,"Mask%03d_min",maskindex);
    if(!file.HasGlobalKey(key)) {
      //cout << "no global key " << key << endl;
      continue;
    }

    double lmin = atof(file.GlobalValue(key).c_str());
    sprintf(key,"Mask%03d_max",maskindex);
    double lmax = atof(file.GlobalValue(key).c_str());

    cout << "INFO MeasuredSpectrum::ApplyMask removing points in mask " << maskindex <<" between " << lmin << " and " << lmax <<  endl;

    for(MeasuredSpectrum::iterator it=begin(); it!=end(); ++it) {
      const double& lambda = (*it)->Wavelength();
      if(lambda>lmin && lambda<lmax) {
	(*it)->SetValid(0);
      }
    }
  }
}
void MeasuredSpectrum::DeweightTelluricLines() {
  
  cout << "INFO MeasuredSpectrum::DeweightTelluricLines" << endl;
  
  // get max flux for error floor
  //====================================
  double maxflux = 0;
  for(MeasuredSpectrum::iterator it=begin(); it!=end(); ++it) {
    if(maxflux<(*it)->Flux())
      maxflux = (*it)->Flux();
  }
  

  // lower weight in regions of atmospheric absoption lines
  // =============================================
  double lmin[] = {6850,7580};
  double lmax[] = {6970,7690};
  double min_error = maxflux/5;
  for(int m=0;m<2;m++) {
    for(MeasuredSpectrum::iterator it=begin(); it!=end(); ++it) {
      const double& lambda = (*it)->Wavelength();
      if(lambda>lmin[m] && lambda<lmax[m]) {
	double err = (*it)->FluxErr();
	if(err<min_error)
	  (*it)->SetFluxErr(min_error);
      }
    }
  }
}

void MeasuredSpectrum::EraseNotValid() {
  
  cout << "INFO MeasuredSpectrum::EraseNotValid" << endl;
  
  MeasuredSpectrum::iterator it=begin();
  while(it!=end()) {
    if((*it)->Valid()==0) {
      it = erase(it);
      continue;
    }
    it++;
  }
}


void MeasuredSpectrum::LoadData(const Vect& wavelength, const Vect& flux, const Vect& error, const Vect& valid) {
  
  size_t size = wavelength.size();
  if(size != flux.size()) { 
    cerr << "fatal error in MeasuredSpectrum::LoadData" << endl;
    exit(-12);
  }
  if(error.size()>0 && (size != error.size())) { 
    cerr << "fatal error in MeasuredSpectrum::LoadData" << endl;
    exit(-12);
  }
  if(valid.size()>0 && (size != valid.size())) { 
    cerr << "fatal error in MeasuredSpectrum::LoadData" << endl;
    exit(-12);
  }
  
  min_lambda =  wavelength(0);
  max_lambda =  wavelength(size-1);
  
  for(size_t i=0;i<size;i++) {
    MeasuredSpectrumPoint *p = new MeasuredSpectrumPoint();
    
    if(isnan(wavelength(i)) || isnan(flux(i)) ) {
      cerr << "WARNING nan in MeasuredSpectrum::LoadData" << endl;
      cout << "WARNING nan in MeasuredSpectrum::LoadData" << endl;
      continue;
    }
    if(error.size()>0) {
      if( isnan(error(i)) ) {
	cerr << "WARNING nan in MeasuredSpectrum::LoadData" << endl;
	cout << "WARNING nan in MeasuredSpectrum::LoadData" << endl;
	continue;
      }
    }


    p->SetWavelength(wavelength(i));
    p->SetFlux(flux(i));
    if(error.size()>0) {
      if(error(i)<=0) {
	cerr << "negative errors !!" << endl;
	p->SetFluxErr(flux(i));
      }else {
	p->SetFluxErr(error(i));
      }
    }else{
      p->SetFluxErr(1.);
    }
    if(valid.size()>0) {
      p->SetValid(int(valid(i)));
    }else{
      p->SetValid(1);
    }
    push_back(p);
  }
}


void MeasuredSpectrum::MilkyWayAbsorptionCorrection(const double& MWEBV) {
  if(milky_way_absorption_is_corrected) return;
  double data[2];
  data[0] = MWEBV;
  data[1] = 3.1; 
  double transmission = 1;
  // erase points if needed
  MeasuredSpectrum::iterator it=begin();
  while(it !=end()) {
    transmission = FluxExtinctionCardelli((*it)->Wavelength(),data);
    if(transmission<0.1) {
      it = erase(it); // huge assumption otherwise
    }else{
      double flux = (*it)->Flux()/transmission;
      double error = (*it)->FluxErr()/transmission;
      (*it)->SetFlux(flux);
      (*it)->SetFluxErr(error); 
      it++;
    }
  }
  milky_way_absorption_is_corrected = true;
}

void MeasuredSpectrum::SetName(const string& n) {
  name = n;
  if(name.size()> MAX_NUMBER_OF_CHARACTERS_FOR_NAME_OF_SPECTRUM) {
    cerr <<"Warning in MeasuredData, keeping only " << MAX_NUMBER_OF_CHARACTERS_FOR_NAME_OF_SPECTRUM << " characters of name " << name << endl;
    name=name.substr(0,MAX_NUMBER_OF_CHARACTERS_FOR_NAME_OF_SPECTRUM);  
  }
}

MeasuredSpectrum::MeasuredSpectrum(const string &FileName, bool erase_not_valid, bool deweight_telluric) {
  
  if (! FileExists(FileName)) {
    cerr << "cannot find " << FileName << endl;
    exit(-12);
  }
  cout << "Loading spectrum " << FileName << endl;
  is_a_galaxy = false;
  from_space = false;
  milky_way_absorption_is_corrected = false;
  galaxy=0;
  number_of_calibration_parameters=-1;
  name="";
  
  // read dictfile
  DictFile file(FileName);
  
  redshift = -1;
  if( file.HasGlobalKey("Z_HELIO")) {
    redshift  = atof(file.GlobalValue("Z_HELIO").c_str());
  } else {
    if( file.HasGlobalKey("Redshift")) {
      redshift  = atof(file.GlobalValue("Redshift").c_str());
    }else {
      cout << "missing redshift information for " << FileName << endl;
      abort();
    }
  } 
    
  date      = atof(file.GlobalValue("Date").c_str());
  if(file.HasGlobalKey("Name")) {
    SetName(file.GlobalValue("Name"));
  }
  if(file.HasGlobalKey("SN")) {
    sn_name = file.GlobalValue("SN");
  }
  
  infos = file.GlobalKeys;
  infos.AddKey("FILENAME",FileName);
  
  
  // phase is now in infos 
  /*
  if(file.HasGlobalKey("Phase")) {
    // cout << "has a phase info" << endl;
    phase = atof(file.GlobalValue("Phase").c_str());
  }else{
    phase     = 0;
  }
  */
  
  

  // check format
  if(file.HasGlobalKey("FROMSPACE")) from_space=true; 
  
  if(file.HasGlobalKey("FORMAT")) {
    
    string format = file.GlobalValue("FORMAT");
    bool ok = false;
    if(format=="VLT0") { ReadVLT0(file); ok=true; } 
    if(format=="VLT1") { ReadVLT1(file); ok=true;  } 
    if(format=="GEM") {ReadGEMINI(file); ok=true;  }
    if(format=="KECK") {ReadKECK(file); ok=true;  }
    if(format=="IUE") { ReadIUE(file); ok=true; } 
    
    if(format=="CFA" || format=="FAKE" || format=="STD") { Read(file); ok=true;} 
    if(!ok){
      cerr << "ERROR MeasuredSpectrum::MeasuredSpectrum unknown FORMAT " << format << " in file " << FileName << endl;
      exit(-12);
    }
    if(size()==0){
      cerr << "ERROR MeasuredSpectrum::MeasuredSpectrum file with no data" << FileName << endl;
      exit(-12);
    }
    
  }else{
    
    // check cfitsio
    // ====================================
#ifndef USECFITSIO
    if(file.HasGlobalKey("Fits_Fluxes") || file.HasGlobalKey("Fits_Errors")) {
      cerr << "ERROR MeasuredSpectrum::MeasuredSpectrum, current installation of software cannot deal with fits files" << endl;
      exit(-12);
    }
#endif
    
    // data temporary container
    Vect Wavelength,Wavelength_for_error,Flux,Error;
    
    // read fluxes
    // ====================================
    if(file.HasGlobalKey("ASCII_Fluxes"))
      read_an_ascii_array(fullpath(FileName,file.GlobalValue("ASCII_Fluxes")),Wavelength,Flux,2);
#ifdef USECFITSIO
    if(file.HasGlobalKey("Fits_Fluxes"))
      read_a_fits_array(fullpath(FileName,file.GlobalValue("Fits_Fluxes")),Wavelength,Flux);
#endif
    if(file.HasGlobalKey("ASCII_Fluxes_and_Errors"))
      read_an_ascii_array(fullpath(FileName,file.GlobalValue("ASCII_Fluxes_and_Errors")),Wavelength,Flux,2); 
    if(Flux.Size()==0) {
      cerr << "No filename for fluxes given in " << FileName << endl;
      exit(-12);
    }
    
    // read errors
    // ====================================
    if(file.HasGlobalKey("ASCII_Errors"))
      read_an_ascii_array(fullpath(FileName,file.GlobalValue("ASCII_Errors")),Wavelength_for_error,Error,2);
#ifdef USECFITSIO
    if(file.HasGlobalKey("Fits_Errors"))
      read_a_fits_array(fullpath(FileName,file.GlobalValue("Fits_Errors")),Wavelength_for_error,Error);
#endif
    if(file.HasGlobalKey("ASCII_Fluxes_and_Errors"))
      read_an_ascii_array(fullpath(FileName,file.GlobalValue("ASCII_Fluxes_and_Errors")),Wavelength_for_error,Error,3);

    if(file.HasGlobalKey("ASCII_Errors") || file.HasGlobalKey("Fits_Errors")) {
      // separate file for errors and fluxes, we have to check stuff in here
      
      double flux_wavelength_step  = Wavelength(1)-Wavelength(0);
      double error_wavelength_step = Wavelength_for_error(1)-Wavelength_for_error(0);
      double eps = 0.001;
      if(fabs(flux_wavelength_step/error_wavelength_step-1)>eps ) {
	cerr << "ERROR MeasuredSpectrum::MeasuredSpectrum flux and error files do not match " << endl;
	exit(-12);
      }
      if(Wavelength.size() != Wavelength_for_error.size()) { // we have to do something here
	cerr << "ERROR MeasuredSpectrum::MeasuredSpectrum flux and error files do not match " << endl;
	
#ifdef STORAGE

	// find range of match
	int first_valid_bin=0;
	while(Wavelength(first_valid_bin)<Wavelength_for_error(0)*(1.-eps)) {first_valid_bin++;};
	int last_valid_bin=Wavelength.size()-1;
	while(Wavelength(last_valid_bin)>Wavelength_for_error(Wavelength_for_error.size()-1)*(1.+eps)) {last_valid_bin--;};
	
	if(last_valid_bin<=first_valid_bin) {
	  cerr << "ERROR MeasuredSpectrum::MeasuredSpectrum flux and error files do not match " << endl;
	  exit(-12);
	}
	
	// find offset, so that Wavelength(i)=Wavelength_for_error(i+offset)*(1+-eps)
	int offset=-first_valid_bin;
	while(Wavelength(first_valid_bin)>Wavelength_for_error(first_valid_bin+offset)*(1.-eps) 
	      && (first_valid_bin+offset<int(Wavelength_for_error.size()))) 
	  {offset++;}; 
	if(offset>=int(Wavelength_for_error.size())) {
	  cerr << "ERROR MeasuredSpectrum::MeasuredSpectrum flux and error files do not match " << endl;
	  exit(-12);
	}
	
	cout << "INFO MeasuredSpectrum::MeasuredSpectrum range of valid bins = " << first_valid_bin << "," << last_valid_bin << " in 0," <<  Wavelength.size()-1 << endl;
	cout << "INFO MeasuredSpectrum::MeasuredSpectrum offset = " << offset << endl;
#endif
      }


    } 
    
  
    // if data are rest-frame, modify this
    //====================================
    if(file.HasGlobalKey("Frame") && (file.GlobalValue("Frame")=="Rest")) {
     
      Wavelength *= (1.+redshift);
      cout << "INFO MeasuredSpectrum::MeasuredSpectrumdata were restframe !! in file " << FileName << endl;
    }
    
    Vect Valid;
    LoadData(Wavelength,Flux,Error,Valid);
    
  } 
  
  // convert mask -> valid for spectra with "@MASK..."
  ApplyMask(file);
  
  if(erase_not_valid)
    EraseNotValid();
  
  if(deweight_telluric)
    DeweightTelluricLines();
  
  if(Galaxy())
    galaxy->ApplyMask(file);
  
  
  if(file.HasGlobalKey("UNITS")) {
    
    double scale = atof(file.GlobalValue("UNITS").c_str());
    cout << "INFO MeasuredSpectrum::MeasuredSpectrumdata will rescale fluxes by " << scale << " according to UNITS" << endl;

    for(list<CountedRef<MeasuredSpectrumPoint> >::iterator it=begin();it!=end();++it) {
      (*it)->SetFlux((*it)->Flux()*scale);
      (*it)->SetFluxErr((*it)->FluxErr()*scale);
    }
    if(galaxy) {
      for(list<CountedRef<MeasuredSpectrumPoint> >::iterator it=galaxy->begin();it!=galaxy->end();++it) {
	(*it)->SetFlux((*it)->Flux()*scale);
	(*it)->SetFluxErr((*it)->FluxErr()*scale);
       }
    }

  }
}

MeasuredSpectrum::~MeasuredSpectrum() {
  //cout << "destroying a measured spectrum" << endl;
  //cout << "clearing data ..." << endl;
  clear();
  //cout << "done" << endl;
  if(galaxy) {
    //cout << "deleting galaxy ... "  << endl;
    delete galaxy;
    //cout << "done" << endl;
  } 
}


bool MeasuredData::LoadSpectrum(const string& filename,double wavelenght_bin_size,bool warning) {
  if ( ! FileExists(filename)) {
    if(warning)
      cerr << "cannot find " << filename << endl;
    return false;
  }
  MeasuredSpectrum *spec = new MeasuredSpectrum(filename,true,true); // rm not valid and deweight telluric lines 
  if(spec->Name()=="") {
    char name[6];
    sprintf(name,"S%d",int(spectra.size()));
    spec->SetName(name);
  }
  spec->Rebin(wavelenght_bin_size);
  spec->MilkyWayAbsorptionCorrection(mwEbv); // divide by Galactic Extinction
  spectra.push_back(spec);
  
  return true;
}



void MeasuredSpectrum::Rebin(double wavelength_step) {
  
  if(wavelength_step<=0)
    return;

  if(galaxy) galaxy->Rebin(wavelength_step);

  // assume data are ordered ...
  double min_lambda = (*begin())->Wavelength();
  double max_lambda = (back())->Wavelength();
  
  MeasuredSpectrum rebinned_spectrum;
  size_t nbins = (size_t)((max_lambda-min_lambda)/wavelength_step);

  
  MeasuredSpectrum::const_iterator it=begin();
  double wavelength_step_in_spectrum_of_first_bin = -(*it)->Wavelength();
  it ++;
  wavelength_step_in_spectrum_of_first_bin += (*it)->Wavelength();
  
  double sumw,sumf;
  double bin_max_lambda,bin_min_lambda,weight;
  int n;
  bin_max_lambda = min_lambda;
  
  
  double wavelength_step_in_spectrum,previous_wavelength,wavelength;
  
  for(size_t ibin = 0; ibin < nbins ; ibin++) {
    sumw=0;
    sumf=0;
    n = 0;
    bin_min_lambda = bin_max_lambda;
    bin_max_lambda = bin_min_lambda + wavelength_step;
    
    //cout << "hello" << endl;

    previous_wavelength = 0;
    wavelength = 0;
    for(it=begin();it!=end();++it) {
      wavelength = (*it)->Wavelength();
      
      // step
      if(it!=begin()) {
	wavelength_step_in_spectrum = wavelength-previous_wavelength;
      }else{
	wavelength_step_in_spectrum = wavelength_step_in_spectrum_of_first_bin;
      }
      previous_wavelength = wavelength;
      
      double wavelength_spec_bin_min = wavelength-0.5*wavelength_step_in_spectrum;
      
      if(wavelength_spec_bin_min >  bin_max_lambda) break;

      double wavelength_spec_bin_max = wavelength+0.5*wavelength_step_in_spectrum;
      
      if(wavelength_spec_bin_max <  bin_min_lambda) continue;
      
      
      double bmin = max(wavelength_spec_bin_min,bin_min_lambda);
      double bmax = min(wavelength_spec_bin_max,bin_max_lambda);
      double frac = (bmax-bmin)/(wavelength_spec_bin_max-wavelength_spec_bin_min);
      
      
      //cout << bin_min_lambda << " " << bin_max_lambda << endl;
      //cout << wavelength_spec_bin_min << " " << wavelength_spec_bin_max << endl;
      //cout << bmin << " " << bmax << endl;
      //cout << "frac = " << frac << endl;
      
      weight = frac/sqr((*it)->FluxErr());
      sumw += weight;
      sumf += weight*((*it)->Flux());
      n++;
    }

#ifdef STORAGE
    while( ( it!=end() ) && ( (*it)->Wavelength() < bin_max_lambda ) ) {
      if( (*it)->Wavelength() >= bin_min_lambda ) {
	weight = 1./sqr((*it)->FluxErr());
	sumw += weight;
	//sumf += (*it)->Flux();
	sumf += weight*((*it)->Flux());
	n++;
      }else{
	cout << "mismatch" << endl;
      }
      ++it;
    }
    //--it;
#endif
    
    if(sumw>0) {
      //cout << n << endl;
      MeasuredSpectrumPoint *p = new MeasuredSpectrumPoint();
      //p->SetFlux(sumf/n);
      p->SetFlux(sumf/sumw);
      //p->SetFluxErr(sqrt(sumv)/n);
      p->SetFluxErr(sqrt(1./sumw));
      p->SetWavelength(0.5*(bin_min_lambda+bin_max_lambda));
      rebinned_spectrum.push_back(p);
    }
  }
  
  // load all of this in current spectrum
  clear();
  for(MeasuredSpectrum::const_iterator it=rebinned_spectrum.begin();it!=rebinned_spectrum.end();it++) {
    push_back(*it);
  }
  
}

bool MeasuredData::LoadSpectralSequence(const string& filename) {
  if( ! FileExists(filename)) {
    cout << "ERROR MeasuredData::LoadSpectralSequence file " << filename << " doesnt exist" << endl;
    return false;
  }
  DictFile file(filename);
  if(! file.HasGlobalKey("FORMAT")) {
    cout << "ERROR MeasuredData::LoadSpectralSequence no key FORMAT in file " << filename << endl;
    return false;
  }
  string format = file.GlobalValue("FORMAT");
  if(format != "SNFACTORY") {
    cout << "ERROR MeasuredData::LoadSpectralSequence read for FORMAT " << format  << " not implemented" << endl;
    return false;
  }
  z_heliocentric  = atof(file.GlobalValue("ZHELIO").c_str());
  z_cmb = atof(file.GlobalValue("ZCMB").c_str());

  if( file.HasGlobalKey("MWEBV")) {
    mwEbv = atof(file.GlobalValue("MWEBV").c_str());
  }else{
    cout << "WARNING MeasuredData::LoadSpectralSequence no key MWEBV, set it to 0" << endl;
    mwEbv = 0;
  }
  
  name = file.GlobalValue("OBJECT");
  
  int nspectra = atof(file.GlobalValue("NSPECTRA").c_str());
  if(nspectra<1) {
    cout << "ERROR MeasuredData::LoadSpectralSequence NSPECTRA =  " << nspectra << endl;
    return false;
  }
  
  
  
  double extinction_data[2];
  extinction_data[0] = mwEbv;
  extinction_data[1] = 3.1; // Rv
      
  

  double current_date = -12;
  
  MeasuredSpectrum *spectrum = 0;

  // now loop on data
  for (DictFileCIterator line = file.begin(); line != file.end(); 
       ++line)
    {
      double date = line->Value("MJD");
      
      if(current_date != date ) {
	// do something here
	if(spectrum) {
	  // save it
	  if(spectrum->size()>0)
	    spectra.push_back(spectrum);
	  else
	    cout << "WARNING MeasuredData::LoadSpectralSequence ignore empty spectrum " << spectrum->name << endl;
	}
	current_date = date;
	spectrum = new MeasuredSpectrum();
	spectrum->redshift = z_heliocentric;
	spectrum->date     = date;
	spectrum->phase    = 0; // unknown
	spectrum->number_of_calibration_parameters = 0; // SNFactory data are well calibrated
	spectrum->is_a_galaxy = false;
	spectrum->milky_way_absorption_is_corrected = false; // i don't know for the moment

	char specname[512];
	sprintf(specname,"%s-%d-%d",name.c_str(),int(date),int(100*date-100*int(date)));
	spectrum->name = specname;
      }
      
      // now add data point to current spectrum
      double flux_error = line->Value("SIG_FLUX");
      if(flux_error<0) {
	cout << "WARNING MeasuredData::LoadSpectralSequence ignore point with SIG_FLUX =  " << flux_error << endl;
	continue;
      }
      MeasuredSpectrumPoint * point = new MeasuredSpectrumPoint();
      point->wavelength = line->Value("WAVE");
      point->flux = line->Value("FLUX");
      point->fluxerr =  flux_error;
      if(mwEbv>0) {
	point->flux *= FluxExtinctionCardelli(point->wavelength,extinction_data);

      }
      spectrum->push_back(point);
    }


  

  return true;
}

const CountedRef<MeasuredPointList>  MeasuredData::GetPointListAndIndex(int index_global, int& index_in_plist) const {
  for(size_t i=0;i<pointLists.size();i++) { 
    const CountedRef<MeasuredPointList> plist = pointLists[i];
    index_in_plist = 0;
    for(MeasuredPointCIterator it = plist->begin();it!=plist->end(); ++it, ++index_in_plist) {
      if((*it)->Index() == index_global) return plist;
    }
  }
  cout << "fatal error in MeasuredData::GetPointListAndIndex, cannot find point for index " << index_global << endl;
  abort();
}

CountedRef<MeasuredPointList> MeasuredData::GetPointList(const string& filter_name, const string& mag_sys, const GlobalVal& globalvals, const double& MWEBV) {
  for(size_t i=0;i<pointLists.size();i++) {  
    CountedRef<MeasuredPointList> plist = pointLists[i];
    //cout << i << " " << plist->GetEffectiveFilter().Name() << " " << plist->magSystem << endl;
    if(filter_name == plist->GetEffectiveFilter().Name() && mag_sys == plist->magSystem) return plist;
  }
  cout << "INFO MeasuredData::GetPointList creating new point list for " << filter_name << endl;
  MeasuredPointList *plist = new MeasuredPointList();
  plist->magSystem = mag_sys;
  
  vector<string> SubStrings;
  DecomposeString(SubStrings,filter_name,"::");
  if(SubStrings.size()!=2) {
    cerr << "ERROR in MeasuredData::GetPointList, cannot decode " << filter_name << endl;
    exit(1);
  }
  string instrument_name = SubStrings[0];
  string band_name = SubStrings[1];
  
  plist->CreateFilter(instrument_name, band_name, globalvals, MWEBV);
  
  string key = "ZPERR_"+filter_name;
  if (globalvals.HasKey(key)){
    plist->zero_point_uncertainty = globalvals.getDoubleValue(key);
    cout << "INFO MeasuredData::GetPointList using " << key << " = " <<  plist->zero_point_uncertainty << endl;
  }
  pointLists.push_back(plist);
  return plist;
}


bool MeasuredData::ReadLightCurvesFile(const string& filename) {

  DictFile file(filename);
  
  //const MagSys* magsys = GetMagSys(file.GlobalValue("MAGSYS"));

  if( ! file.HasGlobalKey("SN")) {
    cout << "ERROR in MeasuredData::ReadLightCurvesFile, missing SN key in file " << filename << endl;
    exit(EXIT_FAILURE);
  }
  
  string sn_name = file.GlobalValue("SN");  
  // check if match name if set before
  if(name!="") {
    if(!CaseInsensitiveCompare(name,sn_name)) {
      cout << "trying to use data from different SNe: " << name << " " << sn_name << endl;
      exit(EXIT_FAILURE);
    }
  }
  name = sn_name;
  
  if(file.HasGlobalKey("Z_HELIO")) z_heliocentric = atof(file.GlobalValue("Z_HELIO").c_str());
  if(file.HasGlobalKey("Z_CMB"))   z_cmb = atof(file.GlobalValue("Z_CMB").c_str());
  if(file.HasGlobalKey("MWEBV"))   mwEbv = atof(file.GlobalValue("MWEBV").c_str());
  if(file.HasGlobalKey("DayMax"))   day_max = atof(file.GlobalValue("DayMax").c_str());
  if(file.HasGlobalKey("SURVEY"))   survey = file.GlobalValue("SURVEY").c_str();
  
  infos = file.GlobalKeys;
  infos.AddKey("FILENAME",filename);
  
  
  // read a covariance or weight matrix if exists
  Mat DataCovMat;  
  if(file.HasMat()) DataCovMat = file.GetMat();
  else {
  if(file.HasGlobalKey("WEIGHTMAT")) {
    
    string tmp_filename =  AddSlash(DirName(FullFileName(filename)))+file.GlobalValue("WEIGHTMAT");
    if(FileExists(tmp_filename)) {
      ifstream stream(tmp_filename.c_str());
      Mat WeightMat;
      WeightMat.readASCII(stream);
      stream.close();
      WeightMat.Symmetrize("U"); // cause may be small differences when reading the file
      for(size_t i=0;i<WeightMat.SizeX();i++) {
	if(WeightMat(i,i)<1.e-20) {
	  cout << "ERROR in MeasuredData::ReadLightCurvesFile, Null diagonal value in weight map" << endl;
	  exit(EXIT_FAILURE);
	}
      }
      
      DataCovMat = WeightMat;
      DataCovMat.SymMatInvert();     
    }else{
      FatalError("in MeasuredData::ReadLightCurvesFile cannot open matrix ASCII file "+filename,true);
    }
  }
  if(file.HasGlobalKey("COVMAT")) {
    
    string tmp_filename =  AddSlash(DirName(FullFileName(filename)))+file.GlobalValue("COVMAT");
    if(FileExists(tmp_filename)) {
      ifstream stream(tmp_filename.c_str());
      DataCovMat.readASCII(stream);
      stream.close();   
    }else{
      FatalError("in MeasuredData::ReadLightCurvesFile cannot open matrix ASCII file "+filename,true);
    }
  }
  }
  if(DataCovMat.SizeX()>0) {// check size
      if(DataCovMat.SizeX() != file.size() || DataCovMat.SizeY() != file.size()) {
	cout << "ERROR in MeasuredData::ReadLightCurvesFile, matrix of size " 
	     << DataCovMat.SizeX() << "x" << DataCovMat.SizeY() 
	     << " and list of size " << file.size()
	     << endl;
	exit(EXIT_FAILURE);
      }
    }
  
  

  int index = 0;

  bool single_band_in_file = true; // for the covariance matrix
  MeasuredPointList* first_plist = NULL;
  
  for(DictFileCIterator line = file.begin(); line != file.end(); ++line, ++index) {
    string filter_name = line->Value("Filter");
    string mag_sys = line->Value("MagSys");
    // find or create MeasuredPointList
    

    CountedRef<MeasuredPointList> plist = GetPointList(filter_name,mag_sys,file.GlobalKeys,mwEbv);

    // check if several bands in this file
    if(!first_plist) {
      first_plist = plist;
    }else{
      single_band_in_file &= (first_plist == plist);
    }


    MeasuredPoint *m = new MeasuredPoint(plist->EffectiveFilterWithMilkyWay,index);
    if (line->HasKey("Flux"))
      {
	m->flux =  line->Value("Flux");
	m->fluxerr =  line->Value("Fluxerr");
	if (m->fluxerr <= 0)
	  FatalError(" negative or null error in "+filename);
	m->zp = line->Value("ZP");
      }
    else if (line->HasKey("Mag"))
      {
	if (line->HasKey("Flux"))
	  FatalError("Cannot mix Flux And Mag in file"+filename);
	// have to compute the reference flux for mags
	double mag = line->Value("Mag");
	double errMag = line->Value("Magerr");
	if(errMag==0)
	  errMag = 0.001;
	m->zp = 0;
	m->flux = pow(10.,0.4*(m->zp-mag));
	// 0.921 = 0.4*log(10), with log(e) = 1.
	m->fluxerr = m->flux*0.921*errMag;
	}
    m->date = line->Value("Date");
    plist->push_back(m);
  }
  
  
  if(DataCovMat.SizeX()>0) { // now we need to deal with the covariance matrix
    if(single_band_in_file && first_plist) {
      first_plist->CovMat = DataCovMat; // easy
    }else{ // not implemented yet
      cout << "INFO in MeasuredData::ReadLightCurvesFile, splitting covariance in covmat per lightcurve" << endl;


      // create matrices for each plist
      for(size_t ip=0;ip<pointLists.size();ip++) {  
	CountedRef<MeasuredPointList> plist = pointLists[ip];
	plist->CovMat.allocate(plist->size(),plist->size());
      }

      
      // for each ij of input matrix, set matching kl in correct plist
      int k,l;
      for(size_t i=0; i<DataCovMat.SizeX(); i++) {
	
	// look for measurement of index=i in plists and get index k in it
	const CountedRef<MeasuredPointList> plist_i =  GetPointListAndIndex(i,k);
	
	
	plist_i->CovMat(k,k) = DataCovMat(i,i);
	
	for(size_t j=i+1; j<DataCovMat.SizeY(); j++) {
	  
	  if(DataCovMat(i,j) == 0) continue; // no need to process null covariances
	  
	  const CountedRef<MeasuredPointList> plist_j =  GetPointListAndIndex(j,l);
	  if( plist_i != plist_j ) {
	    cout << "WARNING in MeasuredData::ReadLightCurvesFile covariant term for measurements in different lightcurves cannot be accounted for " << i << " " << j << endl;
	    continue;
	  }
	  plist_i->CovMat(k,l) = DataCovMat(i,j);
	  plist_i->CovMat(l,k) = DataCovMat(i,j);
	  
	}
      }
      // now modify indices of measurements
      for(size_t ip=0;ip<pointLists.size();ip++) { 
	int index_in_plist = 0;
	for(MeasuredPointIterator it = pointLists[ip]->begin();it!=pointLists[ip]->end(); 
	    ++it, ++index_in_plist) {
	  (*it)->SetIndex(index_in_plist);
	}
      }
      
      // debug
      /*
      for(size_t ip=0;ip<pointLists.size();ip++) {  
	pointLists[ip]->CovMat.writeFits("covmat_"+pointLists[ip]->BandName()+".fits");
      }
      */

    }
  }

  for(size_t i=0;i<pointLists.size();i++) {  
    pointLists[i]->ApplyScaling();
  }
  
  return true;
}

#include <param.h>



bool MeasuredData::WriteLightCurvesFile(const string& filename, const FitParams* params) const {

  ofstream os(filename.c_str());
  os << setprecision(8);
  os << "@SN " << name << endl;

  if(z_heliocentric>-1)
    os << "@Z_HELIO " << z_heliocentric << endl;
  if(z_cmb>-1)
    os << "@Z_CMB " << z_cmb << endl; // TODO: check default value

  os << "@MWEBV " << mwEbv << endl; 
  
  

  if(pointLists.size()>0) {
    const CountedRef<MeasuredPointList> plist = pointLists[0];
    if(plist->has_focal_plane_coord) {
      os << "@X_FOCAL_PLANE " << plist->focal_plane_x_coord << endl; 
      os << "@Y_FOCAL_PLANE " << plist->focal_plane_y_coord << endl;
      
    }
  }

  if(params) {
    const LcParam* p = params->LocateParam("DayMax");
    if(p) {
      os << "@DayMax " << p->val << " " << p->err << endl;
      
    }
  }
  
  for (int i = 0; i<pointLists.size(); i++){
    os << "@ZPERR_"<< pointLists[i]->GetEffectiveFilter().Name() << " " << pointLists[i]->zero_point_uncertainty<< endl;
  }
  
  
  
  

  // check for covariances
  bool has_covmat = false;
  for(size_t i=0;i<pointLists.size();i++) 
    if(pointLists[i]->CovMat.SizeX()>0) {
      has_covmat = true;
      break;
    }
  if(has_covmat) {
    string lc_dirname = DirName(filename);
    string lc_filename = BaseName(filename);
    string covmat_filename = "covmat_"+CutExtension(lc_filename)+".dat";
    string covmat_path = lc_dirname+"/"+covmat_filename;
    os << "@COVMAT " << covmat_filename << endl;

    // fill block diagonal matrix
    int npoints = 0;
    for(size_t i=0;i<pointLists.size();i++)
      npoints += pointLists[i]->size();
    Mat CovMat(npoints,npoints);
    int first_index = 0;
    for(size_t i=0;i<pointLists.size();i++) {
      const CountedRef<MeasuredPointList> plist = pointLists[i];
      for(size_t j=0;j<plist->size();j++) {
	for(size_t i=0;i<plist->size();i++) {
	  CovMat(first_index+i,first_index+j)=plist->CovMat(i,j);
	}
      }
      first_index += plist->size();
    }
    // write it
    cout << "INFO MeasuredData::WriteLightCurvesFile writing covariance matrix " << covmat_path << endl;
    CovMat.writeASCII(covmat_path);
    // debugging
    // string covmat_fits_path = CutExtension(covmat_path)+".fits";
    // CovMat.writeFits(covmat_fits_path);
  }
  
  
 
  write_additional_info(filename,os,infos);
  

  os << "#Date : " << endl;
  os << "#Flux : " << endl;
  os << "#Fluxerr : " << endl;
  os << "#ZP : " << endl;
  os << "#Filter : instrument and band" << endl;
  os << "#MagSys : magnitude system" << endl;
  os << "#end " << endl;
  for(size_t i=0;i<pointLists.size();i++) {  
    const CountedRef<MeasuredPointList> plist = pointLists[i];
    string filter_name = plist->GetEffectiveFilter().Name();
    for(MeasuredPointCIterator ip = plist->begin(); ip!=plist->end(); ++ip) {
      os << (*ip)->Date() 
	 << " " << (*ip)->Flux() 
	 << " " << (*ip)->FluxErr()
	 << " " << (*ip)->Zp()
	 << " " << filter_name
	 << " " << plist->magSystem
	 << endl;
    }
  }
  os.close();
  cout << "INFO MeasuredData::WriteLightCurvesFile wrote " << filename << endl;
    
  return true;
}
