#ifndef MEASUREDDATA__H
#define MEASUREDDATA__H


#include <list>
#include <string>
#include <vector>
#include <cmath>

#include "countedref.h"
#include "filter.h"
#include "matvect.h"
#include "dictfile.h"

class Filter;
//! \file
//! handles reading of photometry results

class Measurement : public RefCount
{
  protected :
    
    double flux, fluxerr;
  double zp;
  double date, errdate;
   
  public :
    Measurement() {}
  
  
  double Flux() const { return flux;}
  double FluxErr() const {return fluxerr;}
  double Zp() const { return zp;}
  double Day() const { return date;}
  double Date() const { return date;}
  double Mag() const { if(flux<=0) return 99; return -2.5*log10(flux)+zp;}
  double MagErr1() const {if((1.-fluxerr/flux)<=0) return 99; return -2.5*log10(1.-fluxerr/flux);}
  double MagErr2() const {if((1.+fluxerr/flux)<=0) return 99; return -2.5*log10(1.+fluxerr/flux);}
  
  void SetFlux(double v)  {  flux=v;}
  void SetFluxErr(double v)  { fluxerr=v;}
  void SetZp(double v)  {  zp=v;}
  void SetDay(double v)  {  date=v;}
  void SetDate(double v)  {  date=v;}
  
  
};


class MeasuredPoint : public Measurement
{ // use refcounts in case one wants to put the same point in several std::lists.


    private :

    const  Filter& filter; // effective filter including Milky Way extinction
    int index;
    typedef std::map<std::string,std::string> MyMap;
    MyMap dict_file_line;

    
    public :
	MeasuredPoint(const Filter &F, int i) : Measurement(), filter(F), index(i) {}
    
    const Filter &GetFilter() const { return filter;}
    int Index() const {return index;} // index of measurement in the lighcturve (useful for fullfit)
    void SetIndex(int i) { index = i;} // index of measurement in the lighcturve (useful for fullfit)

    // store whatever was in the lightcurve file for later access. Borrow mechanism to DictFileEntry
    struct Val
    {
      const std::string &val;
      Val( const std::string &Sval) : val(Sval) {};
      operator double() const { return atof(val.c_str());};
      operator std::string() const { return val;};
    };

    //! this routine enable double toto = line.Value("STUFF");
    Val Value(const std::string &Key, const bool DiesIfAbsent = true) const;
    bool HasKey(const std::string &Key) const;
    void StoreLine(DictFileCIterator&);


    
    friend class MeasuredPointList;
    friend class MeasuredData;
};

class MeasuredSpectrumPoint : public Measurement
{ // use refcounts in case one wants to put the same point in several std::lists.


    private :
      double wavelength;
      int valid; // 0=no 1=yes
    
    public :
	MeasuredSpectrumPoint() : Measurement() {}
    
    double Wavelength() const {return wavelength;} 
    void SetWavelength(double w) {wavelength=w;} 
    int Valid() const {return valid;} 
    void SetValid(int i) {valid=i;} 
    
    friend class MeasuredSpectrum;
    friend class MeasuredData;
};


class DictFile;
class Vect;
class MeasuredSpectrum : public std::list<CountedRef<MeasuredSpectrumPoint> >, public RefCount
{      
  friend class MeasuredData;
  
  private :
  double redshift;
  double min_lambda;
  double max_lambda;
  double date;
  double phase;
  int number_of_calibration_parameters;
  std::string name; // can be used as an id.
  std::string sn_name; 
  
  //double min_calibration_error; // like 0.7 for 30% error, lowe bound
  //double max_calibration_error; // like 1.3
  
  double sigma_calibration; // uncertainty on 'grey'=achromatic calibration df/f , 0 by default
  
  std::string fluxes_fits_filename;
  std::string errors_fits_filename;
  
  bool is_a_galaxy;
  MeasuredSpectrum * galaxy;
  //CountedRef<MeasuredSpectrum> galaxy;
  void ReadVLT0(const DictFile& file);
  void ReadVLT1(const DictFile& file);
  void ReadIUE(const DictFile& file);
  void Read(const DictFile& file);
  void ReadGEMINI(const DictFile& file); // CB 18/04/2007 for Gemini data
  void ReadKECK(const DictFile& file); // CB 21/05/2007
  void LoadData(const Vect& wavelength, const Vect& flux, const Vect& error, const Vect& valid);
  void ApplyMask(const DictFile& file);
  bool milky_way_absorption_is_corrected;

  public :
  MeasuredSpectrum() {redshift=0;date=0;galaxy=0;is_a_galaxy=false;number_of_calibration_parameters=0;from_space=false;milky_way_absorption_is_corrected=false;sigma_calibration=0;};
  MeasuredSpectrum(const std::string &FileName,bool erase_not_valid=true, bool deweight_telluric=true);
  virtual ~MeasuredSpectrum();
  GlobalVal infos; // can contain extra information specific to this data

  std::string Name() const {return name;};
  void SetName(const std::string &n);
  double Redshift() const {return redshift;};
  double Date() const {return date;};
  void SetRedshift(double z)  {redshift=z;};
  void SetDate(double d)  {date=d;};
  double GetMinLambda() const {return min_lambda;};
  double GetMaxLambda() const {return max_lambda;};
  void EraseNotValid();
  void DeweightTelluricLines();
  
  //double GetMinCalibrationError() const {return min_calibration_error;};
  //double GetMaxCalibrationError() const {return max_calibration_error;};
  
  double GetPhase() const {return phase;};
 
  int NumberOfCalibrationParameters() const {return number_of_calibration_parameters;}
  void SetNumberOfCalibrationParameters(unsigned n)  {number_of_calibration_parameters = n;}
  double SigmaCalibration() const {return sigma_calibration;};
  void SetSigmaCalibration(const double& s) {sigma_calibration=s;};
  void Rebin(double wavelength_step=10);
  const MeasuredSpectrum* Galaxy() const {return (const MeasuredSpectrum*)(galaxy);};
  MeasuredSpectrum* Galaxy()  {return ( MeasuredSpectrum*)(galaxy);};
  bool IsAGalaxy() const {return is_a_galaxy;};
  bool from_space;
  void MilkyWayAbsorptionCorrection(const double& MWEBV);
  void Write(const std::string& filename, const std::map<std::string,std::string>* globalkeys=0) const;
  
};


//! stuff of a given band
class MeasuredPointList : public std::list<CountedRef<MeasuredPoint> >, public RefCount
{
    friend class MeasuredData;

 private : 
      
      Filter EffectiveFilter; // effective filter of the instrument 
    Filter EffectiveFilterWithMilkyWay; // effective filter including MilkyWay extinction using Cardelli's law + ...
    
    std::string magSystem;
    std::string instrumentName;
    std::string WeightMatFileName;
    void ApplyScaling();    
    
    bool has_focal_plane_coord;
    double focal_plane_x_coord;
    double focal_plane_y_coord;
    
    
    public :
      MeasuredPointList();
 MeasuredPointList(const Filter& effectivefilter) : EffectiveFilter(effectivefilter),EffectiveFilterWithMilkyWay(effectivefilter) {};
      MeasuredPointList(const std::string &FileName,bool applyscaling=true, double MWEBV=0);
      
      GlobalVal infos; // can contain extra information specific to this data
      
      void CreateFilter(const std::string &instrument_name, const std::string &bandName, const GlobalVal& globalvals, const double& MWEBV);
      double zero_point_uncertainty;

    std::string BandName() const {return EffectiveFilter.Name();}
    const Filter &GetEffectiveFilter() const { return EffectiveFilter;}
    const Filter &GetEffectiveFilterWithMilkyWay() const { return EffectiveFilterWithMilkyWay;}
    
    Mat CovMat; // filled only once from the WeightMat
    
    double GetSimulationZP() const;
    const std::string& GetMagSystem() const {return magSystem;};
    void SetMagSystem(const std::string &magsys){magSystem=magsys;};
    const std::string& GetInstrumentName() const {return instrumentName;};
    const std::string& GetWeightMatFileName() const {return WeightMatFileName;};
    
    MeasuredPointList::iterator erase(MeasuredPointList::iterator toremove);
    //void KeepOnlyPointsBetween(double observer_datemin,double observer_datemax); 
    void ApplyErrorFloor(double errorfloordval = 0.);
    //void AddMeasuredPoint(CountedRef<MeasuredPoint> mp) {push_back(mp);};
    //void CreateWeightMat();
};


typedef MeasuredPointList::iterator MeasuredPointIterator;
typedef MeasuredPointList::const_iterator MeasuredPointCIterator;

std::ostream &operator << (std::ostream &, const MeasuredPointList&);
std::ostream &operator << (std::ostream &, const MeasuredPoint&);



class FitParams;
class GlobalVal;
class MeasuredData 
{


  private :
  
    std::vector<CountedRef<MeasuredPointList> > pointLists;
    std::vector<CountedRef<MeasuredSpectrum> > spectra;
    
    double z_heliocentric;
    double z_cmb;
    double mwEbv; // Milky Way E(B-V)
    double day_max; // if exists in light curves
    std::string name; // of the object(SN)
    std::string ccd; 
    std::string field; 
    std::string snpath;
    std::string survey; 
    

 public :
    CountedRef<MeasuredPointList> GetPointList(const std::string& filter_name, const std::string& mag_sys, const GlobalVal& globalvals, const double& MWEBV);
 private :
    

    void ReadLightFile(const std::string& lightFileName);
    void Init();

    const CountedRef<MeasuredPointList> GetPointListAndIndex(int index_global, int& index_in_plist) const;
    
    public :
    MeasuredData();
    MeasuredData(const std::vector<std::string> &LcNames,bool applyscaling=true, const std::vector<std::string> * discarded_bands=0);




    GlobalVal infos; // can contain extra information specific to this data
    
    bool ReadLightCurvesFile(const std::string& filename);
    bool WriteLightCurvesFile(const std::string& filename, const FitParams* params = 0) const;

    bool LoadSpectralSequence(const std::string& filename);

    bool LoadSpectrum(const std::string& filename="spectrum.dat",double wavelenght_bin_size=-12,bool warning = true);

    //! number of available bands.
    size_t NBands() const { return pointLists.size();}
    
    //! number of spectra
    size_t NSpectra() const { return spectra.size();}
    
    const MeasuredPointList * GetMeasuredPointListByBand(const std::string& band) const;

    //! number of available bands.
    std::string BandName(const size_t N) const 
	{ return PointList(N).BandName();}

    //! access to bands  (labelled from 0 to NBands()-1)
    const MeasuredPointList &PointList(const size_t N) const
	{ return *pointLists[N];}

    //! access to bands  (labelled from 0 to NBands()-1
    MeasuredPointList &PointList(const size_t N) 
      { return *pointLists[N];}

    //! 
    const MeasuredSpectrum &Spectrum(const size_t N) const
	{ return *spectra[N];}

    //! 
    MeasuredSpectrum &Spectrum(const size_t N) 
      { return *spectra[N];}

    //! remove one pointstd::list
    void RemovePointList(const size_t N);
    void AddPointList(CountedRef<MeasuredPointList> mp) {pointLists.push_back(mp);};
    
    const std::string& SNPath() const {return snpath;};

    std::vector<std::string> BandNames() const;

    double Zhelio() const {return z_heliocentric;}
    double Zcmb() const {return z_cmb;}
    double DayMax() const {return day_max;}
    
    double MWEBV() const { return mwEbv;}
    
    void SetZhelio(const double& x)   {z_heliocentric = x;}
    void SetZcmb(const double& x)     {z_cmb = x;}
    void SetDayMax(const double& x)   {day_max = x;}    
    void SetMWEBV(const double& x)    {mwEbv = x;}
    void SetName(const std::string& iname) {name = iname;}
    
    std::string Name() const {return name;}
    std::string Survey() const {return survey;}
    
    std::string CCD() const {return ccd;}

    std::string Field() const {return field;}
    
    
    void Print (const int PrintLevel = 1) const;

    //void KeepOnlyPointsBetween(double observer_datemin,double observer_datemax) const; 

};

std::ostream &operator << (std::ostream &, const std::vector<std::string>);


#endif /* MEASUREDDATA__H */
