
#include "neutrinocosmo.h"
#include "fatalerror.h"

#include <cmath>
#include <iostream>
#include <stdlib.h>

#include <bspline1d.h>

using namespace std;

static double sq(const double &x) { return x*x;}
static double cub(const double &x) { return x*x*x;}

NuCosmo::NuCosmo(const double OmegaM, const double OmegaK, 
		 const double W0, const double Wa, 
		 const double OmegabH2, const double H0, 
		 const double T_cmb_K, 
		 const double mass_neutrino1_eV, const double mass_neutrino2_eV, const double mass_neutrino3_eV, const double n_eff_neutrinos) : 
  Cosmology(OmegaK),
  omegam(OmegaM),
  w0(W0), wa(Wa),
  obh2(OmegabH2), 
  h0(H0),
  tcmb_K(T_cmb_K),
  mc2_nu1_eV(mass_neutrino1_eV),
  mc2_nu2_eV(mass_neutrino2_eV),
  mc2_nu3_eV(mass_neutrino3_eV),
  neff_nu(n_eff_neutrinos) 
{
  
  k_bolz_eVK = 8.617332478e-5; // eV/K physical constant (PDG)
  
  // sigma_bolz = 5.670373e-8 W m−2 K−4 (PDG)
  // rhoc   = 1.05375e-5  h2 (GeV/c2 ) cm−3 (PDG)
  // w_gamma_prefactor = 4 sigma_bolz / c / rhoc = 4*(5.670373e-8/1.602176e-19/1e9/1e4)/2.9979e10/1.05375e-5 = 4.48132733356e-7
  w_gamma_prefactor_K4 = 4.48132733356e-7; // K^{-4}, derived physical constant from PDG
  
}

// not implemented virtual functions unused in sncosmo
void NuCosmo::SetParam(const int i, const double Val) {std::cout << "NuCosmo::SetParam not implemented" <<std::endl; abort();};
double NuCosmo::GetParam(const int IPar) const {std::cout << "NuCosmo::GetParam not implemented" <<std::endl; abort();};
int NuCosmo::NPar() const {std::cout << "NuCosmo::NPar not implemented" <<std::endl; abort();}


double NuCosmo::w_gamma() const { 
  return w_gamma_prefactor_K4*pow(tcmb_K,4);
}

double NuCosmo::Omega_gamma() const { 
  return w_gamma()/(h0*h0/10000);
}
  


void NuCosmo::dump(std::ostream &stream) const
{
  stream << " NuCosmo om=" << omegam << " ok=" << omegak
	 << " w0=" << w0 << " wa=" << wa
	 << " H0=" << h0 << " obh2=" << obh2
	 << " Tcmb=" << tcmb_K 
	 << " m_nu = " << mc2_nu1_eV << "," << mc2_nu2_eV << "," << mc2_nu3_eV
	 << " neff=" << neff_nu;
  }

double NuCosmo::zstar() const
{
// eq 66,67,68 Komatsu 2009
// same as in camb/modules.f90 function CosmomcTheta()
  double omegamh2 = Omegam()*sq(h0/100);
  double g1=0.0783*pow(obh2,-0.238)/(1.+ 39.5*pow(obh2,0.763));
  double g2 = 0.560/(1+21.1*pow(obh2,1.81));
  return 1048*(1.+0.00124*pow(obh2,-0.738))*(1+g1*pow(omegamh2,g2));
}


double NuCosmo::zdrag() const
{ // eq 4 from Eisenstein Hu, 1998
  
  

  double omegamh2 = Omegam()*sq(h0/100);
  double b1=0.313*pow(omegamh2,-0.419)*(1+0.607*pow(omegamh2,0.674));
  double b2 = 0.238*pow(omegamh2,0.223);
  return 1291*(1.+ b1*pow(obh2,b2))* pow(omegamh2,0.251)/
    (1+0.659*pow(omegamh2,0.828));
}

/*
  rho = rho_gamma a^-4 + rho_b a^-3
  p   = 1/3 * rho_gamma a^-4
  
  cs2 = dp/drho = dp/da / (drho/da)
      = (4/3 rho_gamma) / (4 rho_gamma + 3 rho_b)
      = 1/(3 + 9/4 rho_b/rho_gamma)
  cs  = 1/sqrt(3*(1+R)) with R = 3/4*rho_b/rho_gamma
  
  same as used in dsound_da_exact(a) camb/modules.f90
 
  sound speed in units of c
*/
double NuCosmo::cs(const double Z) const {
  double h = h0/100;
  double omegab = OBH2()/(h*h);
  double omega_gamma = Omega_gamma();
  double R   = 3*omegab/(4.*omega_gamma)/(1+Z); 
  return 1/sqrt(3.*(1+R));
}

/* as in dsound_da(a) camb/modules.f90 , approximate form used e.g. by CosmoMC for theta  */
/* very close to above as 1/(4*omega_gamma*h2)=1/(4*4.48132733356e-7*2.7255**4)=1.011e4 */
double NuCosmo::cs_camb(const double Z) const {
  double R   = 3e4*obh2/(1+Z); 
  return 1/sqrt(3.*(1+R));
}


/* it actually misses the c/H0 prefactor */

// rs = int cs(z)*dz/H(z) 
//    = (c/H0)  int (cs(z)/c)*dz/E(z)
// rs' =    (c/H0)  int (cs(a)/c)*da/a2/E(z)

double NuCosmo::rs(const double Z) const
{
  double amax= 1./(1+Z);
  double amin=0;
  int na = 500; 
  double da = (amax-amin)/na;
  double a=amin+0.5*da;
  double sum = 0;
    
  for (int i=0; i<na ; ++i , a+= da)
    {
      // we follow E(z) for clarity (at the expense of speed) but exclude dark energy
      double z=1/a-1;
      double Eval = E_with_or_without_darkenergy(z,false); // remove dark energy here     
      sum += cs(z)/Eval*da/(a*a);
    } 
  return sum;
}

double NuCosmo::rs_camb(const double Z) const
{
  double amax= 1./(1+Z);
  double amin=0;
  int na = 500; 
  double da = (amax-amin)/na;
  double a=amin+0.5*da;
  double sum = 0;
    
  for (int i=0; i<na ; ++i , a+= da)
    {
      // we follow E(z) for clarity (at the expense of speed) but exclude dark energy
      double z=1/a-1;
      double Eval = E_with_or_without_darkenergy(z,false); // remove dark energy here     
      sum += cs_camb(z)/Eval*da/(a*a);
    } 
  return sum;
}


  
double NuCosmo::l_a(const double Z,const int NStep) const
{
  return M_PI*Dm(Z,NStep)/rs(Z);
}

double NuCosmo::R(const double Z,const int NStep) const  
{
  return sqrt(Omegam())*Dm(Z,NStep);
}


//! for dealing with massive neutrinos
//! integral x^3*sqrt(1+(m/x)^2)/(exp(x)+-1) 
//! similar to camb/modules.f90 
//! subroutine nuRhoPres(am,rhonu,pnu)  for computation of splines
//! subroutine Nu_rho(am,rhonu)

#define ZETA3 1.2020569031595942853997
#define ZETA5 1.0369277551433699263313
#define ZETA7 1.0083492773819228268397
#define PI2   9.86960440108935799
#define PI4   97.4090910340024152


static double compute_Im(const double & mc2divkT) {
  int nx = 500;
  double dx=(30.)/nx; // use xmax=30 as in camb
  double sum=0;
  double x=0.00001;
  for(int i=0;i<nx;i++,x+=dx) {
    sum += x*x*x*dx*sqrt(1+mc2divkT*mc2divkT/x/x)/(exp(x)+1);
  }
  return sum;
}

double NuCosmo::Im(const double & mc2divkT) const {

  if(mc2divkT<0.01) return (7.*PI4/120)*(1+(5./(7*PI2))*mc2divkT*mc2divkT); // taylor expansion lower limit
  if(mc2divkT>600) return 3./2*(ZETA3*mc2divkT + (15*ZETA5)/2/mc2divkT); // taylor expansion upper limit
  
  static EqualStepBSpline1D *ImLogSpline = NULL;
  
  if(ImLogSpline) return exp(ImLogSpline->Value(log(mc2divkT)));
  
  {
    cout << "compute imspline ..." << endl;
    int nm = 100;
    double lm_min=log(0.01);
    double lm_max=log(600);
    double dlm=(lm_max-lm_min)/(nm-1);
    
    Vect log_m(nm);
    Vect log_Im(nm);
    for(int i=0;i<nm;i++) {
      log_m(i)  = lm_min+i*dlm;
      log_Im(i) = log(compute_Im(exp(log_m(i))));
    }
    ImLogSpline = new EqualStepBSpline1D();
    ImLogSpline->Fit(nm,log_m.Data(),log_Im.Data(),true);
    cout << "done" << endl;
  }
  // return value
  return Im(mc2divkT);
}


/*
Planck2013 XVI
"
We account for the additional energy density in neutrinos by
assuming that they have a thermal distribution with an effective
energy density rho_nu = Neff 7/8 (4/11)^(4/3) rho_gamma
with Neff = 3.046 in the baseline model (Mangano et al. 2002,
2005). This density is divided equally between three neutrino
species while they remain relativistic.

In our baseline model we assume a minimal-mass normal
hierarchy for the neutrino masses, accurately approximated for
current cosmological data as a single massive eigenstate with
mν = 0.06 eV (omega_nu h2 = sum m_nu /93.04 eV ~ 0.0006)
"
-> omega_nu ~ 0.06/93.04/(0.7**2) = 0.00132

omega_nu_h2 = omega_gamma_h2 * 90/(11 pi^4) * (Neff/3)^(3/4) * zeta(3) * sum_i  (mc2_i/kT_phot) 
omega_nu_h2 = 4/sigma/c*T_phot**4 *  90/(11*3.1415927**4)*(3.046/3)**(3./4)*1.2020569/(8.617332478e-5*2.7255)   * sum_i mc2_i/[1 eV]
omega_nu_h2 = 4.48132733356e-7*2.7255**4 * 90/(11*3.1415927**4)*(3.046/3)**(3./4)*1.2020569/(8.617332478e-5*2.7255)   * sum_i mc2_i/[1 eV]
omega_nu_h2 = sum_i mc2_i/[93.002 eV]

*/

//! rho_nu(z)/rho_crit(z=0)
double NuCosmo::rho_all_neutrinos_div_rhoc(const double& z) const {
  
  // uses only 
  
  /*
    fermions :
    <E> = g*(2pi hbar /( V^1/3))^-3 int d3p E / ( exp (E/(kT)) + 1 )
    <rho_fermions> = g/(2pi hbar)^3 int d3p E / ( exp (E/(kT)) + 1 )
    bosons :
    <rho_bosons>   = g/(2pi hbar)^3 int d3p E / ( exp (E/(kT)) - 1 )
    
    with the integration variable x = pc/kT (and h=2pi*hbar)
    IF we assume the distribution is frozen when particles are still relativistic !
    (p and T redshift at the same time so the product E/kT=(pc)/kT is unchanged)

    <rho_fermions> = g (kT)^4 (4 pi) (hc)^-3  int  dx x^3 sqrt( 1 + ((mc2/kT)/x)^2 )/(exp(x)+1)
    <rho_bosons>   = g (kT)^4 (4 pi) (hc)^-3  int  dx x^3 sqrt( 1 + ((mc2/kT)/x)^2 )/(exp(x)-1)
    

    photons : m=0,g=2
    <rho_phot> = g * (kT)^4 4 pi (hc)^-3  int  dx x^3 /(1-exp(x))
               = g * (kT_phot)^4 4 pi (hc)^-3 * pi^4/15 
               = 8/15 * (kT_phot)^4 * pi^5 (hc)^-3    
	       = 4 * sigma_boltz /c * T^4   ( sigma_boltz = 2/15*k^4*pi^5/(h^3c^2) )

    for neutrinos, when relativistic : 
    
    <rho_nu_rel> = g (kT_nu)^4 4 pi hc^-3  int  dx x^3 /(1+exp(x))
                 = g (kT_nu)^4 4 pi hc^-3 7 pi^4/120
                 = g 7/30 (kT_nu)^4 pi^5 hc^-3      

        
    with g=2 per neutrino , with N_nu neutrinos
    
    <rho_nu_rel> = N_nu * 2 * 7/30 (kT_nu)^4 pi^5 hc^-3
                 = N_nu * 2 * 7/30 * 15/8 (T_nu/T_phot)^4 <rho_phot>
		 = N_nu * 7/8 (T_nu/T_phot)^4 <rho_phot>
    
	     
    introducing an effective number of neutrinos :
    T_nu/T_phot = (4/11)^(1/3) (Neff/3)^(1/4)

    
    <rho_nu_rel>     = N_nu * 7/8 (4/11)^(4/3) (Neff/3) <rho_phot> 
    
    <rho_nu>         = <rho_nu_rel> *( 120/(7 pi^4) ) * int  dx x^3 sqrt( 1 + ((mc2/kT_nu)/x)^2 )/(1+exp(x))
    
    <rho_nu_non_rel> = <rho_nu_rel> *( 120/(7 pi^4) ) * 3/2 * zeta(3) * (mc2/kT_nu)
                     = (N_nu) * <rho_phot>  * 7/8 (4/11)^(4/3) * (Neff/3) * ( 120/(7 pi^4) ) * 3/2 * zeta(3) * (mc2/kT_nu)
                     = (N_nu) * <rho_phot>  * 7/8 (4/11)^(4/3) * (Neff/3) * ( 120/(7 pi^4) ) * 3/2 * zeta(3) * mc2 (4/11)^(-1/3) (Neff/3)^(-1/4) (k T_phot)^-1
		     = (N_nu) * <rho_phot>  * 7/8  * 120/7 * 3/2 * (4/11) * 1/pi^4 * (Neff/3)^(3/4) * zeta(3) * (mc2/kT_phot) 
		     = (N_nu) * <rho_phot>  * 90/(11 pi^4) * (Neff/3)^(3/4) * zeta(3) * (mc2/kT_phot) 
		     
		     
                     
  */
  
  
  double kTphot  = k_bolz_eVK*tcmb_K*(1+z); // eV 
  double kTnu = pow(4./11,1./3)*pow(neff_nu/3,1./4)*kTphot; // eV
  double m1 = mc2_nu1_eV/kTnu; // no dim.
  double m2 = mc2_nu2_eV/kTnu; // no dim.
  double m3 = mc2_nu3_eV/kTnu; // no dim.
  
  double rho_nu_div_rho_gamma_one_relativistic_neutrino = 7./8*pow(4./11,4./3)*(neff_nu/3);
  
  double rho_nu_div_rho_gamma_3_neutrinos = rho_nu_div_rho_gamma_one_relativistic_neutrino*(120./(7*PI4)) * ( Im(m1)+Im(m2)+Im(m3) );
  
  return rho_nu_div_rho_gamma_3_neutrinos*Omega_gamma()*pow(1+z,4);
}

//! rho_nu(z)/rho_crit(z=0)
double NuCosmo::rho_one_neutrino_div_rhoc(const double& z, const double& mc2_nu_eV) const {
  double kTphot  = k_bolz_eVK*tcmb_K*(1+z); // eV 
  double kTnu    = pow(4./11,1./3)*pow(neff_nu/3,1./4)*kTphot; // eV
  double m       = mc2_nu_eV/kTnu; // no dim.
  double rho_nu_div_rho_gamma_one_relativistic_neutrino = 7./8*pow(4./11,4./3)*(neff_nu/3);
  
  double rho_nu_div_rho_gamma_neutrino = rho_nu_div_rho_gamma_one_relativistic_neutrino*(120./(7*PI4)) * ( Im(m) );
  
  return rho_nu_div_rho_gamma_neutrino*Omega_gamma()*pow(1+z,4);
}

double NuCosmo::Omega_neutrino() const { return rho_all_neutrinos_div_rhoc(0);}

double NuCosmo::E_with_or_without_darkenergy(const double z, bool with_dark_energy) const {
  
  
  double omega_nu = Omega_neutrino();
  
  // here we assume that by definition, omegam contains the neutrino relic density
  // it's ok because the relativistic component is anyway bound to be < 0.227*3.04*omega_gamma <= 4e-5
  
  double omega_cdm_and_baryons = omegam-omega_nu; 
  
  double rho = 0;
  double oz=1+z;
  double oz2=oz*oz;
  double oz3=oz2*oz;
  double oz4=oz3*oz;
  
  // add matter
  rho += (omega_cdm_and_baryons)*oz3;
  
  // add neutrinos
  rho += rho_all_neutrinos_div_rhoc(z);
  
  // add photons
  double omega_gamma = Omega_gamma();
  rho += omega_gamma * oz4; 
  
  // curvature
  rho += omegak*oz2;
  
  if(with_dark_energy) { // add dark energy
    double omegax = (1-omega_cdm_and_baryons-omegak-omega_gamma-omega_nu);
    rho += omegax * exp(3.*(log(oz)*(1+w0+wa)-wa*z/(oz)));
  }
  
  
  return sqrt(rho);
}

double NuCosmo::integ_hinv(const double &ZMin, const double &ZMax, 
			    const int NStep) const 
{ 
  // cout << "NuCosmo::integ_hinv" << endl;
  // change of variable u = 1./sqrt(1+z), because it gives more accurate results on all Z ranges.
  // Z = Z_CMB even works pretty well (6e-6, with 100 steps)
  double umax = 1./sqrt(1+ZMin);
  double umin = 1./sqrt(1+ZMax);
  double du = (umax-umin)/NStep;
  double u = umin+0.5*du;
  double sum=0;
  for (int i=0; i<NStep; ++i, u+=du) {
    double z=1/(u*u)-1;
    double dz = 2*du/(u*u*u);
    sum += dz/E(z);
  }
  return sum;
}
