#ifndef NUCOSMO__H
#define NUCOSMO__H

#include <cmath>
#include "iostream"
#include "cosmology.h"

class NuCosmo : public Cosmology
{
  
 public :
  
  double omegam;
  
  double w0;

  double wa;

  double obh2;
  
  double h0;
  
  /* temperature of CMB at z=0 in K, hard-coded in constructor  */
  double tcmb_K;
  
  /* Mass of neutrinos assuming a single degenerate mass of three neutrinos, in eV, 0 by default */
  double mc2_nu1_eV;
  double mc2_nu2_eV;
  double mc2_nu3_eV;
  
  /* Effective number of neutrinos, hard-coded in constructor ? */
  double neff_nu;
  
  
  /* Boltzmann constant, in units of eV/K, hard-coded in constructor */
  double k_bolz_eVK;
  
  /* prefactor to compute omega_photons */
  /* 4 times Stefan-Boltzmann constant divided by c and rho_critical, for H0=100 km/s/Mpc, hard-coded in constructor, in K^{-4} */
  /* w_gamma = omega_gamma * h^2 = w_gamma_prefactor * tcmb_K^4 */
  double w_gamma_prefactor_K4;
  
 private :
  
  double E_with_or_without_darkenergy(const double z, bool with_dark_energy) const;

public:
  NuCosmo(const double OmegaM, const double OmegaK, 
	  const double W0, const double Wa,
	  const double OmegabH2, const double H0, 
	  const double T_cmb_K, 
	  const double mass_neutrino1_eV, const double mass_neutrino2_eV, const double mass_neutrino3_eV, 
	  const double n_eff_neutrinos);

  double Omegam() const { return omegam;}
  double W0() const { return w0;}
  double Wa() const { return wa;}
  double W(const double &Z) const { return w0+wa*Z/(1+Z);}

  // not implemented virtual functions unused in sncosmo, abort if called
  void SetParam(const int i, const double Val);
  double GetParam(const int IPar) const;
  int NPar() const;

  double H0() const { return h0;}
  double OBH2() const { return obh2;}
  double NeutrinoMass1() const { return mc2_nu1_eV;}
  double NeutrinoMass2() const { return mc2_nu2_eV;}
  double NeutrinoMass3() const { return mc2_nu3_eV;}
  double Neff() const { return neff_nu;}
  double Tcmb() const { return tcmb_K;}
  double w_gamma() const;
  double Omega_gamma() const;
  double Omega_neutrino() const;
  double E(const double z) const { return E_with_or_without_darkenergy(z,true);}
  double ns; // for some CMB  prior
  double delta; // normalisation of P(k), for Huterer Planck  prior
  
  double integ_hinv(const double &ZMin, const double &ZMax, 
		    const int NStep=100) const;

  void dump(std::ostream &stream) const;

  double zstar() const;

  /* zdrag */
  double zdrag() const;

  /* comoving sound horizon */
  double cs(const double Z) const;
  double cs_camb(const double Z) const;
  double rs(const double Z) const;
  double rs_camb(const double Z) const;

  double l_a(const double Z, const int NStep=100) const;

  double R(const double Z, const int NStep=100) const;

  //! for dealing with massive neutrinos
  //! integral x^3*sqrt(1+(m/x)^2)/(exp(x)+1)
  double Im(const double & mc2divkT) const;
  
  //! rho_nu(z)/rho_crit(z=0)
  double rho_all_neutrinos_div_rhoc(const double& z) const;
  double rho_one_neutrino_div_rhoc(const double& z, const double& mc2_nu_eV) const;
  
  Cosmology *Clone() const { return new NuCosmo(*this);}
  ~NuCosmo() {}

};


#endif

