#include <iostream>
#include <fstream>
#include <cmath>
#include <time.h>
#include <algorithm>


#include "matvect.h"
#include "intmatvect.h"
#include "sparsevect.h"
#include "sampledfunction.h"
#include "integflux.h"



#include "nurbsbasis.h"


using namespace std;

NURBSBasis::NURBSBasis()  {
  xmin=0;
  xmax=0;
  ymin=0;
  ymax=0;   
  acc_is_synchronized = false;
};
  
int NURBSBasis::Nx() const {return X.SizeX();}
int NURBSBasis::Ny() const {return X.SizeY();}


void NURBSBasis::AllocateParameters() {

  //cout << "in AllocateParameters" << endl;
  
  I.allocate(X.SizeX(),X.SizeY());
  int *index  = I.NonConstData();
  int *fixed  = F.NonConstData();
  
  int nx = I.SizeX();
  int ny = I.SizeY();
  int n  = nx*ny;
  
  // set boundaries
  for(int i=0;i<nx;i++) {
    fixed[i+nx*0] = -2;
    index[i+nx*0] = -1;

    fixed[i+nx*(ny-1)] = -2;
    index[i+nx*(ny-1)] = -1;
  }
  
  for(int j=1;j<ny-1;j++) {
    fixed[0+nx*j] = -2;
    index[0+nx*j] = -1;
    
    fixed[nx-1+nx*j] = -2;
    index[nx-1+nx*j] = -1;
  }
  
  // first pass : count number of parameters and set the indices of free control points
  int running_index = 0;
  for(int i=0;i<n;i++,fixed++) {
    if(*fixed==-1) { // control point that is not fixed
      index[i] = (running_index++); // incremental index
    }else{
      if(*fixed==-2) 
	index[i] = -1; // no parameter index associated to this vertex
    }
  }
  
  if(running_index==0) {
    cout << "FATAL ERROR in AllocateParameters() : there is no free parameter" << endl;
    abort();
  }

  // second pass set the indices of the fixed control points
  fixed  = F.NonConstData();
  for(int i=0;i<n;i++,fixed++) {
    if(*fixed>=0)
      index[i] = index[*fixed];
  }
  

  P.allocate(running_index);
  cout << "number of allocated parameters = " << running_index << endl;
}

bool NURBSBasis::FixPoint(int i, int j, int i_of_other, int j_of_other, int direction_for_nearest) {
    
  int nx   = Nx();
  int ny   = Ny();
  
  if(i<0 || i>=nx || j<0 || j>=ny) {
    cout << "ERROR  point " << i << " " << j << " outside grid" << endl;
    return false;
  } 
  
  int* index = I.NonConstData();
  int* fixed = F.NonConstData();
  int current_index = i+nx*j;
  
  if(fixed[current_index]>=0) {
    cout << "WARNING already fixed point " << i << " " << j << endl;
    return false;
  }
  if(fixed[current_index]==-2) {
    cout << "WARNING cannot fix a point at a boundary vertex " << i << " " << j << endl;
    return false;
  }
  
  if(i_of_other>=0 || j_of_other>=0) {
    if(i_of_other<0 || i_of_other>=nx || j_of_other<0 || j_of_other>=ny) {
      cout << "ERROR  other point " << i_of_other << " " << j_of_other << " outside grid" << endl;
      return false;
    }
    int index_of_other = i_of_other+nx*j_of_other;
    if(fixed[index_of_other]!=-1) {
      cout << "ERROR other point is not a free control point" << endl;
      return false;
    }

    fixed[current_index] = index_of_other;
    AllocateParameters();
    return true;
  }
  
  
  
  
  i_of_other = -1;
  j_of_other = -1;
  
  switch(direction_for_nearest) {
  case 2 : // to the left
    {
      for(int k=i-1;k>=0;k--) {
	if(fixed[k+nx*j]==-1) {i_of_other = k; j_of_other = j; break;}
      }
    }
    break;
  case 0 : // to the right
    {
      for(int k=i+1;k<nx;k++) {
	if(fixed[k+nx*j]==-1) {i_of_other = k; j_of_other = j; break;}
      }
    }
    break;
  case 1 : // up
    { 
      for(int l=j+1;l<ny;l++) {
	if(fixed[i+nx*l]==-1) {i_of_other = i; j_of_other = l; break;}
      }
    }
    break;
  case 3 : // down
    { 
      for(int l=j-1;l>=0;l--) {
	if(fixed[i+nx*l]==-1) {i_of_other = i; j_of_other = l; break;}
      }
    }
    break;
    
  default :
    cout << "ERROR in NURBSBasis::FixPoint unexpected direction = " << direction_for_nearest << endl;
    return false;
  }
  
  if(i_of_other==-1) {
    cout << "ERROR in NURBSBasis::FixPoint could not find a free contol point in direction = " << direction_for_nearest << endl;
    return false;
  }

  
  cout << "fixing point (" << i << "," << j <<") to the node (" << i_of_other << "," << j_of_other << ") with index=" << index[i_of_other+nx*j_of_other] << endl;
  fixed[i+nx*j]=i_of_other+nx*j_of_other;
  
  AllocateParameters();

  return true;
}


// nx and ny here are the number of parameters !!!
// so the number of vertices = (nx+2)*(ny+2)
// first_x , last_x ... define the valid area of the model 
void NURBSBasis::Uniform(int nx, const double& first_x, const double& last_x, 
			 int ny, const double& first_y, const double& last_y ) {
  
  if(nx<3) abort(); // au pif
  if(ny<3) abort();
  
  int n_vertices_x = nx + 2;
  int n_vertices_y = ny + 2;
  
  // total number of segments = n_vertices_x -1
  // number of segments  in valid area = total number of patches - 3
  
  double segment_size_x = (last_x-first_x)/(n_vertices_x-4);
  double segment_size_y = (last_y-first_y)/(n_vertices_y-4);
  
  X.allocate(n_vertices_x,n_vertices_y);
  Y.allocate(n_vertices_x,n_vertices_y);
  F.allocate(n_vertices_x,n_vertices_y);
  
  for(int i=0;i<n_vertices_x;i++) {
    
    double x = first_x+(i-2)*segment_size_x;
    
    for(int j=0;j<n_vertices_y;j++) {
      
      double y = first_y+(j-2)*segment_size_y;
      
      X(i,j) = x;
      Y(i,j) = y;
      if(i>0 && i<n_vertices_x-1 && j>0 && j<n_vertices_y-1)
	F(i,j) = -1;
      else
	F(i,j) = -2;
    }
  }
  ComputeXYRange();
  
  if(xmin != first_x || xmax != last_x) abort();
  
  AllocateParameters();
  
}

double NURBSBasis::NF1D(const double &x, const int i, const int d, const double* k, const int n) const {
  
  if(x<k[0] || x>k[n-1]) return 0;
  if(i+d>=n) return 0;
  
  
  if(d==0) {
    if( k[i] <= x && i+1<n && x < k[i+1] ) { 
      return 1;
    } else {
      return 0;
    }
  }  
  
  double res = (x-k[i])/(k[i+d]-k[i])*NF1D(x,i,d-1,k,n);
  if( i+d+1<n && i+1<n)
    res += (k[i+d+1]-x)/(k[i+d+1]-k[i+1])*NF1D(x,i+1,d-1,k,n);
  
  return res;
}

// this is the function of the basis associated to the control point of indices i,j
double NURBSBasis::NF2D(const double &x, const double &y, int i, int j) const {
  
  int degree=2;
  
  /////////////////////////////
  
  int nx = Nx();
  int ny = Ny();
  const double *xv = X.Data();
  
  int offset=1;
  if(i<offset) return 0;
  if(j<offset) return 0;
  if(i>=nx || j>=ny) abort();
  
  int kmax=4;
  if(kmax+i-offset>nx) kmax=nx-i+offset;
  double xnfunc = NF1D(x,0,degree,&xv[(i-offset)+nx*j],kmax);
  
  if(xnfunc==0) return 0;
  
  /////////////////////////////
  
  
  const double *yv = Y.Data();
  double knot[4];
  kmax=4;
  if(kmax+j-offset>ny) kmax=ny-j+offset;
  for(int k=0;k<kmax;k++) {      
    knot[k]=yv[i+nx*(k+j-offset)]; // = Y(i,k+j-offset);
  }
  double ynfunc = NF1D(y,0,degree,knot,kmax);
    
  return xnfunc*ynfunc;
}


/* this routine assumes a sorted array */
static int LocateBelow(const double *array, int NValues, const double &Where)
{
  if (NValues == 0) return 0;
  int low =0;
  int high = NValues-1;
  if (Where > array[high]) abort(); // return high;
  if (Where < array[low]) abort(); // return -1;
  while (high - low > 1)
    {
      int mid = (low+high)/2;
      if (array[mid] > Where) high = mid;
      else low = mid;
    }
  return low;
}

bool NURBSBasis::Acc_GetIndices(const double& x, const double& y, int& imin, int& imax, int& jmin, int& jmax) const {
  


  if(! Acc_IsSync()) const_cast<NURBSBasis*>(this)->Acc_ComputeIndices();
  
  if(x<xmin || x>=xmax || y<ymin || y>=ymax) return false;

#if 0
  // disabled Acc
  imin=0; imax=Nx();
  jmin=0; jmax=Ny(); 
  return true;
#endif
  
  int table_i = LocateBelow(Acc_X.Data(),Acc_X.Size(),x);
  int table_j = LocateBelow(Acc_Y.Data(),Acc_Y.Size(),y);
  int table_ip1 = min(table_i+1,int(Acc_MinX.SizeX()-1));
  int table_jp1 = min(table_j+1,int(Acc_MinX.SizeY()-1));
       
  imin = min(
	     min(Acc_MinX(table_i,table_j),Acc_MinX(table_i,table_jp1)),
	     min(Acc_MinX(table_ip1,table_j),Acc_MinX(table_ip1,table_jp1))
	     );
  
  jmin = min(
	     min(Acc_MinY(table_i,table_j),Acc_MinY(table_i,table_jp1)),
	     min(Acc_MinY(table_ip1,table_j),Acc_MinY(table_ip1,table_jp1))
	     );
  
  imax = max(
	     max(Acc_MaxX(table_i,table_j),Acc_MaxX(table_i,table_jp1)),
	     max(Acc_MaxX(table_ip1,table_j),Acc_MaxX(table_ip1,table_jp1))
	     );
  
  jmax = max(
	     max(Acc_MaxY(table_i,table_j),Acc_MaxY(table_i,table_jp1)),
	     max(Acc_MaxY(table_ip1,table_j),Acc_MaxY(table_ip1,table_jp1))
	     );
  return true; 
}

bool NURBSBasis::Acc_GetLeadingVertex(const double& x, const double& y, int& ibest, int& jbest) const {
  int imin,imax,jmin,jmax;
  if(! Acc_GetIndices(x,y,imin,imax,jmin,jmax) ) return false;
  
  double vmax=0;
 
  for(int i=imin;i<imax;i++)
    for(int j=jmin;j<jmax;j++) {
      double v = NF2D(x,y,i,j);
      if(v>vmax) {
	vmax=v;
	ibest=i;
	jbest=j;
      }
    }
  return true;
}
  

SparseVect NURBSBasis::Derivatives(const double& x, const double& y) const {
  
  SparseVect H;
  
  int imin,imax,jmin,jmax;
  if(! Acc_GetIndices(x,y,imin,imax,jmin,jmax)) return H;
  
  int nx = Nx();
  int ntot = nx*Ny();
  
  int nhits  = 0;
  
  const int *index = I.Data();

  

  for(int i=imin;i<imax;i++) {
    for(int j=jmin;j<jmax;j++) {
 
#ifdef  NURBS_FORCE_NULL_OUTSIDE_BOUNDARIES     
      if(F(i,j)!=-1) return SparseVect();
#endif
      double h = NF2D(x,y,i,j);
      
      
      if(h>0) {
	nhits ++;
	
	int ii = index[i+j*nx];
	
	if(ii>=ntot) {
	  abort();
	}
	
	if(ii>=0) {
	  H.set(ii,H.get(ii)+h);
	}
      }
    }
  }
  // if(nhits < 10) return SparseVect(); // return empty sparsevect
  double sum = 0;
  for(SparseVectConstIterator k = H.begin(); k != H.end(); k++)
    sum += k->second;
  if(sum>0) {
    double suminv = 1./sum;
    for(SparseVectIterator k = H.begin(); k != H.end(); k++)
      k->second *= suminv;
  }
  
  
  return H;
}
  
bool NURBSBasis::Fit(int nx, const double*x, int ny, const double* y,  double** f) {
  int n = P.size();
  Mat A(n,n);
  Vect B(n);
  double *a = A.NonConstData();
  double *b = B.NonConstData();
  
  for(int i=0;i<nx;i++) {
    
    //cout << "x[" << i << "]=" << x[i] << endl;
      
    for(int j=0;j<ny;j++) {
      
      SparseVect H = Derivatives(x[i],y[j]);
      for(SparseVectConstIterator k = H.begin(); k != H.end(); k++) {
	b[k->first] += f[i][j]*k->second;
	//b[k->first] += 1.*k->second; // TEST OF METHOD
	//B(k->first) += 1.*k->second;
	
	for(SparseVectConstIterator l = H.begin(); l != H.end(); l++) {
	  a[k->first + n *l->first] += k->second * l->second;
	  //A(k->first,l->first) += k->second * l->second; 
	}
      }
    }
  }
  
  //A.writeFits("A.fits");
  int status = cholesky_solve(A,B,"L");
  if(status != 0) return false;
  memcpy(P.NonConstData(),B.Data(),sizeof(double)*n);
  return true;
}
  
bool NURBSBasis::Fit(Grid2DFunction &func) {
  return Fit(func.NstepX(),func.xvals(),func.NstepY(),func.yvals(),func.fvals());
}


double NURBSBasis::Value(const double& x, const double& y) const {
  SparseVect H = Derivatives(x,y);
  const double *p = P.Data();
  double res = 0;
  for(SparseVectConstIterator k = H.begin(); k != H.end(); k++) {
    res += p[k->first]*k->second;
  }
  return res;
}



void NURBSBasis::ComputeXYRange() {
  
  xmin = 1.e30;
  xmax = -1.e30;
  ymin = 1.e30;
  ymax = -1.e30;

  int nx = Nx();
  int ny = Ny();
  
  int margin_low  = 2; // for 2nd order
  int margin_high = 1; // for 2nd order
  
  for(int j=margin_low;j<ny-margin_high;j++) {
    for(int i=margin_low;i<nx-margin_high;i++) {
      if(F(i,j)==-1 && F(i-1,j)==-1 && F(i,j-1)==-1) {
	const double& x = X(i,j);
	const double& y = Y(i,j);
	if(x<xmin) xmin = x;
	if(x>xmax) xmax = x;
	if(y<ymin) ymin = y;
	if(y>ymax) ymax = y;	
      }
     }
  }
  

  cout << "xmin xmax = " << xmin << " " << xmax << endl;
  cout << "ymin ymax = " << ymin << " " << ymax << endl;
}

static Vect SortedUniqueValues(const Mat& M) {
  int n = M.SizeX()*M.SizeY();
  const double* v0 = M.Data();
  double* v1 = new double[n];
  double* v2 = new double[n];
  
  memcpy(v1,v0,sizeof(double)*n);
  sort(v1,v1+n);
  
  int j=0; 
  v2[j]=v1[0];
  for(int i=1;i<n;i++) {
    if(v2[j]!=v1[i])
      v2[++j]=v1[i];
  }
  Vect R(j+1);
  memcpy(R.NonConstData(),v2,sizeof(double)*(j+1));
  
  delete [] v1;
  delete [] v2;
  
  return R;
}



void NURBSBasis::Acc_ComputeIndices() {
  
  cout << "Acc_ComputeIndices ..." << endl;
  
  if(I.SizeX() != X.SizeX() || I.SizeY() != X.SizeY())
    AllocateParameters();
  
  ComputeXYRange(); 
  
  int nx_vertices = Nx();
  int ny_vertices = Ny();
  
  Acc_X = SortedUniqueValues(X);
  Acc_Y = SortedUniqueValues(Y);
  
  int nx_table = Acc_X.size();
  int ny_table = Acc_Y.size();
  
  cout << "index table size = " << nx_table << "x" << ny_table << endl;
  
  Acc_MinX.allocate(nx_table,ny_table);
  Acc_MinY.allocate(nx_table,ny_table);
  Acc_MaxX.allocate(nx_table,ny_table);
  Acc_MaxY.allocate(nx_table,ny_table);
    
  const double *ux = Acc_X.Data();
  const double *uy = Acc_Y.Data();
  
  for(int i=0;i<nx_table;i++) {
    
    const double& x = ux[i];
    
    for(int j=0;j<ny_table;j++) {
      
      const double& y = uy[j];
      
      int kmin=100000;
      int lmin=100000;
      int kmax=0;
      int lmax=0;
      int nhits = 0;
      
      // find first associated vertex 	
      for(int k=0;k<nx_vertices;k++) {
	
	bool has_been_hit_at_k = false;
	for(int l=0;l<ny_vertices;l++) {
	  
	  if(NF2D(x,y,k,l)>0) {
	    
	    has_been_hit_at_k = true;
	    
	    if(k<kmin) kmin = k;
	    if(l<lmin) lmin = l;	      
	    if(k>kmax) kmax = k;
	    if(l>lmax) lmax = l;	      
	    nhits ++;
	    
	    if(nhits==9) break;
	    
	  }else{ // no hit
	    if(has_been_hit_at_k) break; // 
	  }
	}
	if(nhits==9) break;
      }
      
      if(nhits>0) {
	//cout << "dk dl = " << kmax-kmin << " " << lmax-lmin << endl;
	Acc_MinX(i,j)=kmin;
	Acc_MinY(i,j)=lmin;
	Acc_MaxX(i,j)=kmax+1; // exclusive upper bound
	Acc_MaxY(i,j)=lmax+1;
      }else{
	Acc_MinX(i,j)=0;
	Acc_MinY(i,j)=0;
	Acc_MaxX(i,j)=nx_vertices;
	Acc_MaxY(i,j)=ny_vertices;
      }
      
      
      
    }
  }
  acc_is_synchronized = true;
  cout << "done with Acc_ComputeIndices" << endl;
  
}
 

bool NURBSBasis::IsOK() const {
  //cout << "in NURBSBasis::IsOK ... " << endl;
  size_t nx = X.SizeX();
  size_t ny = X.SizeY();
  

  if(  ( Y.SizeX() != nx ) 
       || ( Y.SizeY() != ny ) 
       || ( F.SizeX() != nx ) 
       || ( F.SizeY() != ny ) 
       ) {
    cout << "ERROR in NURBSBasis::IsOK : Size of matrices is not correct " << endl;
    return false;
  }
  for( size_t i=0;i<nx;i++) {
    for( size_t j=0;j<ny;j++) {
      if(j>0 && Y(i,j)<= Y(i,j-1)) {
	cout << "ERROR in NURBSBasis::IsOK : Y(" << i << "," << j << ")=" << Y(i,j) << " <= " << " Y(" << i << "," << j-1 << ")=" << Y(i,j-1) << endl;
	return false;
      }
      if(i>0 && X(i,j)<= X(i-1,j)) {
	cout << "ERROR in NURBSBasis::IsOK : X(" << i << "," << j << ")=" << X(i,j) << " <= " << " X(" << i-1 << "," << j << ")=" << X(i-1,j) << endl;
	return false;
      }
      int fixed = F(i,j);
      if( (fixed < -2) || (fixed >= int(nx*ny) )) {
	cout << "ERROR in NURBSBasis::IsOK :  F(" << i << "," << j << ")=" << fixed << endl;
	return false;
      }
      if( fixed >= 0 ) {
	int k = fixed%nx;
	int l = fixed/nx;
	if( F(k,l) != -1 ) {
	  cout << "ERROR in NURBSBasis::IsOK :  point (" << i << "," << j << ") is anchored to point (" << k << "," << l << ") which is not a free param (F=" << F(k,l) << ")"  << endl;
	  return false;
	}
      }
    }
  }
  //cout << "yes!" << endl;
  return true;
  
}

bool NURBSBasis::InsertColumn(int i_after,int j_first, int j_last , bool shift_right) {

  cout << "NURBSBasis::InsertColumn after i=" << i_after << " for " << j_first << "<=j<=" << j_last << endl;
  
  int nx = Nx();
  int ny = Ny();

  if((i_after<0) || (i_after>=nx) || (j_first > j_last) || (j_first<0) || (j_last>=ny) ) {
      
    cout << "NURBSBasis::InsertColumn problem with indices" << endl;
    return false;
  }
    
  
  
  int new_nx = nx + 1;
  Mat  new_X(new_nx,ny);
  Mat  new_Y(new_nx,ny);
  IntMat new_F(new_nx,ny);
  
  int offset = 0;
  for(int i=0;i<nx;i++) {
    
    if(i>i_after) offset = 1;
    for(int j=0;j<ny;j++) {
      new_X(i+offset,j) = X(i,j);
      new_Y(i+offset,j) = Y(i,j);
      
      int fixed = F(i,j);
      
      if(fixed<0) {
	new_F(i+offset,j) = fixed; 
      }else{
	// have to adapt the index of the other vertex
	int i_other = fixed%nx; 
	if(i_other>i_after)
	  new_F(i+offset,j) = fixed+1;
	else
	  new_F(i+offset,j) = fixed;
      }
      
    }
  }

  for(int j=0;j<ny;j++) {
    if(i_after+2<new_nx) {
      new_X(i_after+1,j) = 0.5*(new_X(i_after,j)+new_X(i_after+2,j));
      new_Y(i_after+1,j) = 0.5*(new_Y(i_after,j)+new_Y(i_after+2,j));
    }else{
      new_X(i_after+1,j) = new_X(i_after,j);
      new_Y(i_after+1,j) = new_Y(i_after,j);
    }

    new_F(i_after+1,j) = -1;
  }
  
  X = new_X;
  Y = new_Y;
  F = new_F;
  
  AllocateParameters();
  Acc_Reset();
  
  // now delete parts of column
  if(j_first>0) {
    DeleteColumn(i_after+1,0,j_first-1,shift_right);
  }
  if(j_last<ny-1) {
    DeleteColumn(i_after+1,j_last+1,ny-1,shift_right);
  }
  
  return true;
}


bool NURBSBasis::InsertRow(int j_after, int i_first, int i_last , bool shift_up) {

  cout << "NURBSBasis::InsertRow after j=" << j_after << " for " << i_first << "<=i<=" << i_last << endl;
  
  int nx = Nx();
  int ny = Ny();

  if((j_after<0) || (j_after>=ny) || (i_first > i_last) || (i_first<0) || (i_last>=nx) ) {
      
    cout << "NURBSBasis::InsertRow problem with indices" << endl;
    return false;
  }
    
  
  
  int new_ny = ny + 1;
  Mat  new_X(nx,new_ny);
  Mat  new_Y(nx,new_ny);
  IntMat new_F(nx,new_ny);
  
  int offset = 0;
  for(int j=0;j<ny;j++) {
    
    if(j>j_after) offset = 1;
    
    for(int i=0;i<nx;i++) {
      new_X(i,j+offset) = X(i,j);
      new_Y(i,j+offset) = Y(i,j);
      
      int fixed = F(i,j);
      if(fixed<0) {
	new_F(i,j+offset) = fixed; 
      }else{
	// have to adapt the index of the other vertex
	int j_other = fixed/nx; 
	if(j_other>j_after)
	  new_F(i,j+offset) = fixed+nx;
	else
	  new_F(i,j+offset) = fixed;
      }
    }
  }
  
  for(int i=0;i<nx;i++) {
    if(j_after+2<new_ny) {
      new_X(i,j_after+1) = 0.5*(new_X(i,j_after)+new_X(i,j_after+2));
      new_Y(i,j_after+1) = 0.5*(new_Y(i,j_after)+new_Y(i,j_after+2));
    }else{
      new_X(i,j_after+1) = new_X(i,j_after);
      new_Y(i,j_after+1) = new_Y(i,j_after);
    }
    new_F(i,j_after+1) = -1;
  }
  
  X = new_X;
  Y = new_Y;
  F = new_F;
  
  AllocateParameters();
  Acc_Reset();

  if(1) {
    // now delete parts of row
    if(i_first>0) {
      DeleteRow(j_after+1,0,i_first-1,shift_up);
    }
    if(i_last<nx-1) {
      DeleteRow(j_after+1,i_last+1,nx-1,shift_up);
    }
  }
  
  
  return true;
}

bool NURBSBasis::DeleteColumn(int i_where,int j_first, int j_last , bool shift_right) {

  cout << "NURBSBasis::DeleteColumn at i=" << i_where << " for " << j_first << "<=j<=" << j_last << endl;
  
  int nx = Nx();
  int ny = Ny();

  if(i_where<0 || i_where>=nx || j_first > j_last || j_first<0 || j_last>=ny ) {
     
      cout << "NURBSBasis::DeleteColumn problem with indices" << endl;
      return false;
    }

  

  

  if(shift_right) {
    for(int j = j_first; j<=j_last ; j++) {
      double last_x =  X(nx-1,j) + (X(nx-1,j)-X(nx-2,j)); // push away last vertex
      
      // shift by one slot the others
      for(int i=i_where; i<nx-1; i++) {
	X(i,j) = X(i+1,j);
	Y(i,j) = Y(i+1,j);
	F(i,j) = F(i+1,j);
      }
      
      X(nx-1,j) = last_x;
    }
  }else{ // shift to the left
    
    for(int j = j_first; j<=j_last ; j++) {
      double first_x =  X(0,j) - (X(1,j)-X(0,j)); // push away first vertex
      
      // shift by one slot the others
      for(int i=i_where; i>0; i--) {
	X(i,j) = X(i-1,j);
	Y(i,j) = Y(i-1,j);
	F(i,j) = F(i-1,j);
      }
      
      X(0,j) = first_x;
      
    }

  }
    
   
  AllocateParameters();
  Acc_Reset();
    
  return true;
}

bool NURBSBasis::DeleteRow(int j_where,int i_first, int i_last , bool shift_up) {

  cout << "NURBSBasis::DeleteRow at j=" << j_where << " for " << i_first << "<=i<=" << i_last << endl;
  
  int nx = Nx();
  int ny = Ny();

  if(j_where<0 || j_where>=ny || i_first > i_last || i_first<0 || i_last>=nx ) {
      
      cout << "NURBSBasis::DeleteRow problem with indices" << endl;
      return false;
    }

    

  
    
  if(shift_up) {
    for(int i = i_first; i<=i_last ; i++) {
      
      
      double last_y =  Y(i,ny-1) + (Y(i,ny-1)-Y(i,ny-2)); // push away last vertex
      
      // shift by one slot the others
      for(int j=j_where; j<ny-1; j++) {	
	X(i,j) = X(i,j+1);
	Y(i,j) = Y(i,j+1);
	F(i,j) = F(i,j+1);
      }
      Y(i,ny-1) = last_y;
    }
  }else{ // shift downward
    
    for(int i = i_first; i<=i_last ; i++) {
      
      double first_y =  Y(i,0) - (Y(i,1)-Y(i,0)); // push away first vertex
      
      // shift by one slot the others
      for(int j=j_where; j>0; j--) {
	X(i,j) = X(i,j-1);
	Y(i,j) = Y(i,j-1);
	F(i,j) = F(i,j-1);
      }
      
      Y(i,0) = first_y;
    }

  }
    
   
  AllocateParameters();
  Acc_Reset();
    
  return true;
}

  void NURBSBasis::Acc_Reset() {
    acc_is_synchronized = false;
  }
  
  bool NURBSBasis::Acc_IsSync() const {
    return acc_is_synchronized;
  }
