#ifndef NURBSBASIS__H

class Vect;
class Mat;
class IntMat;
class SparseVect;
class Grid2DFunction;

/* 
   2D basis using "Non Uniform Rational B-Spline"
   
   - X,Y,and F are matrices that define entirely the basis 
   (x,y coordinates of vertices with or without free or fixed control points)
   (X,Y,F must have the same size, see below for the exact definition of F)
   
   - A matrix I of parameter indices associated to the vertices is computed by 
   the routine AllocateParameters() which also allocate the vector of parameters P.
   This routine must be called after X,Y,F are filled.
   
   - The routine Derivatives(x,y) return a sparse vector 

   - The routine Value(x,y) simply returns the scalar product  P * Derivatives(x,y)

   - The Fit routine uses Derivatives as expected

   - There are few routines to generate X,Y,F : Uniform,FixPoint
 
   - There is a set of buffer & routines to accelerate the thing (Acc_*)
   
*/


#define NURBS_FORCE_NULL_OUTSIDE_BOUNDARIES  


class NURBSBasis {

 private :
  bool acc_is_synchronized; 
  
  
public :
  
  // The following 3 matrices define the basis :

  Mat X;     // x coordinate of control points
  Mat Y;     // y coordinate of control points
  IntMat F;  // tell whether there is a free parameter associated to the vertex i,j or not
  /*
    - F(i,j)=-2 means that this is a boundary vertex for which there is no control point (but the vertex is nonetheless needed to define the shape of the spline functions)
    - F(i,j)=-1 means that this is a control point with a free parameter associated to it
    - F(i,j)=index with index>=0 means that this is a control point which value is equal to that of another control point at the address given by index (index=k+nx*p), so there is no parameter associated to this point)
  */
  


  IntMat I;  // parameter indices of control points: -1 means no parameter (happens for vertices on the boundary)
  
  Vect P;    //  vector of parameters : "flux" scalar values of the active control points
  
  double xmin,xmax; // corners of rectangle including the area where the basis is defined 
  double ymin,ymax; // corners of rectangle including the area where the basis is define

  IntMat Acc_MinX; // to get rapidely the first index along x
  IntMat Acc_MinY; // to get rapidely the first index along y 
  IntMat Acc_MaxX; // to get rapidely the last index along x
  IntMat Acc_MaxY; // to get rapidely the last index along y 
  Vect   Acc_X; // sort (unique) values of x and y coordinates of vertices 
  Vect   Acc_Y; // sort (unique) values of x and y coordinates of vertices 
  
 

  NURBSBasis(); 
  int Nx() const;
  int Ny() const;
  void AllocateParameters();


  // routines to set up the basis :
  // forces the value of the control point (i,j) to that of another point, 
  // hence reducing the number of free parameters.
  // the other point can be given (i_of_other,j_of_other) or found as the nearest in the direction
  // 0=right
  // 1=top
  // 2=left
  // 3=down
  bool FixPoint(int i, int j, int i_of_other=-1, int j_of_other=-1, int direction_for_nearest=2);
  
  // nx and ny here are the number of parameters !!!
  // so the number of vertices = (nx+2)*(ny+2)
  // first_x , last_x ... define the valid area of the model 
  void Uniform(int nx, const double& first_x, const double& last_x, 
	       int ny, const double& first_y, const double& last_y );
  
  // delete column at i=i_where, from j_first to j_last (included)
  // actually vertices are not deleted, they are shifted and the last column is then fixed to the previous
  bool DeleteColumn(int i_where,int j_first, int j_last , bool shift_right = true); 

  bool DeleteRow(int j_where,int i_first, int i_last , bool shift_up = true); 
  
  bool InsertColumn(int i_after, int j_first, int j_last , bool shift_right = true); 

  bool InsertRow(int j_after, int i_first, int i_last , bool shift_up = true); 

  
  /* This is the recursive routine that define NURBS 
     It gives the value at x of the basis function of index 'i'
     x is the coord
     i the index of the control point which weight we want to compute
     d the order of the NURBS (=degree of interpolation polynomial)
     k the (1D) 'knot' vector of control points (it defines the "shape" of the function)
     n the size of k
   */
  double NF1D(const double &x, const int i, const int d, const double* k, const int n) const;
  
  /* 
     Returns NF1D(x,i,...)*NF1D(y,j,...) with the appropriate knot vectors (using X,Y,F...)
  */
  double NF2D(const double &x, const double &y, int i, int j) const;
    
 
  /* Get rapidely the indices of parameters that affect value at point x,y
   */
  bool Acc_GetIndices(const double& x, const double& y, int& imin, int& imax, int& jmin, int& jmax) const;
  bool Acc_GetLeadingVertex(const double& x, const double& y, int& i, int& j) const;
  
  /* Return sparse vector of derivatives at point x,y
   */
  SparseVect Derivatives(const double& x, const double& y) const;
  
  /* Return value = parameters * derivatives at point x,y
   */
  double Value(const double& x, const double& y) const;

 
  bool Fit(int nx, const double*x, int ny, const double* y,  double** f);

  bool Fit(Grid2DFunction &func);

  
  void ComputeXYRange();

  /* Computes  Acc_MinX,... Acc_UX,... Must be called once XYF are set
   */
  void Acc_ComputeIndices();
  
  /* Tell acc indices have to be recomputed 
   */
  void Acc_Reset();
  
  /* Check whether acc stuff is synchronized or not
   */
  bool Acc_IsSync() const;
  
  /* Checks that the vertices (the "tesselation") is fine
   */
  bool IsOK() const;
  
  
};





#endif
