#include "saltnirmodel.h"
#include "integflux.h"
#include <snfitexception.h>
#include <iostream>
#include <vutils.h> // DArrayMedian 
#include <fileutils.h> // FileExists
#include <pathutils.h> 

using namespace std;

static double DayWithMaxFlux(const MeasuredPointList &List, double &maxflux) {
  
  // evaluate median error
  double *errors = new double[List.size()];
  int i=0;
  for (MeasuredPointCIterator iti=List.begin(); iti!= List.end(); ++iti, ++i) {
    errors[i] = (*iti)->FluxErr();
  }
  double median_error = DArrayMedian(errors, List.size());
  delete [] errors;
  
  
  double maxDay = 0;
  maxflux = 0;
  for (MeasuredPointCIterator iti=List.begin(); iti!= List.end(); ++iti) {
    const MeasuredPoint *data= *iti;
    if( (data->FluxErr()<3.*median_error) && (data->Flux()>maxflux) )  {
      maxDay = data->Day();
      maxflux = data->Flux();
    }
  }
  return maxDay;
}

#include <algorithm> // min & max

SaltNirModel::SaltNirModel()
{
  // load templates
  string directory = GetDataPath()+"/saltnir/";
  string meanCorrFileName = directory+"meancorr.list";
  if (!FileExists(meanCorrFileName))
    {
      cout << " cannot find " << meanCorrFileName << endl;
      throw(SnfitException("cannot load SaltNir templates (in $PATHMODEL/saltnir or $SALTPATH/saltnir) ..."));
    }
  meanCorr.Read(meanCorrFileName);
  dm15Corr.Read(directory+"dm15corr.list");
  sed.Read(directory+"ref_sed");
  // read color law
  {
    string filename = directory+"/color_correction.dat";
    if(! FileExists(filename)) {
      cerr << "cannot find " << filename << endl;
      throw(SnfitException(" Cannot find "+filename));
    }
    extinction_law.Read(filename);
  }


  phaseMin = max(meanCorr.Xmin(),dm15Corr.Xmin());
  phaseMin = max(phaseMin,sed.Xmin());
  phaseMax = min(meanCorr.Xmax(),dm15Corr.Xmax());
  phaseMax = min(phaseMax, sed.Xmax());
  // if you change the bluecutoff, have a look at Flux()
  double lambdaMin = max(meanCorr.Ymin(), dm15Corr.Ymin())*0.9;
  BlueCutoff = max(lambdaMin,sed.Ymin());
  RedCutoff= 18000;// j'en ai marre.
  params.SetName("SaltNirModel");
  pDayMax  = params.LocateOrAddNonConstParam("DayMax");  
  pRedshift  = params.LocateOrAddNonConstParam("Redshift");  
  pColor = params.LocateOrAddNonConstParam("Color");  
  pColor->val =0;
  pX1 = params.LocateOrAddNonConstParam("X1");  
  pX1->val = 0;
  pX0 = params.LocateOrAddNonConstParam("X0");  

}


double SaltNirModel::GetPhase(double Date) const
{
  return (Date-pDayMax->val )/(1+pRedshift->val);  
}


bool SaltNirModel::CanComputeFluxFor(const Filter& observerfilter, 
				    bool quiet) const
{
  double mean_lambda=observerfilter.Mean()/(1+pRedshift->val);
  return (mean_lambda<RedCutoff&& mean_lambda>BlueCutoff);
}

void SaltNirModel::InitParams(const MeasuredData *Data)
{
  // compute first guess of maxDay;
  double maxDay = 0;
  //  size_t bandwithbestsampling=0;
  if( Data->NBands()) {
    double maxflux=0;
    for (size_t i=0; i < Data->NBands(); ++i) {   
      double maxflux_for_band=0;
      double maxDay_for_band = DayWithMaxFlux(Data->PointList(i),maxflux_for_band);
      if(maxflux_for_band>maxflux) {
	maxflux=maxflux_for_band;
	maxDay=maxDay_for_band;
      }
    }
  // 
  if(!pDayMax->isfixed) {
     pDayMax->val = maxDay;
    pDayMax->step = 1;
    pDayMax->minv = maxDay-30;
    pDayMax->maxv = maxDay+30;
  }
  }

  //
  if(!pX0->isfixed) {
    pX0->val = 1;
    
    if( Data->NBands()) {
      // compute first guess for dflux 
      
      
      double sumModelFluxes = 0;
      double sumDataFluxes = 0;
      for(size_t i=0; i < Data->NBands(); ++i) {
	const MeasuredPointList &datapoints = Data->PointList(i);
	const Filter* observerfilter = &(datapoints.GetEffectiveFilterWithMilkyWay());
	for (MeasuredPointCIterator iti = datapoints.begin(); iti!= datapoints.end(); ++iti)
	  {
	    const MeasuredPoint *mp= *iti;
	    sumDataFluxes  += mp->Flux();
	    sumModelFluxes += Flux(observerfilter,mp->Day());
	  }
      }
      if(sumModelFluxes<=0) {
	cerr << "in SaltNirModel::InitParams model errors to large for first guess of X0" << endl;
	sumDataFluxes = 0;
	sumModelFluxes = 1;
      }
      pX0->val = sumDataFluxes/sumModelFluxes; 
      pX0->step = 0.2 * pX0->val;
      //pX[0]->minv = 0.;
      //pX[0]->maxv = 100. * pX[0]->val;
      pX0->minv = 0;
      pX0->maxv = 10000. * pX0->val;
    }
  }



  pX1->minv = -3;
  pX1->maxv = 3;
  pX1->step = 0.3;
  if (!pX1->isfixed)  pX1->val = 0;
  
    pColor->step = 0.1;
  pColor->minv = -4;
  pColor->maxv = 4;
    
  if( ! pColor->isfixed) {
    pColor->val = 0.;
  }else{
    pColor->minv += pColor->val;
    pColor->maxv += pColor->val;
  }


  Update();
  
}


double SaltNirModel::Flux(const Filter *F, const double observerday) const
{
  double z = pRedshift->val;
  double phase = (observerday-pDayMax->val )/(1+z);
  if (phase < phaseMin || phase > phaseMax) return 0;
  double mean_lambda = F->Mean()/(1+z);
  
  double flux = IntegFlux(sed.ConstXSlice(phase), *F, z);
  
  /* We decide that he corrections are constant when going bluer than
    the bluest filter (for a small range though) */

  double lambda_for_corr = max(mean_lambda,meanCorr.Ymin());
  //  double dmag = meanCorr.Value(phase,mean_lambda)+(dm15-1.1)*dm15Corr.Value(phase,mean_lambda);
  double dmag = meanCorr.Value(phase,lambda_for_corr)+pX1->val*dm15Corr.Value(phase,lambda_for_corr);

  flux *= exp(-0.92103*dmag);
  flux *= extinction_law.extinction_lambda(pColor->val,mean_lambda);
  flux *= ( (1+z)*(pX0->val)); // (1+z) ?? ask Julien
  return flux;
}

static double  constant =  log(10.0)/2.5;

SimpleExtinction::SimpleExtinction(const string &FileName)
  : extinction(FileName)
{
  
}

void SimpleExtinction::Read(const string &FileName)
{
  extinction.Read(FileName);
}

#include <assert.h>
double SimpleExtinction::extinction_lambda(const double &Color, 
					   const double &Lambda) const
{
  assert (Lambda>=extinction.Xmin() && Lambda<=extinction.Xmax());
  return exp(-Color*constant*extinction.Value(Lambda));
}
