
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

#include "measureddata.h"
#include "saltmodel.h"
#include "salt2model.h"
#include "fileutils.h"
#include "dictfile.h"
#include "magsys.h"



static void usage(const char *pg)
{ 
  std::cerr << pg << " dumps peak observed magnitude" << endl;
  std::cerr << "usage : " << pg << " <lightcurve> <lightcurve> -P <parameter_filename> " << std::endl;
  std::cerr << "option : -d date (mag at this date instead of peak B band" << std::endl;
  std::cerr << "         -p phase (mag at this phase=(date-date_of_max)/(1+z) instead of peak B band" << std::endl;
  std::cerr << "<parameter_filename> is the path of the output file of snfit:  result_salt.dat or result_salt2.dat" << endl; 
  exit(EXIT_FAILURE);  
}

int main(int nargs, char **args) {
  if (nargs <3 ) usage(args[0]);
  
  std::vector<std::string> lc;
  
  string parameters_filename = "";

  double date = -10000;
  double phase = -10000;
  string magsys = "AB";
  
  for (int i=1; i<nargs; ++i)
    {
      char *arg = args[i];
      if (arg[0] != '-')
	{
	  lc.push_back(arg);
	  continue;
	}
      switch (arg[1])
	{
	case 'P' : parameters_filename = args[++i]; break;
	case 'd' : date = atof(args[++i]); break;
	case 'p' : phase = atof(args[++i]); break;
	case 'm' : magsys = args[++i]; break;
	default : usage(args[0]);
	}
    }
  
  if(parameters_filename == "") {
    cout << "ERROR : missing parameters_filename" << endl;
    usage(args[0]);
  }
  if(lc.size()==0) {
    cout << "ERROR : missing light curve file" << endl;
    usage(args[0]);
  }
  cout << "magsys : " << magsys << endl;

  MeasuredData md(lc,false);
  //DictFile file(lc[0]);
  
    
  
  
  // read model name
  string model_name = "";
  {
    std::ifstream stream(parameters_filename.c_str());
    if(!stream) {
      cerr << "error: cannot read paramete file " << parameters_filename << endl;
      usage(args[0]);
    }
    
    if(!(stream >> model_name)) {
      cerr << "error: cannot read model name in " << parameters_filename << endl;
      exit(EXIT_FAILURE);
    }
    stream.close();
  }
  Model2Fit *model=0;
  if(model_name=="SaltModel")
    model = new SaltModel();
  if(model_name=="Salt2Model")
    model = new Salt2Model();
  if(!model) {
    cerr << "error: cannot deal with model '" << model_name << "'" << endl;
    exit(-1);
  }
  model->SetWavelengthRange(0.,100000.);
  model->ReadParams(parameters_filename);
  
  // for salt1 backward compatibility
  if(model->params.LocateParam("Redshift")->err<0) {
    cout << "warning : redshift parameter is not set in file, set it to " << md.Zhelio() << endl;
    model->SetParameterValue("Redshift",md.Zhelio());
  }
  
  
  double day  = model->params.LocateParam("DayMax")->val;
  if(date!=-10000)
    day=date;
  if(phase!=-10000) {
    day += phase*(1+model->params.LocateParam("Redshift")->val);
  }
  
  for(size_t b=0;b<md.NBands();b++) {
    double mag = 99;
    const Filter& filter = md.PointList(b).GetEffectiveFilter();
    
    if(model->CanComputeFluxFor(filter)) {
      double zp   = GetMagSys(magsys)->ZeroPoint(filter);
      double flux = model->Flux(&md.PointList(b).GetEffectiveFilterWithMilkyWay(),day);
      if(flux>0)
	mag  = -2.5*log10(flux)+zp;
    }
    cout << "MAG " << filter.Name() << " " << mag << " " << 1.086*model->params.LocateParam("X0")->err/model->params.LocateParam("X0")->val << endl;
  }
  return EXIT_SUCCESS;
}
